#ifndef _ARRAY_H_
#define _ARRAY_H_

#include "Exception.h"
#include "String.h"


/**
  Array-Template.
  Verwaltet Kopien von Objekten derselben Klasse.

  Das Array kann sich bei Bedarf (es werden mehr Elemente
  hinein getan als urspruenglich angegeben) dynamisch
  vergroessern.
  Diese Klasse erbt absichtlich nicht von StorableObject.
  Gruende: 

  \begin{itemize}
  \item  Die Klasse ist eine template-Klasse.
  \item  StorableObject gibt vor, dass toString() in ihren Kindern
      implementiert werden muss.
  \item  Ein Array soll auch nicht-Objekte verwalten koennen. Fuer 
      nicht-Objekte kann man keine Methoden deklarieren.
  \end{itemize}

  {\small {\bf Quelltext}

  [\URL[Header]{src/Array.h} | [\URL[Implementierung]{src/Array.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Array.h,v $
  *   $Revision: 1.29 $
  *   $Date: 1998/08/14 13:19:55 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
template <class T>
class Array 
{
public:

  /** @name Fehlerklassen */
   //@{
    /** Fehlerklasse fuer allgemeine Fehler.
      Ausgeloest wird er z.B. bei ungueltige Angaben fuer die Arraygroesse. */
    class ArrayException: public Exception {
    public:
      ArrayException(const String& aDescription )
        : Exception( aDescription ) {}
    };

    /** Fehlerklasse fuer Indexfehler. */
    class IndexOutOfBoundsException: public ArrayException {
    public:
      IndexOutOfBoundsException(const String& aDescription )
        : ArrayException( aDescription ) {}
      virtual const String className() const 
      {
        return "IndexOutOfBoundsException";
      }
    };
    
  //@}
  
  /** @name Array anlegen und zerst\"oren */
  //@{
    /** Neues, leeres Array anlegen.
      @param size initiale Groesse des Arrays
      @exception ArrayException falls size<0 */
    Array(int size=5);

    /** Copy-Konstruktor. */
    Array( const Array<T>& otherArray);

    /** Zuweisungsoperator */
    const Array<T>& operator= ( const Array<T>& otherArray);

    /* Destruktor. */
    ~Array();
  //@}
    
  /** @name Elemente lesen*/
  //@{  
    /** Der eigentliche Index-Operator. Es ist sowohl lesender als auch
      schreibender Zugriff m\"oglich.
      @param index Index des Objekts im Array
      @exception IndexOutOfBoundsException, falls index ungueltig ist 
        ( <0 oder >elementsCount() ). 
      @return Schreibbare Referenz.*/
    T& objectAt( int index ) const;

    /** Index-Operator mit Klammern. Ruft intern \Ref{objectAt}() auf.
      @param index Index des Objekts im Array 
      @exception IndexOutOfBoundsException, falls index ungueltig ist 
        ( <0 oder >elementsCount() ). 
      @return Schreibbare Referenz.*/
    T& operator[] ( int index ) const;
  //@}
  
    
  /** @name Elemente ver\"andern */
  //@{
    /** Fuegt ein Element am Ende des Arrays ein.
      Falls das Array voll ist, wird es vergroessert. */
    void addObject( const T& newObject );

    /** Fuegt die Eintraege des uebergebenen Arrays am Ende des Arrays ein.
      Falls das Array voll ist, wird es vergroessert. */	
    void addObjects( const Array<T>& newObjects);

    /** Speichert ein Objekt an einer angegebenen Stelle im Array.
      @param index Der Index, an dem das Objekt erscheinen soll.
      @param newObject Objekt, das gespeichert werden soll
      @exception IndexOutOfBoundsException, 
        falls der angegebene Index ungueltig ist. */
    void setObjectAt( const T& newObject, int index );
    
    /** Loescht den Eintrag an der Stelle index. 
      Nachfolgende Eintraege werden nach vorne verschoben.  
      @param index index des zu loeschenden Objekts.
      @exception IndexOutOfBoundsException falls der angegebene Index 
        ungueltig ist. */
    void removeObjectAt( int index );

    /** Loescht alle Eintraege.
      Der belegte Speicher wird {\bf noch nicht} zurueckgegeben. */
    void removeAll();
  
  //@}
    
    
  /** @name Sonstige Methoden */
  //@{
    /** Anzahl der Elemente */
    int count() const;

    /** Groesse des Arrays. 
      Fuer diese Anzahl von Objekten wurde bereits Speicherplatz reserviert. */
    int size() const;
  
    /** Ausgabe auf Stream */
    void printOn( ostream& outputStream = cout) const;
  //@}
    
private:
  ///
  T* theElements;
  ///
  int theMaximumSize;
  ///
  int theElementsCount;
  
  /** Groesse des Array veraendern.
    Wird intern von addObject und addObjects benutzt um das Array zu vergroessern.
    @exception ArrayException falls die neue Groesse kleiner als 0 ist       */
  void resize(int newSize);
};

/* Konstruktor */
template <class T>
inline Array<T>::Array( int size )
{
  theMaximumSize = size;
  if ( theMaximumSize >= 0 ) {
    theElements = new T[theMaximumSize];
  }
  else {
    theElements = NULL;
    throwException( ArrayException( "Arraysize less than 0" ) );
  }
  theElementsCount = 0;
}


/* Copy-Konstruktor. */
template <class T>
inline Array<T>::Array( const Array<T>& otherArray )
{
  theMaximumSize  = otherArray.theMaximumSize;
  theElementsCount = otherArray.theElementsCount;
  
  if( theElementsCount >= 0 ) {
    theElements = new T[theMaximumSize];
    for (int i = 0; i < theElementsCount; i++ ) {
      theElements[i] = otherArray.theElements[i];
    }
  }
  else {
    theElements = NULL;
    throwException( ArrayException( "Arraysize less than 0" ) );
  }
}

  
/* Zuweisungsoperator */
template <class T>
inline const Array<T>& Array<T>::operator= ( const Array<T>& otherArray )
{
  if ( this == &otherArray ) {
    return *this;
  }
  
  theElementsCount = otherArray.theElementsCount;
  
  if ( otherArray.theMaximumSize > theMaximumSize ) {
    delete [] theElements;
    theMaximumSize = otherArray.theMaximumSize;
    theElements = new T[theMaximumSize];
  }
  for ( int i = 0; i < theElementsCount; i++ ) {
    theElements[i] = otherArray.theElements[i];
  }
  
  return *this;
}

  
/* Und der Destruktor. */
template <class T>
inline Array<T>::~Array()
{
  if( theElements ) {
    delete [] theElements;
  }
}


template <class T>
inline T& Array<T>::operator[] ( int index ) const
{
  return objectAt( index );
}


/* Der eigentliche Index-Operator. */
template <class T>
inline T& Array<T>::objectAt( int index ) const
{
  if( index < 0 || index >= theElementsCount ) {
    throwException( IndexOutOfBoundsException( 
        "Index out of bounds with index " + String( index ) +
        " ( array size " + String( theElementsCount ) + " )" ) );
  }
  return theElements[index];
}


/* Fuegt ein Element am Ende des Arrays ein. */
template <class T>
inline void Array<T>::addObject( const T &newObject )
{
  if( theElementsCount >= theMaximumSize ) {
    resize( theMaximumSize + 5 );
  }
  theElements[theElementsCount] = newObject;
  theElementsCount++;
}

	
/* Fuegt die Eintraege des uebergebenen Arrays am Ende des Arrays ein. */
template <class T>	
inline void Array<T>::addObjects( const Array<T>& otherArray )
{
  if( theElementsCount + otherArray.theElementsCount >= theMaximumSize ) {
    resize( theMaximumSize + otherArray.theElementsCount + 5 );
  }
  for (int i = 0; i < otherArray.theElementsCount; i++) {
    theElements[theElementsCount] = otherArray.theElements[i];
    theElementsCount++;
  } 
}

/* Speichert ein Objekt an einer angegebenen Stelle im Array. */
template <class T>
inline void Array<T>::setObjectAt( const T& newObject, int index )
{
  if( index < 0 || index >= theElementsCount ) {
    throwException( IndexOutOfBoundsException( "Index out of bounds" ) );
  }
  theElements[index] = newObject;
}


/* Loescht den Eintrag an der Stelle index. Nachfolgende Eintraege werden
  nach vorne verschoben.  */
template <class T>
inline void Array<T>::removeObjectAt( int index )
{
  if(index < 0 || index >= theElementsCount ) {
    throwException( IndexOutOfBoundsException( "Index out of bounds" ) );
  }
  for( int i = index; i < theElementsCount - 1; i++ ) {
    theElements[i] = theElements[i+1];
  }
  theElementsCount--;
}


/* Loescht alle Eintraege. Der vom Array zur Speicherung der Verweise
   verwendet Speicherplatz wird dabei nicht reduziert. */
template <class T>
inline void Array<T>::removeAll()
{
  theElementsCount = 0;
}


/* Zahl der Elemente im Array */
template <class T>
inline int Array<T>::count() const
{
  return theElementsCount;
}

/* Groesse des Array */
template <class T>
inline int Array<T>::size() const
{
  return theMaximumSize;
}

/* Groesse des Array veraendern */
template <class T>
inline void Array<T>::resize( int newSize )
{
  if( newSize < 0 ) {
    throwException( ArrayException( "Arraysize less than 0" ) );
  }
  T *newElements = new T[newSize];
  for( int i = 0; i < theElementsCount; i++ )
    newElements[i] = theElements[i];
  if( theElements ) {
    delete [] theElements;
  }
  theElements = newElements;
  theMaximumSize = newSize;
}

template <class T>
void Array<T>::printOn(ostream& outputStream = cout) const
{
  outputStream << "( array with " << count() << " elements:";
  int i;
  for (i = 0; i < count(); i++) {
    outputStream  << endl << objectAt( i );
  }
  outputStream << " )" << endl;
}

template <class T>
inline ostream& operator << (ostream& outputStream, const Array<T>& array) 
{
  array.printOn( outputStream );
  return outputStream;
}




#endif
