
/*
 * Aktueller Stand
 *
 *   $RCSfile: Deliberator.cc,v $
 *   $Revision: 1.19 $
 *   $Date: 1998/08/14 13:20:13 $
 *   $Author: myritz $ 
 *
 */

#include "Deliberator.h"

#include "Common.h"
#include "InvalidIdException.h"
#include "String.h"
#include "Configuration.h"


// Inline Funktionen

inline int Deliberator::minConstraintUtility() const
{
  return globalConf->minConstraintUtility;
}

inline int Deliberator::minDesireUtility() const
{
  return globalConf->minDesireUtility;
}

// Andere Funktionen

Deliberator::Deliberator( 
  Effectors *anEffector )
  : theEffector( anEffector )
{
  initialize( optionIds(), constraintIds() );
}

Deliberator::Deliberator( 
  const Array<int>& optionIds, 
  const Array<int>& potentialConstraintIds,
  Effectors *anEffector )
  : theEffector( anEffector )
{
  initialize( optionIds, potentialConstraintIds );
}


void Deliberator::initialize( 
    const Array<int>& optionIds, 
    const Array<int>& potentialConstraintIds )
{
  INFO( Log::DELIBERATOR, 2, "ENTER Deliberator::initialize\n" );
  int i;
  // Option-Objekte erzeugen
  options = Array<Option *>( optionIds.count() );
  for( i = 0; i < optionIds.count(); i++ ) {
    INFO( Log::DELIBERATOR, 5, "Adding option " +
        String( optionIds[i] ) + "\n" );
    options.addObject( getOptionWithId( optionIds[i] ) );
  }
  
  // Constraint-Objekte erzeugen
  potentialConstraints = Array<Constraint *>( potentialConstraintIds.count() );
  for( i = 0; i < potentialConstraintIds.count(); i++ ) {
    INFO( Log::DELIBERATOR, 5, "Adding potentialConstraint " +
        String( potentialConstraintIds[i] ) + "\n" );
    potentialConstraints.addObject( getConstraintWithId( 
      potentialConstraintIds[i] ) );
  }
  
  currentIntention = new Option();
  previousIntention = new Option();
}


void Deliberator::startDeliberation()
{
  INFO( Log::DELIBERATOR, 2, "ENTER Deliberator::startDeliberation\n" );
  INFO( Log::DELIBERATOR, 5, "current WorldModel: " +
        globalWorldModel->toString() );

  int i;
  Array<Constraint *> requestedConstraints;
  Array<Option *> desires;
  
  delete previousIntention;
  previousIntention = currentIntention;
  
  INFO( Log::DELIBERATOR, 13,  "Previous Intention : " + 
      previousIntention->toString() + "\n" );
  if ( previousIntention->forceContinuation() == YES ) {
    // Nicht nachdenken, einfach weitermachen
    currentIntention = (Option *)previousIntention->copy();
    INFO( Log::DELIBERATOR, 8,  "Previous Intention : " + 
      previousIntention->toString() + "forces continuation." + "\n" );
  }
  else {
    // Richtig nachdenken, was zu tun ist
    
    currentIntention = new Option();
  
    // Nuetzliche Nebenbedingungen auswaehlen
    INFO( Log::DELIBERATOR, 8, "Starting to examine constraints\n" );
    for ( i = 0; i < potentialConstraints.count(); i++ ) {
      INFO( Log::DELIBERATOR, 9, "Examining constraint " );

      if ( potentialConstraints[i]->expectedUtility() >
           minConstraintUtility() ) {
        requestedConstraints.addObject( potentialConstraints[i] );
        INFO( Log::DELIBERATOR, 9, potentialConstraints[i]->toString() +
              " -- is requested\n" );
      } 
      else {
        INFO( Log::DELIBERATOR, 9, potentialConstraints[i]->toString() +
              " -- is rejected\n" );
      }
    }

    // Wuensche festlegen. Dafuer werden alle Handlungsoptionen auf
    // ihre Nuetzlichkeit unter den gegebenen, nuetzlichen 
    // Nebenbedingungen untersucht
    INFO( Log::DELIBERATOR, 8, "Starting to examine options\n" );
    for ( i = 0; i < options.count(); i++ ) {
      Option *option = (Option *)options[i]->copy();
      option->setConstraints( requestedConstraints );
      INFO( Log::DELIBERATOR, 9, "Examining option " );      
      if ( option->expectedUtility() > minDesireUtility() ) {
        desires.addObject( option );
        INFO( Log::DELIBERATOR, 9, option->toString() + " -- is desired\n" );
      }
      else {
        INFO( Log::DELIBERATOR, 9, option->toString() + " -- is rejected\n" );
        delete option;
      }
    }

    // Absicht festlegen. Das haengt von der Realisierbarkeit des
    // Wunsches und von der bisherigen Absicht ab.
    do {
      int desireIndex;

      delete currentIntention;
      try {
        desireIndex = indexOfBestDesire( desires );
        currentIntention = (Option *)desires[desireIndex]->copy();
        INFO( Log::DELIBERATOR, 9,  "Intending to realize " + 
            currentIntention->toString() + "\n" );
        delete desires[desireIndex];
        desires.removeObjectAt( desireIndex );
      }
      catch (...) {
        // es gibt keine Wuensche mehr. Dann die Standardoption nehmen.
        currentIntention = new Option();
        break;
      }
    } while ( !currentIntention->isPossible() );

    // Jetzt den Rest von desires abraeumen
    for ( int i = 0; i < desires.count(); i++ ) {
      delete desires[i];
    }

    // Noch generische Anpassungsmoeglichkeit einraeumen.
    currentIntention->adaptToPreviousIntention( previousIntention );
  }
  
  // Jetzt ist alles entschieden. Absicht umsetzen.
  
  theEffector->clearQueue();  
  INFO( Log::DELIBERATOR, 5,  "Generating plan for " +
      currentIntention->toString() + "\n" );
  Plan plan = currentIntention->generatePlan();
  INFO( Log::DELIBERATOR, 6,  "Executing plan " +
      plan.toString() + "\n" );
  theEffector->queuePlan( plan );
  INFO( Log::DELIBERATOR, 2, "LEAVE Deliberator::startDeliberation\n" );
}

Deliberator::~Deliberator()
{
  INFO( Log::DELIBERATOR, 3, "ENTER Deliberator::~Deliberator\n" );
  delete currentIntention;
  delete previousIntention;
  
  for ( int i = 0; i < options.count(); i++ ) {
    delete options[i];
  }
  for ( int i = 0; i < potentialConstraints.count(); i++ ) {
    delete potentialConstraints[i];
  }
}
  
  
  
int Deliberator::indexOfBestDesire( 
  const Array<Option *>& desires ) const

/*
4 Faelle: 
- die bisherige Absicht ist nach wie vor die beste. 
   -> Die Handlungsoption wird erneut ausgewaehlt.
- die bisherige Absicht ist wieder wuenschenswert, und der Nutzen ist 
  hoechstens cancellationCost schlechter als der des nuetzlichsten 
  Wunsches.
  -> Die Handlungsoption wird erneut ausgewaehlt.
- die bisherige Absicht ist wieder wuenschenswert, aber der Nutzen ist mehr als
  cancellationCost schlechter als der des nuetzlichsten Wunsches.
  -> der nuetzlichste Wunsch wird ausgewaehlt
- die bisherige Absicht ist ueberhaupt nicht unter den Wuenschen.
  -> der nuetzlichste Wunsch wird ausgewaehlt
*/  
{
  INFO( Log::DELIBERATOR, 4, "ENTER Deliberator::indexOfBestDesire\n" );
  int i;
  int bestIndex;
  int switchingUtility;
  int indexOfPreviousIntention = UNDEF_INT;
  int bestUtility = INT_MIN;
  int currentUtilityOfPreviousIntention = UNDEF_INT;
  
  if ( desires.count() == 0 ) {
    // mind. 1 Wunsch ist notwendig!!
    throwException( NoDesiresLeftException() );
  }
    
  for ( i = 0; i < desires.count(); i++ ) {
    if  ( desires[i]->expectedUtility() > bestUtility ) {
      bestIndex = i;
      bestUtility = desires[i]->expectedUtility();
    }
    if ( desires[i]->id() == previousIntention->id() ) {
      currentUtilityOfPreviousIntention = desires[i]->expectedUtility();
      indexOfPreviousIntention = i;
    } 
  }
  
  // Berechnung des Nutzens des Wechsels. Er kann auch negativ sein.
  switchingUtility = bestUtility - currentUtilityOfPreviousIntention;
  switchingUtility -= previousIntention->cancellationCost();

  if ( indexOfPreviousIntention == UNDEF_INT ||
       switchingUtility > NEUTRAL_UTILITY ) {
    INFO( Log::DELIBERATOR, 7, "index: " + String( bestIndex ) + "\n" );
    return bestIndex;
  } 
  else {
    INFO( Log::DELIBERATOR, 7, "index: " + 
        String( indexOfPreviousIntention ) + "\n" );
    return indexOfPreviousIntention;
  }
}
  
  
    
