#ifndef _DELIBERATOR_H_
#define _DELIBERATOR_H_

#include "Array.h"
#include "Options.h"
#include "Constraints.h"
#include "Plan.h"
#include "Effectors.h"

/**
  Die Entscheidungskomponente des Agenten.
  
  Der Deliberator waehlt aus einer Reihe von Handlungsoptionen (Objekte der
  Klasse \Ref{Option} und ihrer Subklassen)
  w\"unschenswerte aus, und setzt dann den nuetzlichsten, unter den momentan
  sinnvollen Nebenbedingungen realisierbaren Wunsch um. Die Liste der
  beruecksichtigten Optionen und Nebenbedingungen ist zur Laufzeit aenderbar.
  
  Die zentrale Methode ist \Ref{startDeliberation}. Sie steuert den Ablauf des
  Entscheidungsprozesses. Dieser Entscheidungsproze\3 wird jedesmal gestartet,
  wenn neue Sichtinformationen eintreffen, oder die Action-Queue von
  \Ref{Effectors} leer ist. Das hei\3t, da\3 jede Handlungsoption jederzeit,
  also auch w\"ahrend ihrer Umsetzung, in der Lage sein mu\3, den erwarteten
  Nutzen ihrer Umsetzung bis zum Ende zu berechnen.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur
  Vererbung geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Deliberator.h} | [\URL[Implementierung]{src/Deliberator.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Deliberator.h,v $
  *   $Revision: 1.18 $
  *   $Date: 1998/08/14 13:20:14 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Deliberator
{
public:
  /** Konstruktor uebernimmt den Effektor.
    Es werden alle moeglichen Wuensche und alle moeglichen (sinnvollen)
    Nebenbedingungen verwendet. */
  Deliberator( Effectors *anEffector );


  /** Konstruktor uebernimmt Handlungsoptionen, potentielle
    Nebenbedingungen und den Effektor.
    @param optionIds
    @param potentialConstraintIds
    @param anEffector Zeiger auf den Effektor */
  Deliberator( 
    const Array<int>& optionIds, 
    const Array<int>& potentialConstraintIds,
    Effectors *anEffector );

  /** Hinzufuegen einer Handlungsoption. 
    Ist die Option bereits in der Liste der Optionen enthalten, geschieht
    nichts.
    @param optionId Die Id der hinzuzufuegenden Option. */
  void addOption(int optionId );
  
  /** Hinzufuegen einer potentiellen Nebenbedingung. 
    Ist sie bereits in der Liste der potentiellen Nebenbedingungen enthalten, 
    geschieht nichts.
    @param constraintId Die Id der hinzuzufuegenden Nebenbedingung. */
  void addPotentialConstraint(int constraintId );
  
  /** F\"uhrt den Entscheidungsproze\3 durch.
  Hier wird eine recht flexible, sog. BDI-Architektur eingesetzt (B - Belief, D
  - Desire, I - Intention). Im Einzelnen passiert bei der Entscheidungsfindung
  folgendes:

  Als erstes werden die geforderten Nebenbedingungen bestimmt. Dazu wird jedes
  Element von \Ref{potentialConstraints} nach seiner 
  \Ref{Constraint::expectedUtility} gefragt. Ist sie groesser als
  \Ref{minConstraintUtility}(), wird sie in die Liste der geforderten
  Nebenbedingungen aufgenommen.
  
  Als n\"achstes werden die W\"unsche bestimmt. Dazu wird von jedem Wunsch in
  \Ref{options} eine Kopie angelegt, der die geforderten Nebenbedingungen
  mitgeteilt werden (per Aufruf von \Ref{Option::setConstraints}() ). Liegt der
  erwartete Nutzen \"uber \Ref{minDesireUtility}(), wird die Handlungsoption in
  die Liste der W\"unsche aufgenommen. Interessant ist hier vor allem die
  Bestimmung der N\"utzlichkeit eines Wunsches durch
  Option::\Ref{Option::computeExpectedUtility}().
  
  Nun wird aus den W\"unschen eine Absicht bestimmt. Dabei spielen die letzte
  Absicht, zu deren Umsetzung sich der Agent entschlossen hat (committed
  intention), die Umsetzbarkeit eines Wunsches, und nicht zuletzt der Nutzen des
  jeweiligen Wunsches eine Rolle. Dieser Schritt ist in die Methode
  \Ref{indexOfBestDesire} ausgelagert. Konnte kein umsetzbarer Wunsch ermittelt
  werden, wird der Standardwunsch \Ref{Option} zur Absicht bestimmt.
  
  Bevor ein konkreter Plan entwickelt wird, hat die Absicht die Gelegenheit,
  Daten mit der bisherigen Absicht abzugleichen (mit dem Aufruf von
  \Ref{Option::adaptToPreviousIntention}() ). 
  
  Als letztes wird die nun festgelegt Absicht aufgefordert, einen Plan zu ihrer
  Umsetzung zu erzeugen (mit \Ref{Option::generatePlan}). Dieser Plan wird nun
  den Effectors zur Ausf\"uhrung \"ubergeben. */
  void startDeliberation();
  
  /** Destruktor. */
  ~Deliberator();


private:
  /** Private Fehlerklasse, die in indexOfBestDesire ausgeloest wird. */
  class NoDesiresLeftException : public Exception {
  public:
    NoDesiresLeftException() : Exception() {}
  };  

  /** Der Effektor */
  Effectors *theEffector;

  /** Die Handlungsoptionen */
  Array <Option *> options;

  /** Die potentiellen Nebenbedingungen */
  Array <Constraint *> potentialConstraints;


  /** Die momentane Absicht */
  Option *currentIntention;

  /** Die vorherige Intention */
  Option *previousIntention;

  // Methoden
  
  /** Initialisiert die internen moeglichen Wuensche und Nebenbedingungen. */
  void initialize( 
    const Array<int>& optionIds, 
    const Array<int>& potentialConstraintIds );

  
  /** Waehlt den Wunsch mit dem groessten Nutzen aus. 
    Dabei werden die Abbruchkosten der Umsetzung der bisherigen Absicht
    beruecksichtigt ( \Ref{Option::cancellationCost}() ). Ist das Array mit den
    Wuenschen leer, wird eine Exception erzeugt, da dann kein sinnvoller Index
    zurueckgegeben werden kann.
    @param desires Array mit den Wuenschen, aus denen der beste
      ausgewaehlt wird.
    @exception Deliberator::NoDesiresLeftException, wenn das \"ubergeben Array
      desires ler ist.
    @return Wert zwischen 0 und desires.count()-1 */
  int indexOfBestDesire( 
    const Array<Option *>& desires ) const;
  
  /** Liefert den minimalen Nutzen, den das Einhalten einer Nebenbedingung
    bringen muss, damit die Bedingung in die Liste der einzuhaltenden 
    Bedingungen aufgenommen wird.
    @return Ein Wert zwischen NEUTRAL_UTILITY und MAX_UTILITY. */
  int minConstraintUtility() const;
  
  /** Liefert den minimalen Nutzen, den die Umsetzung einer Handlungsoption 
    bringen muss, damit sie in die Liste der Wuensche aufgenommen wird.
    @return Ein Wert zwischen NEUTRAL_UTILITY und MAX_UTILITY. */
  int minDesireUtility() const;

};

#endif

