#ifndef _MODELLEDBALL_H_
#define _MODELLEDBALL_H_

#include "Configuration.h"

#include "Common.h"
#include "ModelledObject.h"
#include "Reliability.h"
#include "MovingObject.h"
#include "Vector.h"

class ModelledPlayer;

/**
  Die interne Repraesentation des Balls im Weltmodell.
  Ein ModelledBall hat zusaetzlich zu den Eigenschaften eines ModelledObject
  das Wissen darueber, von welchem Team er als letzter beruehrt worden ist.
  (Bisher wird diese Eigenschaft aber noch nirgendwo gesetzt und nirgendwo
  abgefragt.)

  {\small {\bf Quelltext}

  [\URL[Header]{src/ModelledBall.h} | [\URL[Implementierung]{src/ModelledBall.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ModelledBall.h,v $
  *   $Revision: 1.40 $
  *   $Date: 1998/08/14 13:20:51 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class ModelledBall : public ModelledObject
{
public:

  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. */
    class ModelledBallException: public ModelledObjectException {
    public:
      ModelledBallException(const String& aDescription )
        : ModelledObjectException( aDescription ) {}
    };
    
    /** Fehlerklasse f\"ur illegalen Kickversuch.
      Sie wird ausgel\"ost, wenn der Ball au\3er Reichweite des Spielers 
      ist. */
    class KickImpossibleException: public ModelledBallException {
    public:
      KickImpossibleException(const String& aDescription )
        : ModelledBallException( aDescription ) {}
    };
  //@}

  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Default-Konstruktor und neuer Ball mit Daten.
      @param id Die Id des Balles.
      @param aPos Die Position. Als Default wird unbekannt angenommen.
      @param aSpeed Die Geschwindigkeit. Als Default wird unbekannt angenommen.
      @param aPosReliability Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aPosReliability Die Verlaesslichkeit der Geschwindikeit. Als Default
      wird RANDOM angenommen.
      @param aLastTouchedBy Das Team, von dem der Ball als letztes beruehrt
      worden ist. Default ist UNDEF_TEAM. 
      @param aLastTouchedAtTime Zeit, zu der der Ball zuletzt ber\"uhrt wurde.
      Default ist -1. */
    ModelledBall( int id = UNDEF_ID,
                  const Vector& aPos = Vector(), 
		  const Vector& aSpeed = Vector(),
		  const Reliability& aPosReliability =
                  Reliability(Reliability::RANDOM),
		  const Reliability& aSpeedReliability =
                  Reliability(Reliability::RANDOM),
		  int aLastTouchedBy = UNDEF_TEAM_ID,
		  int aLastTouchedAtTime = -1);

    /** Modellierung eines wahrgenommenen Balls. 
      Dabei werden die Positions- und Geschwindigkeitsvektoren NICHT in das
      absolute Koordinatensystem \"ubertragen, sondern einfach \"ubernommen.
      Die Transformation ist Aufgabe des Weltmodells.
      @param aMovingObject Der gesehene Ball.
      @param aPosReliability Die Verlaesslichkeit der Position. Als Default
      wird RANDOM angenommen.
      @param aPosReliability Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aLastTouchedBy Das Team, von dem der Ball als letztes beruehrt
      worden ist. Default ist UNDEF_TEAM_ID.
      @param aLastTouchedAtTime Zeit, zu der der Ball zuletzt ber\"uhrt wurde.
      Default ist -1. */
    ModelledBall( const MovingObject& aMovingObject,  
		  const Reliability& aPosReliability =
		  Reliability(Reliability::RANDOM),
		  const Reliability& aSpeedReliability =
		  Reliability(Reliability::RANDOM),
		  int aLastTouchedBy = UNDEF_TEAM_ID,
		  int aLastTouchedAtTime = -1);
      
    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    virtual StorableObject* copy() const;
  //@}
  

  /** @name Werte abfragen */
  //@{  
    /** Liefert Team, das zuletzt den Ball beruehrt hat. */
    int lastTouchedBy() const;

    /** Liefert die Zeit, zu der der Ball zuletzt ber\"uhrt wurde. */
    int lastTouchedAtTime() const;

    /** Beschreibung als String fuer die Ausgabe auf einem Terminal.
      Es werden ID, Position, Geschwindigkeit, Zuverlaessigkeitswerte,
      LastTouchedBy und LastTouchedAtTime ausgegeben. */
    virtual String toString() const;
  //@}  


  /** @name Werte setzen */
  //@{         
    /** Setzt Team, das zuletzt den Ball beruehrt hat, neu. */
    void setLastTouchedBy( int  newLastTouchedBy );

    /** Setzt die Zeit, zu der der Ball zuletzt ber\"uhrt wurde. */
    void setLastTouchedAtTime( int  newLastTouchedAtTime );
  //@}
    
  
  /** @name Simulation */
  //@{
    /** Fuehrt Simulationsschritte aus. 
      Dabei wird die Geschwindigkeit auf die Position addiert und die
      Geschwindigkeit mit dem Verlangsamungsfaktor multipliziert. Die
      Verlaesslichkeit wird herabgesetzt. */
    void simulateSteps( unsigned int simSteps = 1 );

    /** Beschleunigt. 
      Wie stark sich die Kraft auf die Beschleunigung auswirkt, haengt vom
      Kraftfaktor (powerRate), von einem Objektbeschleunigungsfaktor
      (maxPowerRate) und von der Position des kickenden Spielers (byPlayer) ab.
      
      Durch Aufruf von accelerate werden nur Werte gesetzt. Um die Aktion
      tats\"achlich zu simulieren, mu\3 anschlie\3end \Ref{simulateSteps}
      aufgerufen werden.
      @param power Die Kraft, mit der beschleunigt wird. 
      @param angle Die Richtung, relativ zur Blickrichtung des Spielers
      byPlayer, in die beschleunigt werden soll.
      @param byPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @exception ModelledBallException bei undefierter Ballgeschwindigkeit */
    void accelerate( double power, double angle,
		    const ModelledPlayer& byPlayer );
  //@}
      
  /** @name Hilfsfunktionen zur Kick-Berechnung */
  //@{
    /** Ueberprueft, ob der Ball im Kick-Bereich ist.
      @return YES, wenn ja, sonst nein */
      bool isInKickRangeOfPlayer( const ModelledPlayer& aPlayer ) const;
      
    /** Es wird die tatsaechlich verfuegbare Kraft
      (effectiveKickPower) ausgerechnet.
      @param power Die Kraft, mit der man den Ball kicken will.
      @param kickingPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @return effectiveKickPower Die Kraft, mit der der Ball tatsaechlich
      beschleunigt wird.
      @exception KickImpossibleException falls Ball nicht im Kick-Bereich (zu 
      weit weg) */
    double effectiveKickPower( double power,
			   const ModelledPlayer& kickingPlayer ) const;

    /** Es wird die Kraft bestimmt, die dem SoccerServer geschickt werden
      muss, damit effektiv der Ball mit der Kraft power beschleunigt wird.
      Diese Funktion ist also genau die Umkehrfunktion von effectiveKickPower.
      Es gilt also: kickPowerForSend(effectiveKickPower( power )) == power,
      wenn power normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert
      und eine Warnung ins Log geschrieben.
      @param power Die Kraft mit der Ball beschleunigt werden soll.
      @param kickingPlayer Der Spieler der die Beschleunigung ausfuehrt.
      @return powerForSend Die Kraft mit dem man den Ball kicken muss.
      @exception KickImpossibleException falls  Ball nicht im Kick-Bereich (zu 
      weit weg) */
    double kickPowerForSend( double power,
			     const ModelledPlayer& kickingPlayer ) const;

    /** Liefert den maximalen sinnvollen effektiven Kick durch den der 
      angegebene Spieler den Ball zur derzeitigen Situation beschleunigen kann. */
    double maxEffectiveKickPower(const ModelledPlayer& kickingPlayer) const;

    /** Normalisiert die Kraft fuer einen kick.
      @exception ModelledBallException falls der uebergebene Wert UNDEF_DOUBLE 
      ist */
    double normalizeKickPower(double power) const;

    /** Normalisiert den Winkel fuer einen kick. */
    double normalizeKickAngle(double angle) const;
  //@}

    
  /** @name Konfigurierbare Werte erfragen */
  //@{
    /** Klassenvariablen BallSize abfragen */
    double size() const;

    /** Klassenvariablen BallDecay abfragen */
    double decay() const;

    /** Klassenvariablen BallMaxSpeed abfragen */
    double maxSpeed() const;

    /** Klassenvariablen KickPowerRate abfragen */
    double kickPowerRate() const;

    /** Klassenvariablen MaxPower abfragen */
    double maxPower() const;

    /** Klassenvariablen MinPower abfragen */
    double minPower() const;

    /** Klassenvariablen MaxMoment abfragen */
    double maxMoment() const;

    /** Klassenvariablen MinMoment abfragen */
    double minMoment() const;

    /** Klassenvariablen KickMargin abfragen */
    double kickMargin() const;

    /** Klassenvariablen BallSize setzen */
    void setSize(double aBallSize);

    /** Klassenvariablen BallDecay setzen */
    void setDecay(double aBallDecay);

    /** Klassenvariablen BallMaxSpeed setzen */
    void setMaxSpeed(double aBallMaxSpeed);

    /** Klassenvariablen KickPowerRate setzen */
    void setKickPowerRate(double aKickPowerRate);

    /** Klassenvariablen MaxPower setzen */
    void setMaxPower(double aMaxPower);

    /** Klassenvariablen MinPower setzen */
    void setMinPower(double aMinPower);

    /** Klassenvariablen MaxMoment setzen */
    void setMaxMoment(double aMaxMoment);

    /** Klassenvariablen MinMoment setzen */
    void setMinMoment(double aMinMoment);

    /** Klassenvariablen KickMargin setzen */
    void setKickMargin(double aKickMargin);
  //@}

  
  /** @name Konfigurierbare Werte setzen */
  //@{ 
    /** liest die Werte aller Klassenvariablen aus dem Konfigurationsfile 
    ein. */
    void setClassVars();
  //@}


private:
  ///
  static double theBallSize;
  ///
  static double theBallDecay;
  ///
  static double theBallMaxSpeed;
  ///
  static double theKickPowerRate;
  ///
  static double theMaxPower;
  ///
  static double theMinPower;
  ///
  static double theMaxMoment;
  ///
  static double theMinMoment;
  ///
  static double theKickMargin;
  ///
  static short  isSetOnce;

  /** Formel mit der die Kickstaerke modifiziert wird */
  double kickPowerMult( double angleFromPlayerToBall,
			double distFromPlayerToBall ) const;

  /** Id des Teams, das den Ball als letztes beruehrt hat. */
  int theLastTouchedBy;

  /** Zeit zu der der Ball zuletzt ber\"uhrt wurde. */
  int theLastTouchedAtTime;
};

inline double ModelledBall::normalizeKickPower( double power ) const {
  if( power == UNDEF_DOUBLE ) {
    throwException( ModelledBallException( "undef power value" ) );
  }
  else if( power < minPower() ) {
    return minPower();
  }
  else if( power > maxPower() ){
    return maxPower();
  }
  else {
    return power;
  }
}

inline double ModelledBall::normalizeKickAngle( double angle ) const {
  angle = normalizeAngle(angle);
  
  if( angle < minMoment() ) {
    return minMoment();
  }
  else if( angle > maxMoment() ) {
    return maxMoment();
  }
  else {
    return angle;
  }
}

inline int ModelledBall::lastTouchedBy() const {
  return theLastTouchedBy;
}

inline void ModelledBall::setLastTouchedBy( int newLastTouchedBy ) {
  theLastTouchedBy = newLastTouchedBy;
}

inline int ModelledBall::lastTouchedAtTime() const {
  return theLastTouchedAtTime;
}

inline void ModelledBall::setLastTouchedAtTime( int newLastTouchedAtTime ) {
  theLastTouchedAtTime = newLastTouchedAtTime;
}

inline double ModelledBall::size() const {
  return ModelledBall::theBallSize;
}

inline double ModelledBall::decay() const {
  return ModelledBall::theBallDecay;
}

inline double ModelledBall::maxSpeed() const {
  return ModelledBall::theBallMaxSpeed;
}

inline double ModelledBall::kickPowerRate() const {
  return ModelledBall::theKickPowerRate;
}

inline double ModelledBall::maxPower() const {
  return ModelledBall::theMaxPower;
}

inline double ModelledBall::minPower() const {
  return ModelledBall::theMinPower;
}

inline double ModelledBall::maxMoment() const {
  return ModelledBall::theMaxMoment;
}

inline double ModelledBall::minMoment() const {
  return ModelledBall::theMinMoment;
}

inline double ModelledBall::kickMargin() const {
  return ModelledBall::theKickMargin;
}

inline void ModelledBall::setSize(double aBallSize) {
  ModelledBall::theBallSize = aBallSize;
}

inline void ModelledBall::setDecay(double aBallDecay) {
  ModelledBall::theBallDecay=aBallDecay;
}

inline void ModelledBall::setMaxSpeed(double aBallMaxSpeed) {
  ModelledBall::theBallMaxSpeed=aBallMaxSpeed;
}

inline void ModelledBall::setKickPowerRate(double aKickPowerRate) {
  ModelledBall::theKickPowerRate=aKickPowerRate;
}

inline void ModelledBall::setMaxPower(double aMaxPower) {
  ModelledBall::theMaxPower=aMaxPower;
}

inline void ModelledBall::setMinPower(double aMinPower) {
  ModelledBall::theMinPower=aMinPower;
}

inline void ModelledBall::setMaxMoment(double aMaxMoment) {
  ModelledBall::theMaxMoment=aMaxMoment;
}

inline void ModelledBall::setMinMoment(double aMinMoment) {
  ModelledBall::theMinMoment=aMinMoment;
}

inline void ModelledBall::setKickMargin(double aKickMargin) {
  ModelledBall::theKickMargin = aKickMargin;
}

#endif
