#ifndef _MODELLEDPLAYER_H_
#define _MODELLEDPLAYER_H_

#include "Configuration.h"

#include "Common.h"
#include "ModelledObject.h"
#include "Reliability.h"
#include "MovingPlayer.h"
#include "Exception.h"
#include "ViewPoint.h"

class ModelledBall;

/**
  Die interne Repraesentation eines Spielers im Weltmodell.
  Ein ModelledPlayer hat zusaetzlich zu den Eigenschaften eines ModelledObject
  noch eine Blickrichtung und eine Kraftreserve. Die Kraftreserve setzt sich
  aus
  den Werten stamina, effort und recovery zusammen. Dazu kommen entsprechende
  Verl\"asslichkeitsinformationen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/ModelledPlayer.h} | [\URL[Implementierung]{src/ModelledPlayer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: ModelledPlayer.h,v $
  *   $Revision: 1.43 $
  *   $Date: 1998/08/14 13:20:55 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class ModelledPlayer : public ModelledObject
{

public:

  /** @name Interne Fehlerklassen. */
  //@{
    /** Allgemeine Fehlerklasse. */
    class ModelledPlayerException: public ModelledObjectException {
    public:
      ModelledPlayerException(const String& aDescription )
        : ModelledObjectException( aDescription ) {}
    };
  //@}
 
       
  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Neuer Spieler mit Daten und Default-Konstruktor.
      @param id Die Id des Spielers
      @param aPos Die Position. Als Default wird unbekannt angenommen.
      @param aSpeed Die Geschwindigkeit. Als Default wird unbekannt angenommen.
      @param aFaceDir Die Blickrichtung. Als Default wird unbekannt angenommen.
      @param aPosReliability Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aSpeedReliability Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aFaceDirReliability Die Verl\"a\3lichkeit der Blickrichtung. Als
      Delault wird RANDOM angenommen.
      @param aViewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen. Als Default wird 
      VIEW_WIDTH_WIDE angenommen.
      @param stamina die Kraftreserve des Spielers. Gueltige Werte: 0 bis
      theStaminaMax. Kleinere Werte werden auf 0 gesetzt, groessere Werte
      (UNDEF_DOUBLE) auf theStaminaMax.
      @param effort der Leistungsfaktor des Spielers. Gueltige Werte:
      theEffortMin bis 1.0. Kleinere Werte werden auf theEffortMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param recovery der Erholungsfactor des Spielers. Gueltige Werte
      theRecoveryMin bis 1.0. Kleinere Werte werden auf theRecoveryMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param aStaminaReliability Die Verl\"asslichkeit der Stamina-Werte. Als
      Default wird RANDOM angenommen. */
    ModelledPlayer( int id = UNDEF_ID,
      const Vector& aPos = Vector(), 
      const Vector& aSpeed = Vector(),
      double aFaceDir = UNDEF_DOUBLE,
      const Reliability& aPosReliability = Reliability(Reliability::RANDOM),
      const Reliability& aSpeedReliability = Reliability(Reliability::RANDOM),
      const Reliability& aFaceDirReliability = Reliability(Reliability::RANDOM),
      int aViewWidth = VIEW_WIDTH_WIDE,
      double stamina = UNDEF_DOUBLE,
      double effort = UNDEF_DOUBLE,
      double recovery = UNDEF_DOUBLE,
		  const Reliability& aStaminaReliability = Reliability(Reliability::RANDOM) 
    );

    /** Modellierung eines wahrgenommenen Spielers.
      @param aMovingPlayer Der gesehene Spieler.
      @param aPosReliability Die Verlaesslichkeit der Position. Als Default wird 
      RANDOM angenommen.
      @param aSpeedReliability Die Verlaesslichkeit der Geschwindikeit. Als
      Default wird RANDOM angenommen.
      @param aFaceDirReliability Die Verl\"a\3lichkeit der Blickrichtung. Als
      Deafault wird RANDOM angenommen.
      @param aViewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen. Als Default wird 
      VIEW_WIDTH_WIDE angenommen.
      @param stamina die Kraftreserve des Spielers. Gueltige Werte: 0 bis
      theStaminaMax. Kleinere Werte werden auf 0 gesetzt, groessere Werte
      (UNDEF_DOUBLE) auf theStaminaMax.
      @param effort der Leistungsfaktor des Spielers. Gueltige Werte:
      theEffortMin bis 1.0. Kleinere Werte werden auf theStaminaMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param recovery der Erholungsfactor des Spielers. Gueltige Werte
      theRecoveryMin bis 1.0. Kleinere Werte werden auf theRecoveryMin gesetzt,
      groessere Werte (UNDEF_DOUBLE) auf 1.0.
      @param aStaminaReliability Die Verl\"asslichkeit der Stamina-Werte. Als
      Default wird RANDOM angenommen. */
    ModelledPlayer( const MovingPlayer& aMovingPlayer,  
      const Reliability& aPosReliability = Reliability(Reliability::RANDOM),
      const Reliability& aSpeedReliability = Reliability(Reliability::RANDOM),
      const Reliability& aFaceDirReliability = Reliability(Reliability::RANDOM),
      int aViewWidth = VIEW_WIDTH_WIDE,
      double stamina = UNDEF_DOUBLE,
      double effort = UNDEF_DOUBLE,
      double recovery = UNDEF_DOUBLE,
      const Reliability& aStaminaReliability = Reliability(Reliability::RANDOM)
    );
  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
    darauf zurueck. */
  virtual StorableObject* copy() const;
  //@}

    
  /** @name Werte abfragen */
  //@{  
    /** Liefert den Stamina-Wert. */
    double stamina() const;
    
    /** Liefert den Effort-Wert. */
    double effort() const;
    
    /** Liefert den Recovery-Wert. */
    double recovery() const;
    
    /** Liefert Verl\"a\3lichkeit der Stamina-Werte. */
    const Reliability& staminaReliability() const;

    /** Liefert Blickrichtung */
    double faceDir() const;

    /** Liefert die Verl\"a\3lichkeit der Blickrichtung. */
    const Reliability& faceDirReliability() const;

    /** Liefert die Groesse des Gesichtsfeldes. 
      @return VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE */
    int viewWidth() const;

    /** Liefert den Wert des Winkels des sichtbaren Bereiches. */
    double visibleAngle() const;
    
    /** Liefert die TeamId des Teams, zu dem der Spieler geh\"ort.
      @return OUR_TEAM_ID, THEIR_TEAM_ID oder UNDEF_TEAM_ID
    */
    int team() const;

    /** Liefert YES wenn die Nummer des Spielers bekannt ist, sonst NO. */
    bool isPlayerNumberKnown() const;

    /** Beschreibung als String fuer die Ausgabe auf einem Terminal.
      Es werden ID, Position, Geschwindigkeit, Zuverlaessigkeitswerte, FaceDir,
      Sichtrichtung und -winkel (in Grad), sowie die Staminawerte 
      ausgegeben. */
    virtual String toString() const;
  //@}  

  
  /** @name Werte setzen */
  //@{         
    /** Setzt Stamina-Modell. 
      @param newStamina Die Kraftreserve. Gueltige Werte: 0 bis theStaminaMax
      @param newEffortFactor Der Leistungsfaktor.
      Gueltige Werte: theEffortMin bis 1.0
      @param newRecoveryFactor Der Erholungsfaktor.
      Gueltige Werte: theRecoveryMin bis 1.0
      @param aStaminaReliability Die Verl\"asslichkeit der Stamina-Werte. Als
      Default wird PERFECT angenommen. */
    void setStaminaValues( double newStamina = UNDEF_DOUBLE,
      double newEffortFactor = UNDEF_DOUBLE,
      double newRecoveryFactor = UNDEF_DOUBLE,
      const Reliability& aStaminaReliability =
      Reliability(Reliability::PERFECT) );

    /** Setzt den Stamina-Wert. */
    void setStamina( double newStamina );

    /** Setzt den Effort-Wert. */
    void setEffort( double newEffortFactor );

    /** Setzt die Recovery. */
    void setRecovery( double newRecoveryFactor );

    /** Setzt die Verl\"a\3lichkeit der Stamina-Werte. */
    void setStaminaReliability( const Reliability& newStaminaReliability );

    /** Setzt Blickrichtung neu. */
    void setFaceDir( double newFaceDir );
    /** Setzt Verl\"a\3lichkeit der Blickrichtung. */
    void setFaceDirReliability( const Reliability& newFaceDirReliability );

    /** Setzt die Groesse des Gesichtsfeldes. 
      Es wird nicht auf Gueltigkeit getestet.
      @param newViewWidth Neue Gesichtsfeldgroesse. Moegliche Werte sind 
      VIEW_WIDTH_NARROW, VIEW_WIDTH_NORMAL, VIEW_WIDTH_WIDE. */
    void setViewWidth( int newViewWidth );

    /** Setzt den Sichtwinkel. Dieser ist abhaengig vom Paramter viewWidth und
      von den normalen Sichtwinkel der aus der Datei server.conf entnommen wird.
      @param viewWidth Sichtbreite darf nur die Werte VIEW_WIDTH_NARROW,
      VIEW_WIDTH_NORMAL oder VIEW_WIDTH_WIDE annehmen.
      @exception ModelledPlayerException bei undefinierter viewWidth */
    void setVisibleAngle( int viewWidth );  
  //@}

        
  /** @name Simulation */
  //@{
    /** Simuliert das Verstreichen von Simulationszeit. Die Routine geht davon
      aus, dass der Spieler waehrend dieser Zeit keine Aktionen unternimmt. Es
      wird simuliert: Die Geschwindigkeitsaenderung, die Positionsaenderung und
      die Staminaaenderung. Die Verlaesslichkeit wird herabgesetzt.
      @param stepCount Zahl der Simulationsschritte. */
    void simulateSteps( unsigned int simSteps = 1 );

    /** Baut die Kraftreserven wieder auf. 
      @param simSteps Die Zahl der Simulationsschritte, die zur Erholung zur
      Verfuegung stehen. Wird von simulateSteps() aufgerufen. */
    void recover( int simSteps = 1 );

    /** Beschleunigt. Wie stark sich die Kraft auf die Beschleunigung auswirkt,
      haengt vom Kraftfaktor (powerRate) und einem Objektbeschleunigungsfaktor
      (maxPowerRate) ab. Dabei wird auch die Kraftreserve herabgesetzt.
      @param aPower Die Kraft, mit der beschleunigt wird. 
      @exception ModelledPlayerException bei undefinierter Geschwindigkeit */
    void accelerate( double aPower );

    /** Dreht die Blickrichtung. Die naechste Beschleunigung geht standardmaessig
      in diese Richtung.
      @param angle Ein Winkel relativ zur bisherigen Blickrichtung.  */
    void turn( double angle );
  //@}
    
    
  /** @name Winkelbezogene Hilfsfunktionen */
  //@{
    /** Ermittelt den Winkel unter dem der Spieler den angegebenen Punkt sieht.
      @exception ModelledPlayerException bei undefinierter Blickrichtung */
    double angleTo( const Vector& point ) const;

    /** Ermittelt ob sich das angegeben Objekt im Sichtbereich des
      Spielers befindet. Wenn sich das Objekt im Sichtbereich befindet
      wird YES zurueckgegeben, andernfalls NO. */
    bool isInFieldOfView( const StableObject& pos ) const;  

    /** Ermittelt ob sich das angegeben Objekt in unmittelbarer N\"ahe des
      Spielers befindet. Wenn sich das Objekt in unmittelbarer N\"ahe des
      Spielers befindet (der Spieler das Objekt f\"uhlen kann)
      wird YES zurueckgegeben, andernfalls NO. */
    bool isInFieldOfFeel( const StableObject& pos ) const;  

    /** Ermittelt ob sich das angegeben Objekt im Wahrnehmungsbereich des
      Spielers befindet. Wenn sich das Objekt im Wahrnehmungsbereich befindet
      wird YES zurueckgegeben, andernfalls NO. Ein Objekt befindet sich im
      Wahrnehmungsbereich des Spielers, wenn es entweder in dessen
      Sichtbereich ist, oder in unmittelbarer N\"ahe des Spielers ist
      @see isInFieldOfView
      @see isInFieldOfFeel */
    bool isInFieldOfSense( const StableObject& pos ) const;  
  //@}  

        
  /** @name Dash-Berechnung */
  //@{
    /** Es wird die tatsaechlich verfuegbare Kraft fuer einen dash
      (effectiveDashPower) ausgerechnet. Diese ist abhaengig vom Effort und vom
      noch zur Verfuegung stehenden Stamina.
      @param power Die Kraft, mit der man beschleunigen will.
      @return effectiveDashPower Die Kraft, mit der man tatsaechlich 
      beschleunigt.  */
    double effectiveDashPower( double power ) const;

    /** Es wird die Kraft bestimmt, die dem SoccerServer geschickt werden muss,
      damit man effektiv mit der Kraft power beschleunigt. 
      Diese Funktion ist also genau die Umkehrfunktion von effectiveDashPower.
      Es gilt also: dashPowerForSend(effectiveDashPower( power )) == power,
      wenn power normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert und
      eine Warnung ins Log geschrieben.
      @param power Die Kraft mit der der Spieler beschleunigen soll.
      @return powerForSend Die Kraft die man den SoccerServer senden muss. */
    double dashPowerForSend( double power ) const;

    /** Liefert den maximalen sinnvollen effektiven Dash mit dem der Spieler
      zur derzeitigen Situation beschleunigen kann.
      @exception ModelledPlayerException bei undefiniertet Geschwindigkeit */
    double maxEffectiveDashPower() const; 
    
    /** Normalisiert die Kraft fuer einen dash. 
      @exception ModelledPlayerException falls der Power-Wert undefiniert ist */
    double normalizeDashPower(double power) const;
  //@}
    
    
  /** @name Turn-Berechnung */
  //@{
    /** Es wird die tatsaechlich ausgefuehrte Drehung
      (effectiveTurnAngle) ausgerechnet. Diese ist abhaengig von der aktuellen
      Geschwindigkeit des Spielers.
      @param angle Der Winkel, um die man sich drehen will.
      @return effectiveTurnAngle Der Winkel, um den man sich tatsaechlich
      dreht.  */
    double effectiveTurnAngle( double angle ) const;

    /** Es wird der Winkel bestimmt, der dem SoccerServer geschickt werden muss,
      damit sich der Spieler effektiv um den angegebenen Winkel dreht.
      Diese Funktion ist als genau die Umkehrfunktion von effectiveTurnAngle.
      Es gilt also: turnAngleForSend(effectiveTurnAngle( angle )) == angle,
      wenn angle normalisiert ist (im definierten Bereich liegt). Liegt das
      zu sendene Kommando nicht im definierten Bereich, wird normalisiert
      und eine Warnung ins Log geschrieben.
      @param angle Der Winkel um den sich der Spieler drehen soll.
      @return angleForSend Den Winkel, den man dem Server uebermitteln muss. */
    double turnAngleForSend( double angle ) const;

    /** Liefert den vom Betrag her maximalen sinnvollen Turn um dessen Winkel
      sich der Spieler in der gegebenen Situation drehen kann. */
    double maxEffectiveTurnAngle() const;
    
    /** Normalisiert die Drehung fuer einen turn. */
    double normalizeTurnAngle(double angle) const;
  //@}

        
  /** @name Entfernunsberechnungen */
  //@{
    /** Gibt den Abstand des Spielers von der angegebenen Position zurueck.
      Dieser Abstand wird vom Mittelpunkt des Spielers gemessen.
      @param position Die Position zu der Abstand ermittelt werden soll.
      @return distance Abstand des Spielers zur Position position. */
    double distTo( const Vector& position ) const;

    /** Schaetzt die Schritte, die der Spieler braucht, um den angegebenen
      Ball zu erreichen. 
      @param ball Der Ball der erreicht werden soll.
      @return steps Schritte die der Spieler braucht. */
    int estimatedStepsToBall( ModelledBall ball ) const;

    /** Bestimmt die Schritte, die der Spieler braucht, um den angegebenen
      Ball zu erreichen. Diese Berechnung geschieht mittels Simulation und
      ist daher sehr zeitaufwendig. Der zweite Parameter gibt daher an, ab
      wieviel Takten erst richtig simuliert werden soll.
      Diese Routine simuliert den Ball soviele Schritte, wie der Spieler
      zum Ball braucht.
      Bei Schrittzahlen gr\"o\3er als {\Ref globalConf->maxStepsToCalculate}
      wird die Simulation abgebrochen und die derzeitige Schrittzahl
      zur\"uckgegeben. Nach 5 durchgef\"uhrten Iterationen, um
      die Shrittzahl zum Ball zu bestimmen, wird ebenfalls abgebrochen und
      die derzeitige Schrittzahl zur"uckgegeben.
      @param ball Der Ball der erreicht werden soll.
      @param estimatedSteps Gesch\"atzte Zahl der Schritte. 
      @return steps Schritte die der Spieler braucht. */
    int calculatedStepsToBall( ModelledBall& ball, int estimatedSteps ) const;
  //@}    


  /** @name Konfigurierbare Werte erfragen */
  //@{
    ///
    double size() const;
    ///
    double maxSpeed() const;
    ///
    double decay() const;
    ///
    double staminaIncMax() const;
    ///
    double staminaMax() const;
    ///
    double recoverDecThr() const;
    ///
    double recoverDec() const;
    ///
    double recoverMin() const;
    ///
    double effortDecThr() const;
    ///
    double effortDec() const;
    ///
    double effortIncThr() const;
    ///
    double effortInc() const;
    ///
    double effortMin() const;
    ///
    double dashPowerRate() const;
    ///
    double maxPower() const;
    ///
    double minPower() const;
    ///
    double maxMoment() const;
    ///
    double minMoment() const;
    ///
    double normVisibleAngle() const;
    ///
    double inertiaMomentOfTurn() const;
  //@}

    
  /** @name Konfigurierbare Werte setzen */
  //@{
    /** liest die Werte aller Klassenvariablen aus dem Konfigurationsfile 
      ein. */
    void setClassVars();

    ///
    void setSize(double aPlayerSize);
    
    ///
    void setMaxSpeed(double aPlayerMaxSpeed);
    
    ///
    void setDecay(double aPlayerDecay);
    
    ///
    void setStaminaIncMax(double aStaminaIncMax);
    
    ///
    void setStaminaMax(double aStaminaMax);
    
    ///
    void setRecoverDecThr(double aRecoverDecThr);
    
    ///
    void setRecoverDec(double aRecoverDec);
    
    ///
    void setRecoverMin(double aRecoverMin);
    
    ///
    void setEffortDecThr(double aEffortDecThr);
    
    ///
    void setEffortDec(double aEffortDec);
    
    ///
    void setEffortIncThr(double aEffortIncThr);
    
    ///
    void setEffortInc(double aEffortInc);
    
    ///
    void setEffortMin(double aEffortMin);
    
    ///
    void setDashPowerRate(double aDashPowerRate);
    
    ///
    void setMaxPower(double aMaxPower);
    
    ///
    void setMinPower(double aMinPower);
    
    ///
    void setMaxMoment(double aMaxMoment);
    
    ///
    void setMinMoment(double aMinMoment);
    
    ///
    void setNormVisibleAngle(double aNormVisibleAngle);
    
    ///
    void setInertiaMomentOfTurn(double inertiaMoment);
  //@}
  
  /// Maximale Schrittzahl die ein Spieler braucht um zum Ball zu gelangen.
  static const int MAX_STEPS_TO_BALL;

  
private:
  ///
  static double thePlayerSize;
  ///
  static double thePlayerMaxSpeed;
  ///
  static double thePlayerDecay;
  ///
  static double theStaminaIncMax;
  ///
  static double theStaminaMax;
  ///
  static double theRecoverDecThr;
  ///
  static double theRecoverDec;
  ///
  static double theRecoverMin;
  ///
  static double theEffortDecThr;
  ///
  static double theEffortDec;
  ///
  static double theEffortIncThr;
  ///
  static double theEffortInc;
  ///
  static double theEffortMin;
  ///
  static double theDashPowerRate;
  ///
  static double theMaxPower;
  ///
  static double theMinPower;
  ///
  static double theMaxMoment;
  ///
  static double theMinMoment;
  ///
  static double theNormVisibleAngle;
  ///
  static double theInertiaMomentOfTurn;
  ///
  static short  isSetOnce;

  /** Die Blickrichtung als absoluter Winkel im Bogenmass. */
  double theFaceDir;

  /** Die der aktuellen Sichtfeldgroesse entsprechende Konstante */
  int theViewWidth;

  /** Der sichtbare Bereich als Winkel im Bogenmass. */
  double theVisibleAngle;
  
  /** Die aktuelle Kraftreserve. Sie sinkt beim Beschleunigen und beim Aufruf
    von usePower. Die Kraftreserve liegt anfangs bei MAX_STAMINA. */
  double theStamina;

  /** Leistungsfaktor, der aktuelle Leistungsfaehigkeit repraesentiert. Er gibt
    an, wieviel Kraft in einem Simulationsschritt tatsaechlich zur Verfuegung
    steht. Er sinkt, wenn die Kraftreserve unter einen Schwellwert faellt und
    steigt wieder, wenn sich der Spieler praktisch ganz erholt hat. */
  double theEffort;
  
  /** Erholungsfaktor, der die Langzeitausdauer repraesentiert. Er gibt an, wie
    stark die Kraftreserve in einer Ruhepause wieder ansteigt. Er wird waehrend
    eines Spiels nicht erhoeht und sinkt, wenn die Kraftreserve unter einen
    Schwellwert faellt. */
  double theRecovery;  

  /** Die Verl\"a\3lichkeit der Stamina-Werte. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Reliability theStaminaReliability;

  /** Die Verl\"a\3lichkeit der Blickrichtung. Sie haengt von der anfaenglichen
    Verlaesslichkeit und von der Zahl der durchgefuehrten Simulationsschritte
    ab. */
  Reliability theFaceDirReliability;
  
  // Methoden
  
  /** Formel mit der der Drehwinkel modifiziert wird */
  double turnAngleMult() const;

  /** Normalisiert den Stamina-Wert. */
  double normalizeStamina(double stamina) const;

  /** Normalisiert den Effort-Wert. */
  double normalizeEffort(double effort) const;

  /** Normalisiert den Recovery-Wert. */
  double normalizeRecovery(double recovery) const;

};

inline double ModelledPlayer::normalizeStamina(double stamina) const 
{
  if( stamina == UNDEF_DOUBLE ) {
    return staminaMax();
  }
  else if( stamina < 0 ) {
    return 0;
  }
  else if( stamina > staminaMax() ) {
    return staminaMax();
  }
  else{
    return stamina;
  }
}

inline double ModelledPlayer::normalizeEffort(double effort) const 
{
  if( effort == UNDEF_DOUBLE ) {
    return 1.0;
  }
  else if( effort < effortMin() ) {
    return effortMin();
  }
  else if( effort > 1 ) {
    return 1;
  }
  else{
    return effort;
  }
}

inline double ModelledPlayer::normalizeRecovery(double recovery) const 
{
  if( recovery == UNDEF_DOUBLE ) {
    return 1.0;
  }
  else if( recovery < recoverMin() ) {
    return recoverMin();
  }
  else if( recovery > 1 ) {
    return 1 ;
  }
  else {
    return recovery;
  }
}

inline double ModelledPlayer::normalizeDashPower(double power) const 
{
  if( power == UNDEF_DOUBLE ) {
    throwException( ModelledPlayerException( "undef power value" ) );
  }
  else if( power < minPower() ) {
    return minPower();
  }
  else if( power > maxPower() ){
    return maxPower();
  }
  else {
    return power;
  }
}

inline double ModelledPlayer::normalizeTurnAngle(double angle) const 
{
  angle = normalizeAngle( angle );
  
  if( angle < minMoment() ) {
    return minMoment();
  }
  else if( angle > maxMoment() ) {
    return maxMoment();
  }
  else {
    return angle;
  }
}

inline double ModelledPlayer::distTo( const Vector& position ) const
{
  return ( (position - pos()).length() );
}

inline double ModelledPlayer::size() const 
{
  return ModelledPlayer::thePlayerSize;
}

inline double ModelledPlayer::maxSpeed() const 
{
  return ModelledPlayer::thePlayerMaxSpeed;
}

inline double ModelledPlayer::decay() const 
{
  return ModelledPlayer::thePlayerDecay;
}

inline double ModelledPlayer::staminaIncMax() const 
{
  return ModelledPlayer::theStaminaIncMax;
}

inline double ModelledPlayer::staminaMax() const 
{
  return ModelledPlayer::theStaminaMax;
}

inline double ModelledPlayer::recoverDecThr() const 
{
  return ModelledPlayer::theRecoverDecThr;
}

inline double ModelledPlayer::recoverDec() const 
{
  return ModelledPlayer::theRecoverDec;
}

inline double ModelledPlayer::recoverMin() const 
{
  return ModelledPlayer::theRecoverMin;
}

inline double ModelledPlayer::effortDecThr() const 
{
  return ModelledPlayer::theEffortDecThr;
}

inline double ModelledPlayer::effortDec() const 
{
  return ModelledPlayer::theEffortDec;
}

inline double ModelledPlayer::effortIncThr() const 
{
  return ModelledPlayer::theEffortIncThr;
}

inline double ModelledPlayer::effortInc() const 
{
  return ModelledPlayer::theEffortInc;
}

inline double ModelledPlayer::effortMin() const 
{
  return ModelledPlayer::theEffortMin;
}

inline double ModelledPlayer::dashPowerRate() const 
{
  return ModelledPlayer::theDashPowerRate;
}

inline double ModelledPlayer::maxPower() const 
{
  return ModelledPlayer::theMaxPower;
}

inline double ModelledPlayer::minPower() const 
{
  return ModelledPlayer::theMinPower;
}

inline double ModelledPlayer::maxMoment() const 
{
  return ModelledPlayer::theMaxMoment;
}

inline double ModelledPlayer::minMoment() const 
{
  return ModelledPlayer::theMinMoment;
}

inline double ModelledPlayer::normVisibleAngle() const 
{
  return ModelledPlayer::theNormVisibleAngle;
}

inline double ModelledPlayer::inertiaMomentOfTurn() const 
{
  return ModelledPlayer::theInertiaMomentOfTurn;
}

inline void ModelledPlayer::setSize(double aPlayerSize) 
{
  ModelledPlayer::thePlayerSize = aPlayerSize;
}

inline void ModelledPlayer::setMaxSpeed(double aPlayerMaxSpeed) 
{
  ModelledPlayer::thePlayerMaxSpeed=aPlayerMaxSpeed;
}

inline void ModelledPlayer::setDecay(double aPlayerDecay) 
{
  ModelledPlayer::thePlayerDecay=aPlayerDecay;
}

inline void ModelledPlayer::setStaminaIncMax(double aStaminaIncMax) 
{
  ModelledPlayer::theStaminaIncMax=aStaminaIncMax;
}

inline void ModelledPlayer::setStaminaMax(double aStaminaMax) 
{
  ModelledPlayer::theStaminaMax=aStaminaMax;
}

inline void ModelledPlayer::setRecoverDecThr(double aRecoverDecThr) 
{
  ModelledPlayer::theRecoverDecThr=aRecoverDecThr;
}

inline void ModelledPlayer::setRecoverDec(double aRecoverDec) 
{
  ModelledPlayer::theRecoverDec=aRecoverDec;
}

inline void ModelledPlayer::setRecoverMin(double aRecoverMin) 
{
  ModelledPlayer::theRecoverMin=aRecoverMin;
}

inline void ModelledPlayer::setEffortDecThr(double aEffortDecThr) 
{
  ModelledPlayer::theEffortDecThr=aEffortDecThr;
}

inline void ModelledPlayer::setEffortDec(double aEffortDec) 
{
  ModelledPlayer::theEffortDec=aEffortDec;
}

inline void ModelledPlayer::setEffortIncThr(double aEffortIncThr) 
{
  ModelledPlayer::theEffortIncThr=aEffortIncThr;
}

inline void ModelledPlayer::setEffortInc(double aEffortInc) 
{
  ModelledPlayer::theEffortInc=aEffortInc;
}

inline void ModelledPlayer::setEffortMin(double aEffortMin) 
{
  ModelledPlayer::theEffortMin=aEffortMin;
}

inline void ModelledPlayer::setDashPowerRate(double aDashPowerRate) 
{
  ModelledPlayer::theDashPowerRate=aDashPowerRate;
}

inline void ModelledPlayer::setMaxPower(double aMaxPower) 
{
 ModelledPlayer::theMaxPower=aMaxPower;
}

inline void ModelledPlayer::setMinPower(double aMinPower) 
{
  ModelledPlayer::theMinPower=aMinPower;
}

inline void ModelledPlayer::setMaxMoment(double aMaxMoment) 
{
  ModelledPlayer::theMaxMoment=aMaxMoment;
}

inline void ModelledPlayer::setMinMoment(double aMinMoment) 
{
  ModelledPlayer::theMinMoment=aMinMoment;
}

inline void ModelledPlayer::setNormVisibleAngle(double aNormVisibleAngle) 
{
  ModelledPlayer::theNormVisibleAngle=aNormVisibleAngle;
}

inline void ModelledPlayer::setInertiaMomentOfTurn(
    double aInertiaMomentOfTurn) 
{
  ModelledPlayer::theInertiaMomentOfTurn=aInertiaMomentOfTurn;
}

inline double ModelledPlayer::stamina() const 
{
  return theStamina;
}

inline void ModelledPlayer::setStamina( double newStamina ) 
{
  theStamina = normalizeStamina( newStamina );
}

inline double ModelledPlayer::effort() const 
{
  return theEffort;
}

inline void ModelledPlayer::setEffort( double newEffortFactor ) 
{
  theEffort = normalizeEffort( newEffortFactor );
}

inline double ModelledPlayer::recovery() const 
{
  return theRecovery;
}

inline void ModelledPlayer::setRecovery( double newRecoveryFactor ) 
{
  theRecovery = normalizeRecovery( newRecoveryFactor );
}  

inline double ModelledPlayer::faceDir() const 
{
  return theFaceDir;
}

inline void ModelledPlayer::setFaceDir( double newFaceDir ) 
{
  theFaceDir=newFaceDir;
}

inline int ModelledPlayer::viewWidth() const 
{
  return theViewWidth;
}

inline double ModelledPlayer::visibleAngle() const 
{
  return theVisibleAngle;
}

inline const Reliability& ModelledPlayer::staminaReliability() const 
{
  return theStaminaReliability;
}

inline void ModelledPlayer::setStaminaReliability(
    const Reliability& newStaminaReliability) 
                               {
  theStaminaReliability = newStaminaReliability;
}

inline const Reliability& ModelledPlayer::faceDirReliability() const 
{
  return theFaceDirReliability;
}

inline void ModelledPlayer::setFaceDirReliability(
    const Reliability& newFaceDirReliability) 
{
  theFaceDirReliability = newFaceDirReliability;
}

#endif


