#ifndef _SENSORS_H_
#define _SENSORS_H_

#include "Communicator.h"
#include "Array.h"
#include "StableObject.h"
#include "MovingObject.h"
#include "MovingPlayer.h"
#include "Message.h"
#include "BodyInfos.h"
#include "Exception.h"

/** Das Sensorik-Objekt. 
  Die Sensorik wartet auf Sensorinformationen vom Server
  und erzeugt eine Objekt-Repraesentation von ihnen. Diese Sensorinformationen
  werden dann vom Weltmodell ausgelesen.

  Bereits bei der Erzeugung der Objektdarstellung der Wahrnehmung wird vom
  Server- in das mathematische Koordinatensystem gewandelt.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur Vererbung
  geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Sensors.h} | [\URL[Implementierung]{src/Sensors.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Sensors.h,v $
  *   $Revision: 1.26 $
  *   $Date: 1998/08/14 13:21:12 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Sensors
{
public:

  /** Interne Fehlerklasse. */
  class SensorsException: public Exception {
  public:
    SensorsException(const String& aDescription )
      : Exception( aDescription ) {}
  };
  
  /** Erzeugt das Sensorobjekt.
    @param aCommunicator Das Communicator-Objekt, von dem die 
    Sensor-Informationen kommen.  */
  Sensors( Communicator* aCommunicator );
		
  /** Initialisieren der Sensordaten. Die Arrays der Sensordaten werden
    geloescht. Somit steht das Sensorobjekt fuer den Empfang neuer
    Daten bereit. Diese Methode muss nach dem kompletten Lesen der
    aufgelaufenen
    Daten aufgerufen werden um damit kenntlich zu machen, dass man nur noch
    an neueren Daten interessiert ist. Diese Methode wird automatisch bei
    eintreffen einer neuen Sensorinformation aufgerufen. */
  void initSensors();

  /** Empfaengt Sensorstrings und verarbeitet sie zu Objekten.
    Versucht vom eigenen Communicator Strings zu lesen. Diese werden zu
    Objekten verarbeitet und spaeter vom Weltmodell ausgelesen.

    Nach einem Aufruf von Sense liegen keine Strings mehr im Communicator
    bereit. 
    @exception SensorsException wenn der Server l\"anger als
    \Ref{Configuration::communicationTimeout}() Sekunden keine
    Sichtinformationen geschickt hat. Man kann dann davon ausgehen, da\3 der
    Server sich terminiert hat. */
  void sense();

  /** Ein String vom SoccerServer wird geparst. Die zu diesem String
    gehoerenden Objekte werden aufgebaut. */
  void parseSensoryData(const String& message);

  /** Anfrage ob eine neue Sichtinformation vorliegt. Falls dies der Fall
    ist wird YES zurueckgegeben, andernfalls NO. Diese Methode dient dem
    Weltmodell um zu erkennen, ob sich die Welt geaendert hat. Neue Daten
    liegen vor, wenn nach dem letzten initSensors() Sensordaten ausgewertet
    wurden. */
  bool isNewSeeInfo();

  /** Anfrage ob eine neue Schiedsrichterentscheidungen vorliegen. Falls dies
    der Fall
    ist wird YES zurueckgegeben, andernfalls NO. Diese Methode dient dem
    Weltmodell um zu erkennen, ob sich die Welt geaendert hat. Neue Daten
    liegen vor, wenn nach dem letzten initSensors() Schiedsrichterdaten
    ausgewertet wurden. */
  bool isNewRefereeInfo();

  /** Liefert ein Array mit den wahrgenommenen Flaggen. Zu diesen gehoeren 
    auch die Tore. Die Informationen entstammen der see-Nachricht des
    SoccerServers. */
  const Array<StableObject>& sensedFlags() const;
  
  /** Liefert ein Array mit den wahrgenommenen Linien. Die Position der Linien
    ist besonders zu interpretieren: Die Laenge ist die Entfernung zwischen der
    eigenen Position und dem Schnittpunkt der eigenen Blickgeraden und der
    Linie. Der Winkel ist der Winkel der Linie relativ zur eigenen
    Blickrichtung. Die Informationen zu den Linien entstammen der
    see-Nchricht des SoccerServers. */
  const Array<StableObject>& sensedLines() const;
  
  /** Die gesehenen Spieler. Diese entstammen der see-Nachricht des
    SoccerServers. */
  const Array<MovingPlayer>& sensedPlayers() const;
  
  /** Der gesehene Ball. Falls keiner gesehen wurde, wird eine 
    NothingSeen Exception ausgeloest. Der Ball wird der see-Nachricht
    des SoccerServers entnommen. */
  const MovingObject& sensedBall() const;
  
  /** Der Simulationsschritt zu der die Sichtinformationen gehoeren. 
    Diese Angabe gilt fuer sensedFlags, sensedLines, sensedPlayers und 
    sensedBall. */
  long simStep();

  /** Der Simulationsschritt in der der sich der SoccerServer befindet.
    Dieser Simulationsschritt basiert auf der letzten mitbekommenden
    Zeit (see, bodyInfos oder Referee ). */
  long simStepOfSoccerServer();

  /** Die gehoerten Nachrichten. Diese entstammen der hear-Nachricht des
    SoccerServers. */
  const Array<Message>& heardMessages() const;

  /** Die Informationen ueber den eigenen Koerper. Diese entstammen der 
    sense_body-Nachricht des SoccerServers. */
  const Array<BodyInfos>& bodyInfos() const;
  
  /** Die letzten Informationen \"uber den eigenen K\"orper. */
  const BodyInfos& lastBodyInfos() const;

private:
  /** Interne Referenz auf Communicator Objekt */
  Communicator *theCommunicator;
  
  /** Die gesehenen Flaggen inclusive Toren. */
  Array<StableObject> theSensedFlags;
  
  /** Die gesehenen Linien. */
  Array<StableObject> theSensedLines;	
		
  /** Die gesehenen Spieler. */
  Array<MovingPlayer> theSensedPlayers;

  /** Der gesehene Ball. */
  MovingObject theSensedBall;

  /** Der Simulationsschritt, zu dem die Sichtinformationen gehoeren. */
  long theSimStep;

  /** Der Simulationsschritt des SoccerServers. */
  long theSimStepOfSoccerServer;

  /** Die gehoerten und aufbereiteten Nachrichten. */
  Array<Message> theHeardMessages;

  /** Die Informationen ueber den eigenen Koerper. */
  Array<BodyInfos> theBodyInfos;
  
  /** Die letzten Informationen \"uber den eigenen K\"orper. */
  BodyInfos theLastBodyInfos;

  /** Liegt neue Sichtinformation vor? */
  bool theIsNewSeeInfo;

  /** Liegen neue Schiederichterentscheidungen vor? */
  bool theIsNewRefereeInfo;

  /** Teil einer Nachricht. Da diese als const uebergeben werden soll,
    wurde sie hier global deklariert. */
  String partOfMessage;

  /** Parsen einer Sichtinformation. */
  void parseSights(char* seeMessage);
};

inline const Array<StableObject>& Sensors::sensedFlags() const {
  return theSensedFlags;
}
  
inline const Array<StableObject>& Sensors::sensedLines() const {
  return theSensedLines;
}
  
inline const Array<MovingPlayer>& Sensors::sensedPlayers() const {
  return theSensedPlayers;
}
  
inline const MovingObject& Sensors::sensedBall() const {
  return theSensedBall;
}
  
inline long Sensors::simStep() {
  return theSimStep;
}

inline long Sensors::simStepOfSoccerServer() {
  return theSimStepOfSoccerServer;
}

inline const Array<Message>& Sensors::heardMessages() const {
  return theHeardMessages;
}

inline const Array<BodyInfos>& Sensors::bodyInfos() const {
  return theBodyInfos;
}

inline const BodyInfos& Sensors::lastBodyInfos() const {
  return theLastBodyInfos;
}

inline bool Sensors::isNewSeeInfo() {
  return theIsNewSeeInfo;
}

inline bool Sensors::isNewRefereeInfo() {
  return theIsNewRefereeInfo;
}

#endif




