#ifndef _SKILLS_H_
#define _SKILLS_H_

#include "WorldModel.h"
#include "Plan.h"
#include "Vector.h"
#include "ModelledBall.h"
#include "ModelledPlayer.h"

/**
  Die F\"ahigkeiten des Agenten. 
  Options verwendet Skills, um Plaene zu erzeugen. Ein Plan ist dabei eine
  Folge von konkreten Aktionen, die an den SoccerServer geschickt werden 
  koennen.

  Im Skills-Objekt gibt es eine Ausgangssituation, die anf\"anglich auf die
  aktuelle Situation des Weltmodells gesetzt ist. Jeder Aufruf einer F\"ahigkeit
  f\"uhrt zu \"Anderungen an dieser Situation, so da\3 der beim Aufruf der
  n\"achsten F\"ahigkeit die neue, modifizierte Situation als Ausgangssituation
  genomment werden kann. Vor der Berechnung jedes Plans mu\3 die
  Ausgangssituation mit \Ref{resetSituation}() auf die Gegenwart zur\"uckgesetzt
  werden. Eine Referenz auf die Situation ist intern \"uber die private Methode
  \Ref{situation}() verf\"ugbar.
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/Skills.h} | [\URL[Implementierung]{src/Skills.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Skills.h,v $
  *   $Revision: 1.30 $
  *   $Date: 1998/08/14 13:21:17 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Skills
{
public:

  /** Minimale Drehung, die ausgef\"uhrt werden soll. */
  const double MIN_TURN_ANGLE_IN_DEGREE = 1.0;

  /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    Die Konstanten k\"onnen addiert werden, um mehrere
    Verhaltensweisen gleichzeitig angeben zu k\"onnen. 
    @see gotoRegion */
  const int STANDARD_BEHAVIOR = 0;
   /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    @see gotoRegion */
  const int AVOID_OBSTACLES = 1;
  /** Konstante zur Steuerung des Verhaltens f"ur gotoRegion.
    @see gotoRegion */
  const int RUN_BACKWARDS = 2;
  

  /** Initialisiert ein Skills-Objekt.
    Es ist nicht sinnvoll mehrere Skills-Objekte zu haben. */
  Skills();

  /**
    Erzeugt einen Plan um den Spieler auf die angegebene Position zu
    beamen. Das Sichtfeld wird dabei wieder auf das maximal moegliche
    Sichtfeld eingestellt.
    @param position Posistion auf die gebeamt werden soll. */
  Plan beamToPosition( const Vector& position );

  /** Berechnet den Plan, um zu einer gegebenen Region zu gelangen.
    Diese Region ist kreisfoermig und wird durch einen Mittelpunkt und einen
    Radius bestimmt. Der Radius darf nicht kleiner als TRIANGULATION_ACCURACY
    sein.
    @param centerOfTarget Mittelpunkt der Zielregion.
    @param targetRadius Der Radius der Zielregion. 
    @param staminaLimit die untere Staminagrenze (default 0)
    @param specialBehavior. Hier kann eine Summe von Verhaltenskonstanten
    angegeben werden. M\"oglich sind bisher: AVOID_OBSTACLES; wenn dieser Wert 
    angegeben ist, werden Hindernisse (also andere Spieler) umgangen (default).
    RUN_BACKWARDS; ist dieser Wert angegeben, rennt der Spieler r\"uckw\"arts
    (was wesentlich langsamer ist). STANDARD_BEHAVIOR: Halt nix besondres.
    @param desiredMaxLengthOfPlan gew\"unschte L\"ange des Planes. 
    (default UNDEF_INT was dann maxLengthOfPlan ergibt). */
  Plan gotoRegion( const Vector& centerOfTarget, 
      double targetRadius = TRIANGULATION_ACCURACY,
      const double staminaLimit = 0, 
      const int specialBehavior = AVOID_OBSTACLES,
      int desiredMaxLengthOfPlan = UNDEF_INT );

  /** Berechnet den Plan, um einen gegebenen Kick auszufuehren.
    Kraft und Richtung sind absolute Werte. */
  Plan kickBall( const double power, const double direction );
  
  /** Berechnet einen Plan, um die alleinige Ballkontrolle zu erlangen.
    Dabei wird einfach um 90 Grad gedreht, dann in die Richtung geschossen,
    und dann hinterhergerannt. Durch die vorherige Drehung in die 
    Laufrichtung erhalten wir (hoffentlich) einen Vorsprung, der der
    InterceptBallOption gen\"ugen sollte, um zuerst am Ball zu sein. */
  Plan gainBallControl();

  /** Berechnet den Plan fuer ein Dribbel-Manoever.
    Die Richtung ist ein absoluter Wert. 
    @param staminaLimit Der Staminawert, der nicht unterschritten werden darf.
    @param direction Die Richtung, in die gedribbelt werden soll.
    @param side Die Seite des Spielers (in Blickrichtung), auf der der Ball
      gehalten werden soll. M\"ogliche Werte sind LEFT_SIDE_ID, RIGHT_SIDE_ID,
      UNDEF_SIDE_ID. */
  Plan dribble( 
      const double staminaLimit, 
      const double direction, 
      const int side );

  /* Alternative Dribble-Faehigkeit, die einfach nur schwach nach vorne
    schiesst. */
  Plan trivialDribble( 
      const double staminaLimit, 
      const double direction, 
      const int side );

  /** Berechnet den Plan fuer eine Drehung auf einen Winkel.
    @param finalDirection Die gewuenschte neue Blickrichtung. Das ist der Winkel
    eines Vektors von der absoluten Spielerposition zu einer absoluten
    Zielposition. direction ist also unabh\"angig von der momentanen 
    Blickrichtung des Spielers.
    @return Ein Plan, der aus mehreren Turn-Aktionen besteht. 
    @see turnBy */
  Plan turnTo( double finalDirection );

  /** Berechnet den Plan fuer eine Drehung um einen Winkel.
    @param direction Die gewuenschte neue Blickrichtung. Das ist der Winkel
    eines Vektors von der absoluten Spielerposition zu einer absoluten
    Zielposition. direction ist also unabh\"angig von der momentanen 
    Blickrichtung des Spielers.
    @return Ein Plan, der aus mehreren Turn-Aktionen besteht. 
    @see turnTo */
  Plan turnBy( double turnAngle );
  
  /** Erzeugt einen Plan zum Umstellen der Breite des Blickwinkels.
    Der Plan kann auch leer sein. */
  Plan setViewWidth( const desiredViewWidth );

  
  /** Setzt die Rahmenparameter f\"ur die Suche nach einem freien Sektor.
    Es k\"onnen mehrere Suchen mit denselben Rahmenparametern (z.B. in
    verschiedenen Situationen, also auch mit verschiedenen eigenen Positionen
    durchgef\"uhrt werden. 
    @param aFinalTarget Die absolute Zielposition. Hier ist ein Absolutwert
    notwendig, damit das Ziel unabh\"angig von der eigenen Position angegeben
    werden kann.
    @param aSectorAngle die gew\"unschte Breite des freien Sektors im 
    Bogenma\3. 
    @param aSectorLength die gew\"unschte Laenge des Sektors in Metern.
    @param aMaxDeviation die maximal akzeptable Abweichung der Sektorrichtung
    von der Richtung zum Ziel.
    @param aStepAngle die gew\"unschte Genauigkeit, mit der gesucht werden soll,
    im Bogenma\3. Ist der Wert klein, wird evtl. ein Sektor gefunden, der mehr
    zum Ziel hinzeigt als bei gr\"o\3eren Werten. Eine h\"ohere Genauigkeit
    (also kleinere stepAngle-Werte) f\"uhrt allerdings auch zu einer l\"angeren
    Rechenzeit. */
  void initEmptySectorSearch( 
      const Vector& aFinalTarget,
      const double aSectorAngle,
      const double aSectorLength,
      const double aMaxDeviation,
      const double aStepAngle );

  /** Sucht einen freien Sektor auf dem Weg zu einem Ziel.
    Vor der Suche m\"ussen die Rahmenparameter mit der Funktion
    \Ref{initEmptySectorSearch} gesetzt werden.
    Es wird ein Sektor gesucht, in dem keine Spieler stehen. Die abgeschnittene
    Spitze des Sektors liegt beim Spieler und ist dort so breit wie der Spieler.
    Die Mitte des Sektors zeigt so weit wie m\"oglich zum Ziel hin.
    @param Die Situation, in der der Sektor gesucht wird.
    @return Vektor, der einem Dash- oder Kick-Kommando \"uebergeben werden mu\3,
    damit der Spieler durch die Mitte des freien Sektors l\"auft/schie\3t. Gibt
    es keinen freien Sektor, wird ein Vektor zur\"uckgegeben, dessen
    \Ref{isValid}-Methode NO liefert. */
  Vector findEmptySector( const Situation& aSituation ) const;
    
  /** Berechnet den Plan fuer den WorldModell-Ruf. */
   Plan sayWorldModel();


  
  /** Berechnet einen leeren Plan. 
    @param steps L\"ange des Plans. (default: maxLengthOfPlan )*/
  Plan doNothing(int steps = UNDEF_INT );
  

  /** Liefert die Aktion fuer eine Sichtaenderung. */
  Action simpleChangeView( const int width, const int quality );

  /** Liefert die Aktion fuer ein Move auf die angegebene Position. */ 
  Action simpleMoveTo( const Vector& position );

  /** Liefert den Plan fuer eine Drehung um den angegebenen Winkel. */
  Action simpleTurn( const double angle );

  /** Liefert die Aktion fuer eine Beschleunigung um die angegebene Staerke. */
  Action simpleDash( const double power );

  /** Liefert die Aktion fuer ein Kick.
    Der angegebene Winkel ist relativ zum Spieler. */
  Action simpleKick( const double power, const double angle );

  /** Liefert die Aktion fuer das Rufen einer Nachricht. */
  Action simpleSay( const String& message );

  /** Liefert die Aktion fuer das Greifen nach dem Ball (nur fuer Torwart). */
  Action simpleCatch( const double angle );
  
  /** Liefert die Aktion zur Anfrage ueber den Koerper. */
  Action simpleSenseBody( );

  /** Liefert eine leere Aktion. */
  Action simpleEmptyAction( );
  
  /** Setzt die Ausgangssituation f\"ur die einzelnen Skills zur\"uck.
    Nach dem Aufruf dieser Routine ist die Ausgangssituation die aktuelle
    Situation des Weltmodells. 
    
    Diese Routine mu\3 unbedingt jedesmal aufgerufen werden, bevor ein 
    Plan erstellt wird, der von der Gegenwart ausgeht. Jeder Aufruf
    einer F\"ahigkeit \"andert dann die Ausgangssituation.
    @param scenario Szenario welches die Ausgangssituation auf ein anderes
    als das aktuelle Weltbildszenario setzt. */
  void resetSituation( Situation scenario = globalWorldModel->situation() );
  
private:
  /** Liefert die Situation, von der die n\"achste Skill-Routine ausgeht.
    Der simStep dieser Situation ist gr\"osser/gleich der aktuellen Situation
    im Weltmodell.
    @return eine Referenz auf die Ausgangssituation. Sie darf modifiziert 
    werden. */
  Situation& situation();

  /** Liefert den ModelledPlayer des Agenten der Situation. */
  const ModelledPlayer& player() const;

  /** Liefert den ball der Situation. */
  const ModelledBall& ball() const;

  /** Liefert einen wahren Schu\3vektor, d.h. einen Vektor der
    die wahre Zielrichtung ergibt. */
  Vector getKickVector(const Vector& desiredKick, const ModelledBall& ball, 
		       const ModelledPlayer& player) const;
  
  /** Liefert den f"ur einen Pass aus den Zielvektor desiredKick
    n"otigen wahren Kickvektor. */
  Vector getPassVector(const Vector& desiredKick, 
		       const ModelledBall& ball, 
		       const ModelledPlayer& player) const;

  /** Test ob im n"achsten Takt eine Kollision stattfindet. */
  bool checkBodyContact(const Vector& desiredKick, 
			const ModelledBall& cBall, 
			const ModelledPlayer& cPlayer) const;

  /** Beste Alternative wenn der urspr"ungliche Kickvektor
    zur Kollision fuehrt. */
  Vector bestKickVector(const Vector& desiredKick, 
                        const Vector& calculatedKick,
                        const ModelledBall& cBall, 
			const ModelledPlayer& cPlayer) const;

  /** Berechnet den KickVector aus dem Entfernungsvector, 
    den der Ball in steps Schritten zuruecklegen soll,
    unter Zuhilfenahme von getKickVector.
   */
  Vector calcKickToRelPos(  const Vector& aRelativeVector, 
				 const ModelledBall& aBall, 
				 const ModelledPlayer& aPlayer, 
				 int steps = 1) const;

  /** YES wenn im naechsten Schritt bei Kick mit desiredKick, 
    der Ball noch im Kickbereich und auf richtiger Seite und 
    nicht hinter dem Spieler. NO sonst 
    */
  bool checkBallInDribbleKickRange( const Vector& desiredKick, 
				    const ModelledBall& cBall, 
				    const ModelledPlayer& cPlayer, int side) const;
  /** YES wenn im naechsten Schritt bei Kick mit desiredKick
    Ball noch im Kickbereich. NO sonst
    */
  bool checkBallInKickRange( const Vector& desiredKick, 
			     const ModelledBall& cBall, 
			     const ModelledPlayer& cPlayer ) const;

  /** Liefert beim dribbeln den relativen Vector zur gewuenschten Ballposition pos
   */
  Vector desiredBallPos( int apos, const int aside ) const;
  
  /** Ermittelt einen Kick, der am Kickradius endet.
    @param desiredKick Vector, der die gewuenschte Kickrichtung enthaelt */
  Vector Skills::getKickToRadius( const Vector& desiredKick, 
			        const ModelledBall& aBall, 
			        const ModelledPlayer& aPlayer ) const;

  /** Macht den Kick relativ zum Spieler.
    Die Laenge des Vektors ist der Power-Wert, der simpleKick uebergeben werden
    kann. */
  Vector makeRelativeKick( const Vector& desiredKick, 
			   const ModelledBall& ball, 
			   const ModelledPlayer& player ) const;

  /** untere Grenze fuer Kick-Entfernung (playerRadius + ballRadius). */
  double minKickRadius;
  /** obere Grenze fuer Kick-Entfernung (kickRange). */
  double maxKickRadius;
  /** optimale Entfernung von Ball und Spieler fuer kick (optimalKickDist). */
  double optimalKickRadius;
  /** maximale Geschwindigkeit eines Spielers. */
  double maxSpeedOfPlayer;
  /** maximale Geschwindigkeit des Balls. */
  double maxSpeedOfBall;
  /** maximale Laenge eines Planes. */
  int maxLengthOfPlan;
  /** KickPowerRate. */
  double kickPowerRate;
  /** Maximale Abweichung/Toleranz f"ur Kickroutine am Au\3enradius. */
  double outerKickCircleTolerance;
  /** Maximale Abweichung/Toleranz f"ur Kickroutine am Innenradius. */
  double innerKickCircleTolerance;
  /** Maximale Abweichung/Toleranz f"ur Kickroutine f\"ur Zwischenziele. */
  double intermediateKickTolerance;
  
  /** Ziel fuer die Suche nach einem freien Sektor */
  Vector theFinalTarget;
  /** Breite des freien Sektor im Bogenma\3 */
  double theSectorAngle;
  /** L\"ange des freien Sektor in Metern */
  double theSectorLength;
  /** Schrittweite, in der nach einem Sektorwinkel gesucht wird */
  double theStepAngle;
  /** Maximaler Winkel, um den der gefundene Sektor von der Zielrichtung
    abweichen darf (im Bogenma\3). */
  double theMaxDeviation;
  
  /** Ziel fuer die Suche nach einem freien Sektor */
  const Vector& finalTarget() const;
  /** Breite des freien Sektor im Bogenma\3 */
  double sectorAngle() const;
  /** L\"ange des freien Sektor in Metern */
  double sectorLength() const;
  /** Schrittweite, in der nach einem Sektorwinkel gesucht wird */
  double stepAngle() const;
  /** Maximaler Winkel, um den der gefundene Sektor von der Zielrichtung
    abweichen darf (im Bogenma\3). */
  double maxDeviation() const;
  
  /** Situation, auf der die einzelnen Skills arbeiten */
  Situation theSituation;

  const double DELTA_KICK_ANGLE = 0.05;
  const int MIN_DRIBBLE_PLAN_LENGTH = 5;
  };

/** Global verfuegbare Skills. */
extern Skills *globalSkills;


inline const Vector& Skills::finalTarget() const
{
  return theFinalTarget;
}

inline double Skills::sectorAngle() const
{
  return theSectorAngle;
}

inline double Skills::sectorLength() const
{
  return theSectorLength;
}

inline double Skills::stepAngle() const
{
  return theStepAngle;
}

inline double Skills::maxDeviation() const
{
  return theMaxDeviation;
}

inline Situation& Skills::situation()
{
  return theSituation;
}

inline const ModelledPlayer& Skills::player() const
{
  return theSituation.playerWithId( globalWorldModel->id() );
}

inline const ModelledBall& Skills::ball() const
{
  return theSituation.ball();
}
#endif
