#ifndef _STABLEOBJECT_H_
#define _STABLEOBJECT_H_

#include <stream.h>

#include "Common.h"
#include "Vector.h"
#include "String.h"
#include "StorableObject.h"

/**
  Die Objekt-Repraesentation eines in der Aussenwelt vorkommenden Objektes.
  Solche Objekte sind Flaggen, Linien, Tore, Spieler und der Ball.

  Ein StableObject hat eine Position (ein 2D-Vektor) und eine Id, in der
  kodiert ist, um welchen Typ von Objekt es sich handelt, und welches der
  Objekte dieses Typs das StableObject repraesentiert.

  StableObject ist gleichzeitig die Basisklasse fuer weitere spezialisierte
  Repraesentationsklassen wie MovingObject oder MovingPlayer.

  StableObjects kommen vor allem in der Wahrnehmung vor. Sensors erzeugt sie.
  Das Weltmodell erstellt aus ihnen eine neue Situation.

  Die in einem StableObject aufgenommenen Daten sind immer im mathematischen
  Koordinatensystem. Das Weltmodell muss sie also bei der Berechnung der
  absoluten Koordinaten nicht mehr umwandeln. 

  Ob die Daten als relativ zum Agenten und seiner Blickrichtung oder als
  absolut (d.h. relativ zur Spielfeldmitte bei zum Gegner zeigender positiver
  x-Achse) interpretiert werden, ist nicht festgelegt. \Ref{Sensors} setzt die
  Daten relativ zum Agenten, \Ref{WorldModel} setzt die Daten in Subklassen
  (\Ref{ModelledObject} und davon abgeleitet \Ref{ModelledPlayer} und
  \Ref{ModelledBall} von StableObject absolut.

  {\small {\bf Quelltext}

  [\URL[Header]{src/StableObject.h} | [\URL[Implementierung]{src/StableObject.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: StableObject.h,v $
  *   $Revision: 1.19 $
  *   $Date: 1998/08/14 13:21:19 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class StableObject: public StorableObject
{
public:

  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Default-Konstruktor. 
      Id ist UNDEF_ID und die Position ist der undefinierte Vektor.*/
    StableObject();
  
    /**  Initialisierung mit einem Vektor. 
      @param anId die Id des Objektes. Moegliche Werte sind in \Ref{Objekt-Ids}
      aufgelistet.
      @param aVector die Position des Objektes. Das Koordinatensystem h\"angt
      vom Verwendungszweck ab.*/
    StableObject( int anId, const Vector& aVector );  

    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    virtual StorableObject* copy() const;
  
    /** Virtueller Destruktor. */  
    virtual ~StableObject();
  //@}
  
  /** @name Werte setzen */
  //@{
    /** Setzt die Position. */
    void setPos(const Vector& newPos);

    /** Setzt die ObjectId. */
    void setId(int);
  //@}
  
  /** @name Werte abfragen */
  //@{
    /** Liefert die Position des Objekts. 
      Das Koordinatensystem haengt von der Verwendung ab. */
    const Vector& pos() const;
	
    /** Liefert die Id des Objekts. */
    int id() const;
  
    /** Test auf Gueltigkeit der Id. */
    bool isValid() const;
  
    /** Prueft, ob das Objekt komplett definiert ist.. 
      @return YES falls die Id verschieden von UNDEF_ID ist und die Position
        complete ist, sonst NO */
    bool isComplete() const;
  
    /** Prueft, ob das StableObject Informationen enthaelt. 
      @return YES, falls die Id gleich UNDEF_ID ist, sonst NO */
    bool isUndef() const;

    /** Beschreibung als String fuer die Ausgabe auf einem Terminal.
      Es werden ID und Position ausgegeben. */
    virtual String toString() const;
  //@}

  
private:
  /** Die Id des Objekts. Das ist KEINE Id im Sinne objektorientierter
    Programmierung. Hier stehen vielmehr Zahlen, die fuer bestimmte Objekte in
    der Aussenwelt stehen. */
  int theId;
  
  /** Die Position. Hier koennen nur relative Koordinaten enthalten
    sein. Winkel sind immer im Bogenmass, das Koordinatensystem ist das
    mathematische. */
  Vector thePos;
  
};

inline const Vector& StableObject::pos() const 
{
  return thePos;
}

inline int StableObject::id() const 
{
  return theId;
}

inline void StableObject::setPos(const Vector& newPos) 
{
  thePos = newPos;
}

inline void StableObject::setId(int newId) 
{
  theId = newId;
}

inline bool StableObject::isValid() const
{
  return (theId != UNDEF_ID);
}

inline bool StableObject::isComplete() const
{
  return (theId != UNDEF_ID && thePos.isValid() );
}

inline bool StableObject::isUndef() const
{
  return (theId == UNDEF_ID);
}

#endif






