#ifndef _STORABLE_OBJECT_H_
#define _STORABLE_OBJECT_H_

#include <stream.h>

class String;

/** Abstrakte Superklasse aller dynamischen Objekte. 
  Dynamische Objekte werden zum Datenaustausch zwischen den und innerhalb der
  Hauptkomponenten (Deliberator, WorldModel, Sensors, Effectors) eingesetzt.

  Der Name StorableObject deutet auf eine wesentliche Eigenschaft der Klasse
  hin: StorableObjects haben einen virtuellen Destruktor und einen "virtuellen
  Konstruktor". Genauer gesagt, eine virtuelle Methode \Ref{copy}(), die eine
  Kopie des Objekts auf dem Heap anlegt und einen Zeiger darauf
  zur\"uckliefert. Da der Name copy bei allen abgeleiteten Klassen gleich ist
  (im Gegensatz zum Copy-Konstruktor), k\"onnen in einem Array Zeiger auf
  Objekte verschiedener Klassen verwaltet werden. 

  In StorableObject ist copy() eine pure abstract function, d.h. sie mu\3 in
  der
  abgeleiteten Klasse \"uberschrieben werden, damit Instanzen der Klasse
  erzeugt
  werden k\"onnen.

  Ein StorableObject kann ausserdem einen Beschreibungsstring von sich erzeugen
  und sich auf einen Stream ausgeben. Zur Erzeugung des Strings dient die
  Funktion \Ref{toString}(), die ebenfalls pure abstract ist.

  {\small {\bf Quelltext}

  [\URL[Header]{src/StorableObject.h} | [\URL[Implementierung]{src/StorableObject.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: StorableObject.h,v $
  *   $Revision: 1.8 $
  *   $Date: 1998/08/14 13:21:21 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class StorableObject
{
public:
  /** @name Objekte anlegen und zerst\"oren */
  //@{
    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
    darauf zur\"uck.

    Copy ist eine "pure virtual function" (daher das seltsame = 0 im Code), die 
    also von allen Nachkommen ueberschrieben werden muss. Damit sie richtig
    funktioniert, mu\3 sie von JEDER abgeleiteten Klasse neu \"uberschrieben
    werden. Hier ein Beispiel f\"ur die Klasse Foo:

    \begin{verbatim}
    Foo.h:
    class Foo : public StorableObject {
      public:
        ...
        virtual StorableObject* copy() const;
        ...
    }
    Foo.c:
    StorableObject* Foo::copy() const {
      return new Foo( *this );
    }

   \end{verbatim} 
   @param Ein Zeiger vom Typ (StorableObject *) auf den neu allozierten
      Speicher. */
    virtual StorableObject* copy() const = 0;
   
    /** Der virtuelle Destruktor. */
    virtual ~StorableObject();
  //@}
   

  /** @name Ausgabe auf dem Terminal */
  //@{ 

    /** Ausgabe der Beschreibung auf einen Stream. 
      Standardm\"a\3ig wird auf stdout ausgegeben. Der auszugebende String wird
      von der Methode \Ref{toString}() erzeugt. 
      
      Am einfachsten kann man das Objekt durch Umleitung nach cout ausgeben. */
    void printOn (ostream& = cout) const;
  
    /** Beschreibung als String fuer die Ausgabe auf einem Terminal. 
      Der String sollte kein abschlie\3sendes newline enthalten. Die Ausgabe
      sollte in () eingschlossen sein und als erstes den Klassennamen 
      enthalten. 

      Die Methode mu\3 in der abgeleiteten Klasse \"uberschrieben werden.*/
    virtual String toString() const = 0;
  //@}

};


/** Umleitungsoperator fuer die Ausgabe von StorableObjects. Alle von
StorableObject abgeleiteten Klassen k\"onnen mit einem Kommando wie
\begin{verbatim}
cout << object << endl;
\end{verabtim}
auf dem Terminal ausgegeben werden. Der String kann durch \"Uberschreiben
der Methode \Ref{StorableObject::toString}() beliebig festgelegt werden. */ 
inline ostream& operator<< (ostream& aStream, const StorableObject& object){
  object.printOn( aStream );
  return aStream;
}

#endif
