#ifndef _TIMER_H_
#define _TIMER_H_

#include "Effectors.h"
#include "Sensors.h"
#include "Deliberator.h"
#include "Exception.h"

/**
  Diese Klasse koordiniert das Timing zwischen den Komponenten
  Effectors, Sensors und Deliberator. Sie kann die gegenwaertige 
  Abarbeitung unterbrechen und die Abarbeitung an eine der drei
  Komponenten uebergeben. Nach Abarbeitung dieser Komponenten wird
  dann die unterbrochene Routine fortgesetzt.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Timer.h} | [\URL[Implementierung]{src/Timer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Timer.h,v $
  *   $Revision: 1.17 $
  *   $Date: 1998/08/14 13:21:27 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Timer
{
public:

  /** @name Fehlerklassen */
  //@{
  
    /** Generelle Timerfehler-Klasse. */
    class TimerException: public Exception {
    public:
      TimerException(const String& aDescription )
        : Exception( aDescription ) {}
    };
  //@}
  
  /** @name Timer erzeugen und entfernen */
  //@{

    /** Erzeugt einen neuen Timer. 
      Dieser ist noch nicht aktiv.Er wird erst aktiv, wenn er mit run() 
      gestartet wird.
      @param sensors Zeiger auf das Sensorik-Objekt.
      @param effectors Zeiger auf das Effector-Objekt.
      @param deliberator Zeiger auf das Deliberator-Objekt. */
    Timer( Sensors *sensors, Effectors *effectors, Deliberator *deliberator );

    /** Beendet den Timer. 
      Dieser Aufruf stoppt erst den Timer und beendet ihn dann ganz. */
    ~Timer();
  
  //@}
    
  /** @name Normaler Betrieb */
  //@{
    
    /** Aktiviert den Timer. 
      Die Methode schedule wird als Schedulemethode aktiviert. 
      Diese wird dann alle 12.5 ms aktiviert. 
      @exception \Ref{Timer::TimerException}, wenn ein interner Signalfehler
       auftritt. */
    void run();

    /** Uebernimmt das Scheduling zwischen den Sensors-,
      den Effectors- und dem Deliberatorobjekt. 
      Diese Methode wird nach Aufruf von \Ref{run}() alle 12.5 ms durch das
      UNIX-Signal SIGALRM  aktiviert. */
    void schedule();

    /** 
     */
    void mainLoop();

    /** Stopt den Timer. 
      Der Timer wird wieder abgestellt. */
    void stop();
  
    /** Gibt die aktuelle Timer-Zeit in Millisekunden zurueck.
      Nach Aufruf von run() wird mit Zaehlen begonnen. Beim Anstoss des Spieles
      wird die Timer-Zeit wieder auf 0 zurueckgesetzt. In der zweiten Halbzeit
      passt sie sich selbst\"andig an die Zeit des Servers an. Ansonsten laeuft
      sie autonom weiter. */
    unsigned long time() const;

    /** Liefert den aktuellen Zeittakt des SoccerServers */
    int soccerServerSimStep() const;
        
  //@}

  
  /** @name Bereiche vor Unterbrechung sch\"utzen */
  //@{
    /** Es wird ein geschuetzter Bereich gestartet. 
      Alle Befehle bis zum naechsten \Ref{unlock}() werden ohne Unterbrechung
      ausgefuehrt. Der  Signalhandler wird fuer diesen Zeitraum abgestellt. 
      @exception \Ref{Timer::TimerException}, wenn ein Lock versucht wurde,
        obwohl bereits ein Lock gesetzt ist. */
    void lock();
  
  /** Ein geschuetzter Bereich wird beendet. 
    Danach kann die aufrufende Routine wieder vom Timer unterbrochen werden. 
      @exception \Ref{Timer::TimerException}, wenn versucht wurde, einen Lock 
        zu entfernen, obwohl gar kein Lock gesetzt war. */
  void unlock();
  
  //@}


private:

  /** Zeit, die seit senden der letzten SubAction vergangen ist. */
  long timeSinceLastSentSubAction() const;

  /** Zeit, die seit senden der letzten MainAction vergangen ist. */
  long timeSinceLastSentMainAction() const;

  /** Hier wird die Zeit in x ms-Schritten gezaehlt. */
  long theTimeCounter;

  /** Zeitschritte, die seit senden der letzten MainAction vergangen ist. */
  long theTimeSinceLastSentMainAction;

  /** Zeitschritte, die seit senden der letzten SubAction vergangen ist. */
  long theTimeSinceLastSentSubAction;

  /** Zeittakt des SoccerServers. */
  long theSoccerServerSimStep;
  
  /** Zeittakt des SoccerServers zu dem zuletzt eine MainAction gesendet
    wurde. */
  long theLastSentMainActionSoccerServerTime;

  /** Variable, die ansagt, dass der gegenwaertige Prozess nicht unterbrochen
    werden soll. Wird von \Ref{lock}() und \Ref{unlock}() veraendert. */
  bool isLocked;

  /** Indiz f\"ur den Scheduler, ob die Sensoren auf neue Informationen 
    \"uberpr\"uft werden sollen. */
  bool haveToSee;

  /** Indiz f\"ur die Effectors , ob sie eine neue MainAction senden sollen. */
  bool haveToSendMainAction;

  /** Indiz f\"ur die Effectors , ob sie eine neue SubAction senden sollen. */
  bool haveToSendSubAction;

  /** Indiz f\"ur den Scheduler, ob der Deliberator aktiviert werden soll. */
  bool haveToDecideNew;

  /** Interner Zeittakt des Timers in Nanno-Sekunden. */
  int theInternTimeStep;

  /** Aller wieviel Schritte ein Sub-Kommando gesendet wird. */
  int theSentSubStep;

  /** Aller wieviel Schritte ein Main-Kommando gesendet wird. */
  int theSentMainStep;

  /** Zeiger auf das Sensorik-Objekt */
  Sensors *theSensors;
  
  /** Zeiger auf das Effector-Objekt */
  Effectors *theEffectors;

  /** Zeiger auf das Deliberator-Objekt */
  Deliberator *theDeliberator;
 
};

inline int Timer::soccerServerSimStep() const
{
  return theSoccerServerSimStep;
}
/** Globaler Zeiger auf das aktive Timer-Objekt. */
extern Timer *globalTimer;


void globalSchedule( int signalNumber );

#endif

