#ifndef _COMMUNICATOR_H_
#define _COMMUNICATOR_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <unistd.h>

#include "Common.h"
#include "String.h"
#include "Exception.h"
#include "Configuration.h"

/**
  Der UDP/IP-level Kommunikationsverwalter. 
  Der Communicator sorgt f"ur die Kommunikation mit dem SoccerServer auf der
  UDP/IP Ebene. Er schickt Strings an den Server und liest blockierend 
  Strings vom Server.

  Von der Klasse wird nur ein Objekt erzeugt. Die Klasse ist nicht zur
  Vererbung geeignet.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Communicator.h} | [\URL[Implementierung]{src/Communicator.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Communicator.h,v $
  *   $Revision: 1.21 $
  *   $Date: 1998/08/14 13:20:03 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Communicator
{
public:

  /** Interne Fehlerklasse. */
  class CommunicatorException: public Exception {
  public:
    CommunicatorException(const String& aDescription )
      : Exception( aDescription ) {}
  };

  /** Exception die ansagt, da"s keine Nachricht erhalten wurde. 
   */
  class NothingReceived: public Exception {
  public:
    NothingReceived() : Exception() {}
  };  
  
  /** Verbindungsaufbau.
    Aufbau einer UDP/IP Verbindung.
    @param hostName Der Name des Rechners, auf dem der SoccerServer laeuft
    @param port Der Port, an dem der SoccerServer auf Spieler wartet. 
    @exception CommunicatorException Noch nicht genauer spezifiziert. */
  Communicator(const String& hostName, int port );
  
  /** Verbindungsabbau.
    Die Verbindung zum SoccerServer wird beendet. */
  ~Communicator();
  
  /** Schicken einer Nachricht an den Server.
    Die Nachricht wird sofort per UDP an den Server geschickt.
    @param theMessage String, der die Nachricht enthaelt
    @exception CommunicatorException String konnte nicht abgesendet werden */
  void sendMessage(const String& theMessage);
  
  /** Liefert die naechste Nachricht, die vom Server kommt.
    Ggf. wird gewartet, bis eine Nachricht eintrifft.
    @param milliSeconds Maximale Wartezeit. Default ist WAIT_FOREVER	
    @exception NothingReceived Keine Nachricht eingetroffen. */
  String getNextMessage(int milliSeconds = WAIT_FOREVER);

  /** Liefert die zuletzt erhaltene Nachricht, die vom Server kamm.
    Ggf. wird gewartet, bis eine Nachricht eintrifft.
    @exception CommunicatorException Keine Nachricht eingetroffen. */
  String getLastMessage();  
  
private:
  int ownSocket;
  struct sockaddr_in hostAddress;
  int lengthOfHostAddress;
  bool communicationAvailable;
  
  void openConnection(const String& hostName, int port);
  void closeConnection();
};

#endif
