#include "Effectors.h"
#include "ModelledBall.h"
#include "WorldModel.h"
#include "Log.h"
#include "Array.h"
#include "Timer.h"

/**
 * Aktueller Stand
 *
 *   $RCSfile: Effectors.cc,v $
 *   $Revision: 1.36 $
 *   $Date: 1998/08/14 13:20:26 $
 *   $Author: myritz $ 
 *
 */

Effectors::Effectors( Communicator* aCommunicator )
  : theCommunicator( aCommunicator ) {
    hasEnteredField = NO;
    theQueue = Array<Action>(20);
    indexOfCurrentAction = 0;
 }

void Effectors::enterField( const String& aTeamName, bool isGoalie ) {
  String message;
  String answer;
  
  if ( hasEnteredField ) {
    WARNING( Log::EFFECTORS, "Field already entered. Ignoring command." );
    return;
  }
  
  if(isGoalie) {
    message = "(init " + aTeamName + " (version 4.00) (goalie))";
  }
  else {
    message = "(init " + aTeamName + " (version 4.00))";
  }
  
  theCommunicator->sendMessage( message );
  try {
    answer = theCommunicator->getNextMessage( 
        globalConf->communicationTimeout );
  }
  catch( Communicator::NothingReceived ) {
    throwException( EffectorsException( "Server didn't respond" ) );
  }
  if( answer.beginsWith( "(error ") ) {
    throwException( 
        EffectorsException( answer ) );
  }

  char internalSide;
  int playerNumber;	
  sscanf (answer.cString(), "(init %c %i ", &internalSide, &playerNumber);
  
  globalWorldModel->setPlayerNumber( playerNumber );
  globalWorldModel->setTeamName( aTeamName );
  globalWorldModel->setSide( (internalSide == 'l') ? 
			     LEFT_SIDE_ID : RIGHT_SIDE_ID );
  globalWorldModel->setSimStep( 0 );
  globalWorldModel->setGameState( WorldModel::BEFORE_KICK_OFF );
  globalWorldModel->setInitiative( LEFT_SIDE_ID );
  if ( isGoalie == YES ) {
    if ( globalWorldModel->role() != GOALIE_ROLE ) {
      throwException( EffectorsException( 
          "Role of Goalie is not GOALIE_ROLE" ) );
    }
  } 
  else {
    if ( globalWorldModel->role() == GOALIE_ROLE ) {
      globalWorldModel->setRole( ASSISTANT_GOALIE_ROLE );
      throwException( EffectorsException( 
          "Goalie didn't connect as goalie." ) );
      //WARNING( Log::EFFECTORS, String( "Goalie didn't connect as goalie." ) +
      //    "Setting role to ASSISTANT_GOALIE_ROLE.\n" );
    }
  }
  hasEnteredField = YES;
}

void Effectors::reEnterField( const String& aTeamName, int aPlayerNumber ) {
  String message;
  String answer;
  
  if ( hasEnteredField ) {
    WARNING( Log::EFFECTORS, "Field already entered. Ignoring command." );
    return;
  }
  
  message = "(reconnect " + aTeamName + " " + aPlayerNumber + ")";
  
  theCommunicator->sendMessage( message );
  try {
    answer = theCommunicator->getNextMessage( 5 );
  }
  catch( Communicator::NothingReceived ) {
    throwException( EffectorsException( "Server didn't respond" ) );
  }
  if( answer.beginsWith( "(error" ) ) {
    throwException( EffectorsException( "Server refused reconnect." ) );
  }
  
  char internalSide;
  sscanf (answer.cString(), "(reconnect %c ", &internalSide);

  globalWorldModel->setPlayerNumber( aPlayerNumber );
  globalWorldModel->setTeamName( aTeamName );
  globalWorldModel->setSide( (internalSide == 'l') ? 
			     LEFT_SIDE_ID : RIGHT_SIDE_ID );
  globalWorldModel->setSimStep( globalConf->halfTime );
  globalWorldModel->setGameState( WorldModel::BEFORE_KICK_OFF );
  globalWorldModel->setInitiative( RIGHT_SIDE_ID );
  hasEnteredField = YES;
}

void Effectors::leaveField() {
  if ( !hasEnteredField ) {
    WARNING( Log::EFFECTORS, "Not on field. Ignoring command." );
    return;
  }
  
  theCommunicator->sendMessage( "(bye)" );
  hasEnteredField = NO;
}

void Effectors::queueAction( const Action& anAction ) {
  theQueue.addObject( anAction );
}

void Effectors::queuePlan( const Plan& aPlan ) {
  theQueue.addObjects( aPlan.actions() );
}	

void Effectors::clearQueue() {
  theQueue.removeAll();
  indexOfCurrentAction = 0;
}

void Effectors::setPlan( const Plan& aPlan ) {
  clearQueue();
  queuePlan( aPlan );
}

void Effectors::performAction( const Action& anAction ) {
  String command;
  
  if ( anAction.isValid() == NO ) {
    throwException( EffectorsException( 
        "You tried to let me do the invalid action " + anAction.toString() ) );
  }
  
  command = anAction.asCommandString();
  if ( command != "" ) {
    try {
      if ( anAction.simStep() != globalTimer->soccerServerSimStep() &&
           anAction.isMainAction() ) {
        WARNING( Log::EFFECTORS, "Sending an action that was calculated for " +
            String( anAction.simStep() ) + " at simStep " + 
            String(  globalTimer->soccerServerSimStep() ) + "\n" );
        Action modifiedAction( anAction );
        modifiedAction.setSimStep( globalTimer->soccerServerSimStep() );
        globalWorldModel->registerExecution( modifiedAction );
      }
      else {
        globalWorldModel->registerExecution( anAction );
      }
      theCommunicator->sendMessage( command );
    }
    catch ( ModelledBall::KickImpossibleException message ) {
      WARNING( Log::EFFECTORS, 
               String( "Kick is illegal and will not be sent." ) +
               " ( Caught exception " + message.toString() + " )." );
    }
  }
}

void Effectors::performMainAction( const BodyInfos& aBodyInfo ) {
  INFO( Log::EFFECTORS, 4, "ENTER: Effectors::performMainAction\n" );
  INFO( Log::EFFECTORS, 13, "Starting with Plan: " + toString() );
  Action nextAction = Action( EMPTY_ACTION, globalWorldModel->simStep() );
  int numberOfActions = theQueue.count();
  Plan modifiedPlan;
  int oldIndexOfCurrentAction = indexOfCurrentAction;

  while( indexOfCurrentAction < numberOfActions ) {
    nextAction = theQueue.objectAt( indexOfCurrentAction );
    if(nextAction.isMainAction()) {
      indexOfCurrentAction++;
      break;
    }
    modifiedPlan.addAction( nextAction );
    indexOfCurrentAction++;
  }
  if( nextAction.isMainAction() ) {
    if( modifiedPlan.length() > 0 ) { // neuen Plan zusammenbauen
      INFO( Log::EFFECTORS, 15, "Modify plan.\n" );
      while( indexOfCurrentAction < numberOfActions ) {
        modifiedPlan.addAction( theQueue.objectAt( indexOfCurrentAction ) );
        indexOfCurrentAction++;
      }
      setPlan( modifiedPlan );
    }
    if( nextAction.id() != EMPTY_ACTION ) {
      INFO( Log::EFFECTORS, 15, "Send mainAction.\n" );
      switch( nextAction.id() ) {
        case DASH_ACTION:
          nextAction.setActionCount( aBodyInfo.dashs() );
          break;
        case TURN_ACTION:
          nextAction.setActionCount( aBodyInfo.turns() );
          break;
        case KICK_ACTION:
          nextAction.setActionCount( aBodyInfo.kicks() );
          break;
      }
      performAction( nextAction );
    }
    else {
      globalWorldModel->registerExecution( nextAction );
      INFO( Log::EFFECTORS, 15, "Send nothing.\n" );
      INFO( Log::EFFECTORS, 10, "Remainder of Plan: " + toString() );
      INFO( Log::EFFECTORS, 4, 
            "LEAVE: Effectors::performMainAction with Exception\n" ); 
      throwException( NothingToSendException() );
    }    
  }
  else { // keine MainActions mehr im Plan -> also n"achste SubAction senden.
    WARNING( Log::EFFECTORS,
             String( "No main Action left at performMainAction, " ) +
             "send first subAction instead\n" );
    indexOfCurrentAction = oldIndexOfCurrentAction;
    performSubAction();
  }
  INFO( Log::EFFECTORS, 10, "Remainder of Plan: " + toString() );
  INFO( Log::EFFECTORS, 4, "LEAVE: Effectors::performMainAction\n" ); 
}

void Effectors::performSubAction() {
  INFO( Log::EFFECTORS, 4, "ENTER: Effectors::performSubAction\n" );
  INFO( Log::EFFECTORS, 13, "Starting with Plan: " + toString() ); 
  Action nextAction;
  int numberOfActions = theQueue.count();  

  if( indexOfCurrentAction < numberOfActions ) {
    nextAction = theQueue.objectAt( indexOfCurrentAction );
    if( !(nextAction.isMainAction()) ) {
      indexOfCurrentAction++;
      performAction( nextAction );
    }
    else {
      INFO( Log::EFFECTORS, 4, 
            "LEAVE: Effectors::performSubAction with Exception\n" ); 
      throwException( NothingToSendException() );
    }
  }
  else {
    INFO( Log::EFFECTORS, 4, 
          "LEAVE: Effectors::performSubAction with Exception\n" ); 
    throwException( NothingToSendException() );
  }
  INFO( Log::EFFECTORS, 10, "Remainder of Plan: " + toString() );
  INFO( Log::EFFECTORS, 4, "LEAVE: Effectors::performSubAction\n" ); 
}

bool Effectors::isActionLeft()
{
  if ( indexOfCurrentAction < theQueue.count() ) {
    return YES;
  }
  else {
    return NO;
  }
}

String Effectors::toString() const
{
  int numberOfActions = theQueue.count(); 

  String output = "( Effectors actions: ";
  for( int i = indexOfCurrentAction; i < numberOfActions; i++ ) {
    output = output + "\n" + String( i ) + ".: " + 
      theQueue.objectAt( i ).toString();
  }
  output = output + " )\n";
  return output;
}

