
#include "ModelledBall.h"
#include "ModelledPlayer.h"
#include "Log.h"
#include <math.h>

/**
 * Aktueller Stand
 *
 *   $RCSfile: ModelledBall.cc,v $
 *   $Revision: 1.32 $
 *   $Date: 1998/08/14 13:20:50 $
 *   $Author: myritz $ 
 *
 */

double ModelledBall::theBallSize;
double ModelledBall::theBallDecay;
double ModelledBall::theBallMaxSpeed;
double ModelledBall::theKickPowerRate;
double ModelledBall::theMaxPower;
double ModelledBall::theMinPower;
double ModelledBall::theMaxMoment;
double ModelledBall::theMinMoment;
double ModelledBall::theKickMargin;
short ModelledBall::isSetOnce = NO;


ModelledBall::ModelledBall( int id,
  const Vector& aPos , 
  const Vector& aSpeed,
  const Reliability& aPosReliability,
  const Reliability& aSpeedReliability,
  int aLastTouchedBy,
  int aLastTouchedAtTime ) 
  : ModelledObject( id, aPos, aSpeed, aPosReliability, aSpeedReliability),
    theLastTouchedBy( aLastTouchedBy ),  
    theLastTouchedAtTime( aLastTouchedAtTime )
{
  if (isSetOnce==NO){
    setClassVars();
  }
}

ModelledBall::ModelledBall( 
  const MovingObject& aMovingObject,  
	const Reliability& aPosReliability,
	const Reliability& aSpeedReliability,
	int aLastTouchedBy,
	int aLastTouchedAtTime )
  : ModelledObject()
{
  if (isSetOnce==NO){
    setClassVars();
  }
  setId( aMovingObject.id() );
  setPos( aMovingObject.pos() );
  setSpeed( aMovingObject.speed() );
  setPosReliability( aPosReliability );
  setSpeedReliability( aSpeedReliability );
  theLastTouchedBy = aLastTouchedBy;
  theLastTouchedAtTime = aLastTouchedAtTime;
}


void ModelledBall::setClassVars() 
{
  setSize(globalConf->ballSize);
  setDecay(globalConf->ballDecay);
  setMaxSpeed(globalConf->ballSpeedMax);
  setKickPowerRate(globalConf->kickPowerRate);
  setMaxPower(globalConf->maxPower);
  setMinPower(globalConf->minPower);
  setMaxMoment(radFromDegree(globalConf->maxMoment));
  setMinMoment(radFromDegree(globalConf->minMoment));
  setKickMargin(globalConf->kickMargin);
  isSetOnce=YES;
}

void ModelledBall::simulateSteps( unsigned int simSteps = 1 ) 
{
  if( simSteps > 100 ) {
    WARNING( Log::MODELLED_OBJECTS, 
        "ModelledBall::simulateSteps: more than 100 steps." );
  }
  if( posReliability().get() < Reliability::RANDOM &&
      speedReliability().get() < Reliability::RANDOM ) {
    while( simSteps > 0 ) {
      setPos( pos() + speed() );
      setSpeed( decay() * speed() );
   
    reduceReliability();
    simSteps--;
    }
  }
}

void ModelledBall::accelerate( double power, double angle, 
    const ModelledPlayer& byPlayer )
{
  Vector myAcceleration;

  power = effectiveKickPower( power, byPlayer );
  angle = normalizeKickAngle( angle );

  if( speedReliability().get() < Reliability::RANDOM ) {

    myAcceleration.setPolar( power, angle + byPlayer.faceDir() );

    setSpeed( speed() + (kickPowerRate() * myAcceleration) );
    if( speed().length() > maxSpeed() ) {
      Vector mySpeed = speed();
      mySpeed.setPolar( maxSpeed(), mySpeed.angle() );
      setSpeed(mySpeed);
    }
  }
  else {
    throwException(
        ModelledBallException( "ModelledBall has no valid speed value" ) );
  }
}

bool ModelledBall::isInKickRangeOfPlayer( const ModelledPlayer& aPlayer ) const
{
  double distFromPlayer = (pos() - aPlayer.pos()).length();
  return ( distFromPlayer <= globalConf->kickRange );
}

double ModelledBall::effectiveKickPower( double power,
    const ModelledPlayer& kickingPlayer ) const 
{
  double theEffectivePower;
  double angleFromPlayer = kickingPlayer.angleTo( pos() );
  double distFromPlayer = ( pos() - kickingPlayer.pos() ).length();
  distFromPlayer -= size() + kickingPlayer.size();
  if( distFromPlayer > kickMargin() + DEFAULT_COMPARISON_ACCURACY ) {
    throwException( 
        KickImpossibleException( "player can't kick ball (out of range)" ) );
  }  

  power = normalizeKickPower(power);
  theEffectivePower = power * kickPowerMult( angleFromPlayer,
					     distFromPlayer );

  return theEffectivePower;
}

double ModelledBall::kickPowerForSend( double power,
    const ModelledPlayer& kickingPlayer ) const 
{
  double thePowerForSend;
  double angleFromPlayer = kickingPlayer.angleTo( pos() );
  double distFromPlayer = (pos() - kickingPlayer.pos()).length();
  distFromPlayer -= size() + kickingPlayer.size();
  if( distFromPlayer > kickMargin() + DEFAULT_COMPARISON_ACCURACY ) {
    throwException( 
        KickImpossibleException( "player can't kick ball (out of range)" ) );
  }

  power = normalizeKickPower( power );
  power = power / kickPowerMult( angleFromPlayer, distFromPlayer );
  thePowerForSend = normalizeKickPower(power);
  if( !isAlmostEqual( thePowerForSend, power ) ) {
    WARNING( Log::MODELLED_OBJECTS, 
             "kickPowerForSend: power value may be out of range" );
  }

  return thePowerForSend;
}

double ModelledBall::maxEffectiveKickPower( 
    const ModelledPlayer& kickingPlayer) const 
{
  return effectiveKickPower( maxPower(), kickingPlayer );
}


double ModelledBall::kickPowerMult( double angleFromPlayerToBall,
    double distFromPlayerToBall ) const 
{
  if( distFromPlayerToBall + globalComparisonAccuracy < 0  ) {
    distFromPlayerToBall = 0;
    WARNING( Log::MODELLED_OBJECTS, 
             "kickPowerMult: distFromPlayerToBall is less then 0" );
  }
  return( 1 - 
	  0.25 * (fabs(angleFromPlayerToBall) / PI) -
	  0.25 * (distFromPlayerToBall / kickMargin() ) );
}

StorableObject* ModelledBall::copy() const 
{
  return new ModelledBall( *this );
}


String ModelledBall::toString() const 
{
  String output;
 
  output = String( "( ID: " ) + String( id() ) + 
    String(", Pos: ") + pos().toString() + 
    String( ", Pos" ) + posReliability().toString() +
    String( ", Speed: " ) + speed().toString() + 
    String( ", Speed" ) + speedReliability().toString() +
    String( ", LastTouchedBy: " ) + String( (long)theLastTouchedBy ) +
    String( ", LastTouchedAtTime: " ) + String( (long)theLastTouchedAtTime ) +
    String( " )" );
  return output;
}







