#ifndef _OPTION_H_
#define _OPTION_H_

#include "Common.h"
#include "Constraint.h"
#include "Array.h"
#include "Plan.h"

const int MIN_OPTION_ID = 700;
const int UNDEF_OPTION_ID = 700;

/** 
  Basisklasse fuer alle umsetzbaren Handlungsmoeglichkeiten des Agenten. 
  Jedes Objekt dieser
  Klasse steht fuer eine Option, die eine konkrete Aktionsfolge zur Umsetzung
  erzeugen kann. Ebenso wie bei Constraints haengt die Utilityberechnung von
  Wuenschen von
  der aktuellen Situation ab. Zusaetzlich werden aber einschraenkende Wuensche
  (Nebenbedingungen) beachtet. Nebenbedingungen muessen von Constraint
  abstammen.

  Jedesmal, nachdem sich die aktuelle Situation im Weltmodell oder die Liste
  der zu beruecksichtigenden Nebenbedingungen geaendert hat, wird die 
  expectedUtility (der Nutzen) neu berechnet.

  Auch von Option selbst koennen Instanzen erzeugt werden. Sie werden zur
  Initialisierung im Deliberator und als letzter Fallback bei der
  Intentionerzeugung verwendet. Normalerweise werden aber nur von Nachkommen
  von
  Option Instanzen erzeugt.

  Mit den uebergebenen Arrays von Zeigern auf Constraints wird keine
  Verantwortung fuer Constraint-Objekte uebergeben. Das die Option besitzende
  Objekte muss garantieren, dass Zeiger waerend der Lebensdauer der Option
  gueltig bleiben.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Option.h} | [\URL[Implementierung]{src/Option.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Option.h,v $
  *   $Revision: 1.13 $
  *   $Date: 1998/08/14 13:21:02 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Option : public Constraint
{
public:
  /** Erzeugt eine UNDEF_OPTION. */
  Option();
  
  /** Fordert, dass der Nutzen unter Beruecksichtigung der angegebenen 
    Nebenbedingungen berechnet wird. 
    Das heisst NICHT, dass die Nebenbedingungen beachtet werden, sondern nur,
    dass die durch ihre Verletzung entstehenden Nachteile in die Berechnung des
    Nutzens einfliessen.
    @param constraints Container mit den Nebenbedingungen */
  void setConstraints( const Array<Constraint *>& constraints );
  
  /** Gibt die Abbruchkosten an. 
    Das sind Kosten, die entstehen, wenn bereits  mit der Umsetzung begonnen
    wurde und nun eine andere Option gewaehlt wird. Hohe Abbruchkosten
    entstehen vor allem, wenn eine gemeinsame Absicht des  Teams aufgegeben
    wird.
    @return Im Allgemeinen ein Wert zwischen NEUTRAL_COST und MAX_COST, bei
    Option selbst NEUTRAL_COST. */
  virtual int cancellationCost() const;
  
  /** Prueft, ob die Option tatsaechlich umsetzbar ist.  
    Es kann sein, dass sie nicht umsetzbar ist, da zur Berechnung des Nutzens
    Heuristiken  verwendet werden. Moeglicherweise laesst sich also kein Plan
    zur Umsetzung der Option erzeugen.

    Die Berechnung von isPossible ist deutlich aufwendiger als die Berechnung
    der expectedUtility. Wurde die Utility noch nicht berechnet, muss sie
    zunaechst berechnet werden, um festzulegen, welche Constraints
    beruecksichtigt werden muessen. Danach wird ein Planskelett erzeugt (wenn
    nicht bereits eines existiert), das dann garantiert umsetzbar ist.

    Das Ergebnis wird gecacht, so dass ein wiederholter Aufruf der Funktion in
    derselben Situation schnell zu demselben Ergebnis fuehrt.
    @return YES oder NO. */
  virtual bool isPossible() const;
  
  /** Soll die Absicht unbedingt beibehalten werden?
    Manche Optionen koennen oder sollen nicht jederzeit abgebrochen werden.
    Diese Methode sollte wird vom Deliberator vor jedem Entscheidungsvorgang
    gefragt, ob die aktuelle Absicht weitergemacht werden soll. Beim
    \"Uberschreiben in Subklassen sollte man Vorsicht walten lassen.
    Standardm\"a\3ig liegert die Methode NO zur\"uck.
      @return YES oder NO. */
  virtual bool forceContinuation() const;

  /** Anpassung an die bisherige Absicht. 
    Die Anpassung kann die tatsaechliche Utility des Agenten verbessern, da es
    so aussieht, als wuerde er den bisherigen Plan weiterverfolgen, obwohl
    inzwischen sein Wunsch ein neuer ist.
    Die Default-Methode nimmt keine Anpassungen vor. */
  virtual void adaptToPreviousIntention( const Option *previousIntention );

  /** Erzeugt einen Plan zur Umsetzung der Option. 
    Wurde die Umsetzung des Plans bereits teilweise erledigt, so stellt dies
    generatePlan() fest und erzeugt die naechsten Schritte ausgehend von der
    aktuellen Situation im Weltmodell.  
    @return Plan, der die ersten Schritte zur Umsetzung der Absicht enthaelt.
    Die Zahl der Schritte laesst sich konfigurieren (Variable maxLengthOfPlan
    in agent.conf). */
  virtual Plan generatePlan() const;

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. 
      Diese Methode muss fuer JEDE Unterklasse neu definiert  werden. */
  virtual StorableObject* copy() const;

  /** Virtueller Destruktor */
  virtual ~Option();

protected:
  // Variablen
  
  /** Hier stehen die ueber setConstraints angeforderten Nebenbedingungen. */
  Array<Constraint *> requestedConstraints;
  
  /** Hier stehen die bei der Nutzenberechnung akzeptierten Nebenbedingungen. */
  mutable Array<Constraint *> acceptedConstraints;
  
  
  // Methoden
  
  /** Fuehrt die eigentliche Berechnung der expectedUtility durch. 
    Der Mechanismus ist wesentlich aufwendiger als bei reinen Constraints, da
    die Nebenbedingungen mitberuecksichtigt werden muessen. Diese Methode ist
    const, obwohl sie interne Werte umsetzt (acceptedConstraints). */
  int computeExpectedUtility() const;
  
  /** Der Grundnutzen der Option. 
    Dieser Teil der Nutzenberechnung ist nur  vom Weltmodell abhaengig.
    @return Im Allgemeinen ein Wert zwischen MAX_UTILITY und MIN_UTILITY.
    Option selbst liefert MIN_UTILITY. */
  virtual int baseUtility() const;
  
  /** Liefert die Kosten, die durch die Einhaltung einer Nebenbedingung   
    entstehen. 
    Es gibt drei Arten von Nebenbedingungen.  (1) Solche, die die zugehoerige
    Absicht einbeziehen kann, (2) neutrale, die ohne weiteren Aufwand nebenbei
    miterfuellt werden, und (3) andere, die entweder spaeter hinzugekommen oder
    unerfuellbar sind.

    Die Kosten fuer (1) sind abhaengig von der Option und vom Weltmodell.

    Nebenbedingungen vom Typ (2) haben die Kosten NEUTRAL_COST.

    Die Kosten der Einhaltung einer unbekannten Nebenbedingung sind MAX_COST.
    Dadurch versucht die Option nie, sie zu erfuellen.
    @param constraint Die einzuhaltende Nebenbedingung
    @return Im allgemeinen Fall ein Wert zwischen NEUTRAL_COST und MAX_COST.
    Option selbst liefert fuer alle Constraints MAX_COST. */
  virtual int constraintCost( const Constraint *constraint ) const;
  
  /** Gibt an, ob eine Nebenbedingung unter den akzeptierten ist.
    Gibt YES zuru\"uck, wenn eines der Constraints in acceptedConstraints die
    als Parameter angegebene Id hat.
    @param constraintId eine ConstraintId
    @return YES oder NO */
  bool isConstraintAccepted( int constraintId ) const;
  
  /** Setzt, ob es m\"oglich ist diese Option durchzuf\"uhren. */
  void setIsOptionPossible( bool isOptionPossible ) const;

  /** Erfragt ob die Option prinzipiell moeglich ist.
    @return NO, wenn schon bei der Utility-berechnung die Unmoeglichkeit
    festgestellt wurde, sont YES */
  bool isOptionPossible() const;

private:

  /** Variabel die aussagt, ob die Option durchf\"uhrbar ist. */
  mutable bool theIsOptionPossible;

};

inline void Option::setConstraints( const Array<Constraint *>& constraints )
{
  requestedConstraints = constraints;
  simStepOfUtilityCalculation = UNDEF_INT;
}

inline Option::~Option()
{
}

inline bool Option::isOptionPossible() const
{
  return theIsOptionPossible;
}

inline void Option::setIsOptionPossible( bool isOptionPossible ) const
{
  theIsOptionPossible = isOptionPossible;
}  
#endif
