
#include <stdio.h>
#include <string.h>

#include "Sensors.h"
#include "WorldModel.h"
#include "String.h"
#include "Log.h"
#include "Timer.h"

Sensors::Sensors( Communicator* aCommunicator ) 
  : theCommunicator( aCommunicator ) {
    theSensedFlags = Array<StableObject>(FLAG_COUNT);
    theSensedLines = Array<StableObject>(LINE_COUNT);
    theSensedPlayers = Array<MovingPlayer>(22);
    theSensedBall = MovingObject();
    theSimStep = 0;
    theSimStepOfSoccerServer = 0;
    theHeardMessages = Array<Message>(5);
    theBodyInfos = Array<BodyInfos>(5);
    
    initSensors();
}

void Sensors::initSensors() {
  theIsNewSeeInfo = NO;
  theIsNewRefereeInfo = NO;
  theHeardMessages.removeAll();
  theBodyInfos.removeAll();
}

void Sensors::sense() {
  String message;
  static int timeOfLastMessage;
  
  try {
    while(1) {
      message = theCommunicator->getNextMessage(0);
      timeOfLastMessage = globalTimer->time();
      parseSensoryData(message);
    }
  }
  catch( Communicator::NothingReceived ) {
    int timeSinceLastMessage = globalTimer->time() - timeOfLastMessage;
    if ( timeSinceLastMessage > globalConf->communicationTimeout * 1000 ) {
      throwException( SensorsException( "Server has probably terminated." ) );
    }
  }
}

void Sensors::parseSensoryData(const String& message) {
  
  if( message.beginsWith( "(see" ) ) {
    theSensedFlags.removeAll();
    theSensedLines.removeAll();
    theSensedPlayers.removeAll();
    theSensedBall.setId(UNDEF_ID);
    char seeMessage[MAX_MESSAGE];

    sscanf( message.cString(), "(see %ld%[^\n]", &theSimStep, seeMessage);
    parseSights(seeMessage);
    theSimStepOfSoccerServer = theSimStep;
    theIsNewSeeInfo = YES;
  }
  else if( message.beginsWith( "(hear" ) ) {
    Message heardMessage = Message( message );
    theHeardMessages.addObject( heardMessage );
    theSimStepOfSoccerServer = heardMessage.time();
    if( heardMessage.senderInfo() == REFEREE_ID ) {
      theIsNewRefereeInfo = YES;
    }
  }
  else if( message.beginsWith( "(sense_body" ) ) {
    theLastBodyInfos = BodyInfos( message );
    theBodyInfos.addObject( theLastBodyInfos );
    theSimStepOfSoccerServer = theLastBodyInfos.time();
  }
  else if( message.beginsWith( "(error" ) ) {
    WARNING( Log::SENSORS, "Error message received." );
  }
  else {
    throwException( SensorsException("Unknown answer from the SoccerServer") );
  }

}

void Sensors::parseSights(char* seeMessage) {
  double dist, dir;
  double distChange, dirChange;
  double faceDir;
  char objectArray[256];
  char* object;
  char information[30];

  INFO( Log::SENSORS, 9, "parseSights: " + String(seeMessage) + "\n" );
  while( seeMessage[0] != ')' ) {
    object = objectArray;
    sscanf( seeMessage, " ((%[^)]) %[^)]) %[^\n]", object,
	    information, seeMessage);
    
    INFO( Log::SENSORS, 13, "parseSights: object: " + String(object) +
          " \tinformation: " + String(information) + "\n" );

    strcat( object, ")" );
    strcat( information, ")" );

    sscanf( information, "%lf %[^\n]", &dist, information );
    if( information[0] == ')' ) {
      // nur Richtung wurde erhalten, ignorieren
      continue;
    }    
    sscanf( information, "%lf %[^\n]", &dir, information );
    dir *= -1.0;   // ins mathemaische System umwandeln.

    distChange = UNDEF_DOUBLE;
    dirChange = UNDEF_DOUBLE;
    faceDir = UNDEF_DOUBLE;
    if( information[0] != ')' ) {
      sscanf( information, "%lf %[^\n]", &distChange, information );
      if( information[0] != ')' ) {
	sscanf( information, "%lf %[^\n]", &dirChange, information );
        dirChange *= -1.0;   // ins mathematische System umwandeln.
	if( information[0] != ')' ) {
	  sscanf( information, "%lf %[^\n]", &faceDir, information );
	  faceDir *= -1.0;   // ins mathematische System umwandeln.
	}
      }
    }
  
    if( strncasecmp (object, "ball", 4) == 0 ) { // ball oder Ball
      theSensedBall = MovingObject( BALL_ID, Vector(dist, dir, POLAR, DEGREE),
				    Vector(distChange, dirChange, POLAR,
					   DEGREE) );
    }
    else if( strncasecmp(object, "player", 6) == 0 ) { // player oder Player
      char team[256];
      int playerNumber;
      int playerId = UNDEF_PLAYER_ID;
      
      if( strncmp(object, "Player", 6) == 0 ) { // Player
        sscanf( object, "Player%[^\n]", object);
      }
      else { // player
        sscanf( object, "player%[^\n]", object);
      }

      if( object[0] != ')' ) {
	    sscanf( object, " %[^) \n\t\r\f]%[^\n]", team, object);
	if( globalWorldModel->teamName() == String(team) ) {
	  playerId = MIN_TEAMMATE_ID;
	}
	else {
	  playerId = MIN_OPPONENT_ID;
	}

	if( object[0] != ')' ) {
	  sscanf( object, " %i%[^\n]", &playerNumber, object);
	  playerId += playerNumber - 1;
	}
	else {
	  if( playerId == MIN_TEAMMATE_ID ) {
	    playerId = UNDEF_TEAMMATE_ID;
	  }
	  if( playerId == MIN_OPPONENT_ID ) {
	    playerId = UNDEF_OPPONENT_ID;
	  }
	}
      }
      if( faceDir != UNDEF_DOUBLE ) {
        faceDir = radFromDegree( faceDir );
      }
      theSensedPlayers.addObject( MovingPlayer( playerId,
          Vector(dist, dir, POLAR, DEGREE),
          Vector(distChange, dirChange, POLAR, DEGREE), faceDir ) );
    }
    else if( strncasecmp(object, "goal", 4) == 0 ) { // goal oder Goal
      if( strncmp(object, "goal r", 6) == 0) {
	theSensedFlags.addObject( 
            StableObject( RIGHT_GOAL_CENTER_FLAG_ID,
                          Vector(dist, dir, POLAR, DEGREE) ) );
      }
      else if( strncmp(object, "goal l", 6) == 0) {
	theSensedFlags.addObject( 
            StableObject( LEFT_GOAL_CENTER_FLAG_ID,
                          Vector(dist, dir, POLAR, DEGREE) ) );
      }
      else {
	// anonymes Tor, ignorieren.
	continue;
      }
    }
    else if( strncasecmp(object, "flag", 4) == 0 ) { // flag oder Flag
      if( strcmp(object, "Flag)") == 0 ) {
	// anonyme Flagge, ignorieren.
	continue;
      }
      
      int flagId = UNDEF_FLAG_ID;
      sscanf( object, "flag %[^\n]", object);
      
      switch( object[0] ) {
        case 'l':
          object++; object++;
          switch( object[0] ) {
            case '0':
              flagId = OUTSIDE_LEFT_CENTER_FLAG_ID;
              break;
            case 't':
              object++; object++;
              switch( object[0] ) {
                case '\0':
                  flagId = UPPER_LEFT_FLAG_ID;
                  break;
                case '1':
                  flagId = OUTSIDE_LEFT_TOP_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_LEFT_TOP_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_LEFT_TOP_30_FLAG_ID;
                  break;
              }
              break;
            case 'b':
              object++; object++;
              switch( object[0] ) {
                case '\0':
                  flagId = LOWER_LEFT_FLAG_ID;
                  break;
                case '1':
                  flagId = OUTSIDE_LEFT_BOTTOM_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_LEFT_BOTTOM_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_LEFT_BOTTOM_30_FLAG_ID;
                  break;
              }
              break;
          }
          break;
        case 'r':
          object++; object++;
          switch( object[0] ) {
            case '0':
              flagId = OUTSIDE_RIGHT_CENTER_FLAG_ID;
              break;
            case 't':
              object++; object++;
              switch( object[0] ) {
                case '\0':
                  flagId = UPPER_RIGHT_FLAG_ID;
                  break;
                case '1':
                  flagId = OUTSIDE_RIGHT_TOP_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_RIGHT_TOP_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_RIGHT_TOP_30_FLAG_ID;
                  break;
              }
              break;
            case 'b':
              object++; object++;
              switch( object[0] ) {
                case '\0':
                  flagId = LOWER_RIGHT_FLAG_ID;
                  break;
                case '1':
                  flagId = OUTSIDE_RIGHT_BOTTOM_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_RIGHT_BOTTOM_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_RIGHT_BOTTOM_30_FLAG_ID;
                  break;
              }
              break;
          }
          break;
        case 'c':
          object++; object++;
          switch( object[0] ) {
            case '\0':
              flagId = CENTER_FLAG_ID;
              break;
            case 't':
              flagId = UPPER_MIDDLE_FLAG_ID;
              break;
            case 'b':
              flagId = LOWER_MIDDLE_FLAG_ID;
              break;
          }
          break;
        case 'p':
          object++; object++;
          switch( object[0] ) {
            case 'l':
              object++; object++;
              switch( object[0] ) {
                case 't':
                  flagId = UPPER_LEFT_PENALTY_FLAG_ID;
                  break;
                case 'c':
                  flagId = MIDDLE_LEFT_PENALTY_FLAG_ID;
                  break;
                case 'b':
                  flagId = LOWER_LEFT_PENALTY_FLAG_ID;
                  break;
              }
              break;
            case 'r':
              object++; object++;
              switch( object[0] ) {
                case 't':
                  flagId = UPPER_RIGHT_PENALTY_FLAG_ID;
                  break;
                case 'c':
                  flagId = MIDDLE_RIGHT_PENALTY_FLAG_ID;
                  break;
                case 'b':
                  flagId = LOWER_RIGHT_PENALTY_FLAG_ID;
                  break;
              }
              break;
          }
          break;
        case 'g':
          object++; object++;
          switch( object[0] ) {
            case 'l':
              object++; object++;
              switch( object[0] ) {
                case 't':
                  flagId = LEFT_GOAL_TOP_FLAG_ID;
                  break;
                case 'b':
                  flagId = LEFT_GOAL_BOTTOM_FLAG_ID;
                  break;
              }
              break;
            case 'r':
              object++; object++;
              switch( object[0] ) {	
                case 't':
                  flagId = RIGHT_GOAL_TOP_FLAG_ID;
                  break;
                case 'b':
                  flagId = RIGHT_GOAL_BOTTOM_FLAG_ID;
                  break;
              }
              break;
          }
          break;
        case 't':
          object++; object++;
          switch( object[0] ) {
            case '0':
              flagId = OUTSIDE_TOP_CENTER_FLAG_ID;
              break;
            case 'l':
              object++; object++;
              switch( object[0] ) {
                case '1':
                  flagId = OUTSIDE_TOP_LEFT_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_TOP_LEFT_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_TOP_LEFT_30_FLAG_ID;
                  break;
                case '4':
                  flagId = OUTSIDE_TOP_LEFT_40_FLAG_ID;
                  break;
                case '5':
                  flagId = OUTSIDE_TOP_LEFT_50_FLAG_ID;
                  break;
              }
              break;
            case 'r' :
              object++; object++;
              switch( object[0] ) {
                case '1':
                  flagId = OUTSIDE_TOP_RIGHT_10_FLAG_ID;  
                  break;
                case '2':
                  flagId = OUTSIDE_TOP_RIGHT_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_TOP_RIGHT_30_FLAG_ID;
                  break;
                case '4':
                  flagId = OUTSIDE_TOP_RIGHT_40_FLAG_ID;
                  break;
                case '5':
                  flagId = OUTSIDE_TOP_RIGHT_50_FLAG_ID;
                  break;	    
              }
              break;
          }
          break;
        case 'b':
          object++; object++;
          switch( object[0] ) {
            case '0':
              flagId = OUTSIDE_BOTTOM_CENTER_FLAG_ID;
              break;
            case 'l':
              object++; object++;
              switch( object[0] ) {
                case '1':
                  flagId = OUTSIDE_BOTTOM_LEFT_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_BOTTOM_LEFT_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_BOTTOM_LEFT_30_FLAG_ID;
                  break;
                case '4':
                  flagId = OUTSIDE_BOTTOM_LEFT_40_FLAG_ID;
                  break;
                case '5':
                  flagId = OUTSIDE_BOTTOM_LEFT_50_FLAG_ID;
                  break;
              }
              break;
            case 'r' :
              object++; object++;
              switch( object[0] ) {
                case '1':
                  flagId = OUTSIDE_BOTTOM_RIGHT_10_FLAG_ID;
                  break;
                case '2':
                  flagId = OUTSIDE_BOTTOM_RIGHT_20_FLAG_ID;
                  break;
                case '3':
                  flagId = OUTSIDE_BOTTOM_RIGHT_30_FLAG_ID;
                  break;
                case '4':
                  flagId = OUTSIDE_BOTTOM_RIGHT_40_FLAG_ID;
                  break;
                case '5':
                  flagId = OUTSIDE_BOTTOM_RIGHT_50_FLAG_ID;
                  break;	    
              }
              break;
          }
          break;
      }
      
      if(flagId == UNDEF_FLAG_ID) {
        WARNING( Log::SENSORS, "Unknown flag received." );
      }
      else {
	theSensedFlags.addObject( 
            StableObject( flagId, Vector(dist, dir, POLAR, DEGREE) ) );
      }
    }
    else if( strncasecmp (object, "line", 4) == 0 ) {
      int lineId = UNDEF_LINE_ID;
      sscanf( object, "line %[^\n]", object);

      switch( object[0] ) {
      case 'l':
	lineId = LEFT_LINE_ID;
	break;
      case 'r':
	lineId = RIGHT_LINE_ID;
	break;
      case 't':
	lineId = UPPER_LINE_ID;
	break;
      case 'b':
	lineId = LOWER_LINE_ID;
	break;
      }

      if(lineId == UNDEF_LINE_ID) {
        WARNING( Log::SENSORS, "Unknown line received." );
      }
      else {
	theSensedLines.addObject( 
            StableObject( lineId, Vector(dist, dir, POLAR, DEGREE) ) );
      }
    }
    else {
      throwException( SensorsException("Unknown type of object.") );
    }
  }
}
