#ifndef _SITUATION_H_
#define _SITUATION_H_

#include "Common.h"
#include "StorableObject.h"
#include "ModelledBall.h"
#include "ModelledPlayer.h"
#include "Array.h"
#include "Exception.h"


/**
  Eine Spielsituation.  Eine Situation besteht aus bis zu 22 Spieler-Objekten
  (ModelledPlayer) und evtl. einem Ball-Objekt (ModelledBall). Eine Situation
  bezieht sich immer auf einen Zeittakt (simStep).

  Das Weltmodell erzeugt jeden SimStep eine neue Situation (per Simulation bzw.
  durch Auswerten von Sensorinformationen). Es verwendet Situationen, um den
  Ablauf des Spielgeschehens aufzuzeichnen.

  Eine Situation kann auch das Verstreichen von Zeittakten simulieren, wobei
  davon ausgegangen wird, dass keine Aktionen der Spieler mehr erfolgen. Dabei
  erhoeht sich auch der Zeittakt, auf den sich die Situation bezieht.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Situation.h} | [\URL[Implementierung]{src/Situation.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Situation.h,v $
  *   $Revision: 1.35 $
  *   $Date: 1998/08/14 13:21:14 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Situation : public StorableObject
{
public:

  /** @name Fehlerklassen */
    
  //@{
    /** Interne Fehlerklasse. */
    class NoPlayerWithThisIdException: public Exception {
    public:
      NoPlayerWithThisIdException()
        : Exception() {}
    };
  //@}

          
  /** @name Situationen anlegen und zerst\"oren */
    
  //@{
    /** Neue Situation. Objekte muessen spaeter hinzugefuegt werden. */
    Situation( int aSimStep = UNDEF_INT);
  
    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    virtual StorableObject* copy() const;
  //@}
 
       
  /** @name Werte abfragen */
    
  //@{
    /** Liefert den Zeittakt, auf den sich die Simulation bezieht. */
    int simStep() const;
  
    /** Gibt den Ball zurueck. Enthaelt die Situation keinen Ball,
      wird ein BallObjekt mit undefinierten Werten zurueckgegeben. */
    const ModelledBall& ball() const;
    
    /** Liefert den Spieler mit der Id aPlayerId.  Enthaelt die Situation
      keinen Spieler mit dieser Id, wird eine NoPlayerWithThisIdException
      erzeugt.

      Enthaelt die Situation (fehlerhafterweise) mehrere Spieler mit derselben
      PlayerId, wird der erste Spieler zurueckgegeben.

      Diese Routine ist langsam, d.h. die Laufzeit waechst linear mit der
      Anzahl der Spieler in der Situation.
      @param aPlayerId Die PlayerId des gesuchten Spielers.
      @exception NoPlayerWithThisIdException wenn kein Spieler mit dieser
      Id in der Situation vorkommt. */
    const ModelledPlayer& playerWithId( int aPlayerId ) const;

    /** Liefert die ben\"otigten Schritte zum Ball des Spielers mit der
      Id aPlayerId. Enthaelt die Situation
      keinen Spieler mit dieser Id, wird eine NoPlayerWithThisIdException
      erzeugt. \\
      Diese Routine ist langsam, d.h. die Laufzeit waechst linear mit der
      Anzahl der Spieler in der Situation.
      @param aPlayerId Die PlayerId des Spielers.
      @exception NoPlayerWithThisIdException wenn kein Spieler mit dieser
      Id in der Situation vorkommt. */
    const int stepsOfPlayerWithId( int aPlayerId ) const;

    /** Liefert den index-ten Player in der Situation. Ruft man diese Methode
      nacheinander mit allen Indizes zwischen 0 und playerCount()-1 auf, so
      werden insgesamt alle Spieler der
      Situation geliefert. Die Reihenfolge ist die, in der Spieler der
      Situation hinzugefuegt werden.

      Ist index < 0 oder >= playerCount(), wird eine
      Array::IndexOutOfBoundsException erzeugt.

      Diese Routine ist schnell, d.h. die Laufzeit ist unabhaengig von der Zahl
      der Spieler in der Situation. */
    const ModelledPlayer& playerWithIndex( int index ) const;

    /** Liefert den index, ueber den mit playerWithIndex( index ) auf einen
      Spieler zugegriffen werden kann. Enthaelt die Situation
      keinen Spieler mit dieser Id, wird eine NoPlayerWithThisIdException
      erzeugt.

      Diese Routine ist langsam, d.h. die Laufzeit waechst linear mit der
      Anzahl der Spieler in der Situation.
      @param aPlayerId Die PlayerId des Spielers mit dem gesuchten Index.
      @exception NoPlayerWithThisIdException wenn kein Spieler mit dieser
      Id in der Situation vorkommt. */
    int indexOfPlayerWithId( int aPlayerId ) const;

    /** Checkt, ob die Situation einen Spieler mit der angegebenen Id enthaelt.
      Diese Routine ist langsam, d.h. die Laufzeit waechst linear mit der
      Anzahl der Spieler in der Situation.
      @param aPlayerId Die PlayerId des gesuchten Spielers. */
    bool containsPlayerWithId( int aPlayerId ) const;

    /** Zahl der Spieler in der Situation. */
    int playerCount() const;
    
    /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
    virtual String toString() const;
  //@}
 
     
  /** @name Werte \"andern */
    
  //@{
    /** Setzt den Zeittakt, auf den sich die Simulation bezieht. */
    void setSimStep(int newSimStep) ;
  
    /** Setzt den Ball. */
    void setBall( const ModelledBall& newBall );

    /** Setzt die Daten des Spielers. 
      Ist bereits ein Spieler mit derselben PlayerId enthalten, wird er durch
      newPlayer ausgetauscht. Ist er noch nicht in der Situation enthalten, wird
      er hinzugefuegt.

      Diese Routine ist langsam, d.h. die Laufzeit waechst linear mit der
      Anzahl der Spieler in der Situation.  */
    void setPlayer( const ModelledPlayer& newPlayer );

    /** Ersetzt den index-ten Spieler durch newPlayer. 
      Ist der index playerCount(), wird der Spieler ohne weitere Tests (auf
      bereits vorhandene Id hinzugefuegt.

      Diese Routine ist schnell, d.h. die Laufzeit ist unabhaengig von der Zahl
      der Spieler in der Situation.
      @param newPlayer Spieler-Objekt, das in die Situation eingefuegt wird.
      @param index Stelle, an der abgelegt werden soll. index darf Werte 
      zwischen 0 und playerCount() enthalten.  */
    void setPlayer( const ModelledPlayer& newPlayer, int index );

    /** Fuegt einen Spieler hinzu. 
      Dies ist dasselbe wie \Ref{setPlayer}( newPlayer, playerCount() ) */
    void addPlayer( const ModelledPlayer& newPlayer );

    /** Entfernt einen Spieler aus der Situation.
      Alle Spieler mit einem hoeheren Index werden um 1 nach unten verschoben.
      D.h. evtl. vorher ermittelte Spielerindex-Werte sind nicht mehr g\"ultig.
      @param playerId Die Id des Spielers, der entfernt werden soll.
      @exception NoPlayerWithThisIdException, wenn kein Spieler mit dieser
      Id in der Situation vorkommt. */
    void removePlayerWithId( int aPlayerId );

    /** Entfernt einen Spieler aus der Situation.
      Alle Spieler mit einem hoeheren Index werden um 1 nach unten verschoben.
      D.h. evtl. vorher ermittelte Spielerindex-Werte sind nicht mehr g\"ultig.
      @param index Der Index des Spielers, der entfernt werden soll. */
    void removePlayerWithIndex( int index );

    /** Simuliert das Verstreichen von Simulationszeit. Die Routine geht davon
      aus, dass kein Spieler waehrend dieser Zeit Aktionen unternimmt. Es wird
      also nur die Positonsaenderung und die Gesschwindigkeitsabnahme von 
      Spielern und Ball berechnet.
      @param stepCount Zahl der Simulationsschritte. */
    void simulateSteps( unsigned int stepCount = 1);
  //@}
    
    
  /** @name Abfragen zur Erreichbarkeit des Balls */
  
  //@{
    /** Gibt den Index des Spielers zur\"uck, der den Ball am schnellsten 
      erreichen kann. 
      @param teamId TeamId des Teams dessen schnellster Spieler gefragt ist.
      Zul\"assige Werte sind OUR_TEAM_ID, THEIR_TEAM_ID und BOTH_TEAMS_ID.
      Letzterer Wert gilt per Default.
      @return indexOfPlayer index des Spielers der den Ball am schnellsten
      erreichen kann. 
      @exception NoPlayerWithThisIdException() falls es keinen Spieler in der
      Situation gibt, der den Anforderungen gen\"ugt.
      @exception InvalidIdException, wenn die teamId einen ung\"ultigen Wert 
      hat. */
    int indexOfNearestPlayer( int teamId = BOTH_TEAMS_ID ) const; 

    /** Gibt den n\"achsten Spieler zur\"uck, der den Ball am schnellsten 
      erreichen kann. 
      @param teamId TeamId des Teams dessen schnellster Spieler gefragt ist.
      Zul\"assige Werte sind OUR_TEAM_ID, THEIR_TEAM_ID und BOTH_TEAMS_ID.
      Letzterer Wert gilt per Default.
      @return ModelledPlayer Spieler der den Ball am schnellsten
      erreichen kann.
      @exception NoPlayerWithThisIdException() falls es keinen Spieler in der
      Situation gibt, der den Anforderungen gen\"ugt.
      @exception InvalidIdException, wenn die teamId einen ung\"ultigen Wert 
      hat. */
    const ModelledPlayer& nearestPlayer( int teamId = BOTH_TEAMS_ID ) const;

    /** Gibt die ben\"otigten Schritte zum Ball des Spielers zur\"uck der den
      Ball am schnellsten  erreichen kann. 
      @param teamId TeamId des Teams dessen schnellster Spieler gefragt ist.
      Zul\"assige Werte sind OUR_TEAM_ID, THEIR_TEAM_ID und BOTH_TEAMS_ID.
      Letzterer Wert gilt per Default.
      @return stepsOfPlayer Schritte des Spielers der den Ball am schnellsten
      erreichen kann. Wenn es keinen Spieler in der Situation gibt, der den
      Anforderungen gen\"ugt wird ModelledPlayer::MAX_STEPS_TO_BALL 
      zur\"uckgegeben.
      @exception InvalidIdException, wenn die teamId einen ung\"ultigen Wert 
      hat. */
    int stepsOfNearestPlayer( int teamId = BOTH_TEAMS_ID ) const; 
    
    /** Sagt aus, ob der angegebene Spieler einer derjenigen ist, der den Ball
      als schnellstes erreichen kann. YES wird zur\"uckgeben, wenn er der
      Spieler des eigenen Teams mit den wenigsten Schritten zum Ball ist
      oder er nur \Ref{ globalConf->interceptionTolerance } Schritte langsamer,
      als der schnellste Teamkamerad ist.
      @param id Id des Spielers 
      @return YES oder NO. 
      @exception NoPlayerWithThisIdException wenn kein Spieler mit dieser
      Id in der Situation vorkommt. */
    bool isOneOfTheNearestTeammatesToBall( int id ) const;

    /** Gibt das nearestTeammatesSorted-Array als const zurueck. */
    const Array<int>& getOurTeamSorted() const;

    /** Gibt das nearestEnemiesSorted-Array als const zurueck. */
    const Array<int>& getTheirTeamSorted() const;

    /** Gibt das stepsToBallOfPlayers-Array als const zurueck. */
    const Array<int>& getStepstoBall() const;
    
    /** Gibt die Entfernung des besten Spielers in Takten zum Ball zurueck. */
    const int playerStepsToBall(int index) const;
  //@}
  
  
  /** @name Abfragen zur Erreichbarkeit einer Position */
  
  //@{
    /** Gibt den Index des Spielers zur\"uck, der eine Position am schnellsten 
      erreichen kann.
      @param teamId TeamId des Teams dessen schnellster Spieler gefragt ist.
      Zul\"assige Werte sind OUR_TEAM_ID, THEIR_TEAM_ID und BOTH_TEAMS_ID.
      Letzterer Wert gilt per Default.
      @param pos Die Position, zu der der n\"achste Spieler gesucht ist.
      @return indexOfPlayer index des Spielers der die Position am schnellsten
      erreichen kann. 
      @exception NoPlayerWithThisIdException() falls es keinen Spieler in der
      Situation gibt, der den Anforderungen gen\"ugt.
      @exception InvalidIdException, wenn die teamId einen ung\"ultigen Wert 
      hat. */
    int indexOfNearestPlayerToPos( const Vector& otherPos, 
      const int teamId = BOTH_TEAMS_ID ) const;

    /** Gibt die Entfernung des n\"achsten Spielers zu einer Position  zur\"uck.
      @param teamId TeamId des Teams dessen n\"achster Spieler gefragt ist.
      Zul\"assige Werte sind OUR_TEAM_ID, THEIR_TEAM_ID und BOTH_TEAMS_ID.
      Letzterer Wert gilt per Default.
      @param pos Die Position, zu der der n\"achste Spieler gesucht ist.
      @return Entfernung des Spielers. der die Position am schnellsten
      erreichen kann, zu der Position. 
      @exception NoPlayerWithThisIdException() falls es keinen Spieler in der
      Situation gibt, der den Anforderungen gen\"ugt.
      @exception InvalidIdException, wenn die teamId einen ung\"ultigen Wert 
      hat. */
    double distOfNearestPlayerToPos( const Vector& otherPos, 
        const int teamId = BOTH_TEAMS_ID ) const;
  //@}
  
  
  /** @name Abfragen zum Abseits */
  
  //@{
    /** Testet, ob ein Teammate im Abseits steht.
      Die Routine liefert YES, wenn sich der angegebene Spieler jenseits der
      Abseitslinie befindet, sonst NO.

      Hinweis: Diese Routine kann unkorrekte Werte liefern. So kann sie YES
      liefern, obwohl der Spieler nicht im Abseits steht, wenn die Situation
      unvollst\"andig ist, also Spieler fehlen oder nicht als Gegner erkannt sind.
      Andererseits kann es sein, da\3 Abseits gepfiffen wird, weil ein Spieler
      dem Ball zu nahe kommt, f\"ur den diese Routine NO liefert. Das passiert
      dann, wenn sich seit der Ballabgabe die Abseitslinie \"uber den Spieler
      hinweg nach vorne (auf das eigene Tor zu) bewegt hat. In der Klasse
      Situation kann diese Abseitsfalle nicht erkannt werden.
      @param index Der Index des Spielers, von dem bestimmt werden soll, ob er im
      Abseits steht. Dieser Spieler darf nicht vom gegnerischen Team sein.
      @return YES oder NO
      @exception Array<ModelledPlayer>::IndexOutOfBoundsException, wenn der Index
      zu gro\3 oder zu klein ist.
      @exception InvalidIdException, wenn team() des Spielers THEIR_TEAM_ID
      @see offsideLineX */
    bool isOffsideWithIndex( int index ) const;
    
    /** Testet, ob ein Teammate im Abseits steht.
      @param index Die Id des Spielers, von dem bestimmt werden soll, ob er im
      Abseits steht. Dieser Spieler darf nicht vom gegnerischen Team sein.
      @return YES oder NO
      @exception InvalidIdException, wenn team() des Spielers THEIR_TEAM_ID ist
      oder nicht in der Situation enthalten ist.
      @see offsideLineX
      @see isOffsideWithIndex */
    bool isOffsideWithId( int playerId ) const;

    /** Liefert die X-Koordinate der Abseitslinie.
      Die Abseitslinie ist diejenige der beiden Geraden, die parallel zur
      Mittellinie  verlaufen und durch den Ball bzw. durch den vorletzten
      Spieler des gegnerischen Teams  gehen, die n\"aher am gegnerischen Tor
      liegt. Dabei werden nur die als Gegner identifizierten Spieler
      ber\"ucksichtigt. Die Abseitslinie darf allerdings nicht n\"aher am
      eigenen Tor liegen als die Mittellinie.    

      Der Wert wird intern zwischengespeichert und eigentlich von
      \Ref{computeOffsideX} berechnet.

      Hinweis: Diese Routine kann unkorrekte Werte liefern. Der Wert kann zu nahe
      an der Mittellinie liegen, wenn die Situation unvollst\"andig ist, also
      Spieler fehlen oder nicht als Gegner erkannt sind. Der Wert kann zu weit an
      der gegnerischen Torlinie liegen, wenn sich seit der Ballabgabe die wahre
      Abseitslinie nach vorne (auf das eigene Tor zu) bewegt hat. In der Klasse
      Situation kann diese Abseitsfalle nicht erkannt werden.
      @return x-Koordinate der Abseitslinie, ein Wert zwischen 0 (Mittellinie) und
      52.5 (gegnerische Torlinie). */
    double offsideLineX() const;
    
    /** Verlaesslichkeit der Abseitsdaten. 
      Es wird die schlechtere der Positionsverlaesslichkeiten der letzten beiden
      Spieler zurueckgegeben, wenn die fuer die Berechnung relevant waren, ist
      der Ball relevant, wird die Ballverlaesslichkeit, andernfalls die 
      eigene Positionsverlaesslichkeit zurueckgegeben. */
    const Reliability& Situation::offsideLineXReliability() const;

  //@}
  
private:
    
  // Instanzvariablen

  /** Der Simulationstakt der Situation. */
  int theSimStep;
  
  /** Array mit den Spielern.  */
  Array<ModelledPlayer> thePlayers;
  
  /** Der Ball. */
  ModelledBall theBall;

  /** Zustand, ob die Anzahl der Schritte, die die Spieler zum Ball brauchen,
    bereits gesch\"atzt wurden. Wird von den Methoden, die die Situation
    ver\"andern auf NO gesetzt und von estimateForAllStepsToBall auf YES
    gestzt */
  mutable bool areStepsToBallEstimated;

  /** Array, welches die Anzahl der Schritte der Spieler zum Ball enth\"alt.
    Die Indizes entsprechen dabei, den der Spieler im Array thePlayers. */
  mutable Array<int> stepsToBallOfPlayers;

  /** Array, welches die Indizes aller Spieler enth\"alt. Die Indizes sind
    dabei nach den Werten, die in stepsToBallOfPlayers stehen sortiert,
    d.h. der erste Index im Array entspricht dem Spieler, der den Ball am
    schnellsten erreichen kann (am wenigsten Schritte zu ihm ben\"otigt). */
  mutable Array<int> nearestAllPlayerSorted;

  /** Array, welches die Indizes aller Spieler des eigenen Teams enth\"alt.
    Die Indizes sind
    dabei nach den Werten, die in stepsToBallOfPlayers stehen sortiert,
    d.h. der erste Index im Array entspricht dem Spieler des eigenen Teams,
    der den Ball am
    schnellsten erreichen kann (am wenigsten Schritte zu ihm ben\"otigt). */
  mutable Array<int> nearestTeammatesSorted;

  /** Array, welches die Indizes aller Spieler des gegnerischen Teams
    enth\"alt. Die Indizes sind
    dabei nach den Werten, die in stepsToBallOfPlayers stehen sortiert,
    d.h. der erste Index im Array entspricht dem Spieler des gegnerischen
    Teams, der den Ball am
    schnellsten erreichen kann (am wenigsten Schritte zu ihm ben\"otigt). */
  mutable Array<int> nearestEnemiesSorted;
  
  /** X-Koordinate der Abseitslinie */
  mutable double theOffsideLineX;
  
  /** Verlaesslichkeit der X-Koordinate der Abseitslinie. */
  mutable Reliability theOffsideLineXReliability;
  
  /** Der Mindestabstand, der zwischen den Mittelpunkten von Spielern und
    Ball eingehalten werden mu\3. */
  static double theMinDistPlayerBall;
  
  // Hilfsfunktionen

  /** F\"ur alle Spieler wird gsch\"atzt, wieviele Schritte sie brauchen, um
    zum Ball zu gelangen, wenn sie sich optimal verhalten. Es werden die
    entsprechenden Arrays mit den gesch\"atzten Daten gef\"ullt. Diese Methode 
    setzt die Variable computeForAllStepsToBall auf YES */
  void estimateForAllStepsToBall() const;

  /** Sortiert den playerIndex in das sortedField ein. Sortiert wird dabei
    nach der Anzahl der Schritte, die die Spieler zum Ball brauchen.
    @param playerIndex index des Spielers der eingef\"ugt werden soll
    @param sortedField in das Feld in das einsortiert werden soll
    */
  void sortIntoField( int playerIndex, Array<int>& sortedField ) const;
  
  /** F\"uhrt die eigentliche Berechnung der Abseitslinie durch.
    @return X-Koordinate der Abseitslinie.
    @see offsideLineX */
  double computeOffsideLineX() const;
};

inline int Situation::simStep() const
{
  return theSimStep;
}

inline void Situation::setSimStep( int newSimStep )
{
  theSimStep = newSimStep;
}

inline const Reliability& Situation::offsideLineXReliability() const
{
  return theOffsideLineXReliability;
}

inline const ModelledBall& Situation::ball() const
{
  return theBall;
}

inline void Situation::setBall( const ModelledBall& newBall )
{
  theBall = newBall;
  areStepsToBallEstimated = NO;
}

inline int Situation::playerCount() const
{
  return thePlayers.count();
}

inline void Situation::addPlayer( const ModelledPlayer& newPlayer )
{
  thePlayers.addObject( newPlayer );
  areStepsToBallEstimated = NO;
  if ( newPlayer.team() == THEIR_TEAM_ID ) {
    theOffsideLineX = UNDEF_DOUBLE;
  }
}

inline void Situation::setPlayer( const ModelledPlayer& newPlayer, int index )
{
  thePlayers[index] = newPlayer;
  areStepsToBallEstimated = NO;
  if ( newPlayer.team() == THEIR_TEAM_ID ) {
    theOffsideLineX = UNDEF_DOUBLE;
  }
}

#endif











