#ifndef _WORLDMODEL_H_
#define _WORLDMODEL_H_

#include "Action.h"
#include "Common.h"
#include "Field.h"
#include "ModelledBall.h"
#include "ModelledPlayer.h"
#include "MovingObject.h"
#include "MovingPlayer.h"
#include "RasteredPlayground.h"
#include "Sensors.h"
#include "Situation.h"
#include "StableObject.h"
#include "ViewPoint.h"

/**
  Kapselung des Wissens eines Agenten ueber seine Umwelt und sich selbst. 

  Bemerkungen zu einzelnen Wissenselementen:
  
  Situation: Ein Objekt, in Modelle der dem Spieler bekannten, sich auf dem
  Spielfeld bewegenden Objekte enthalten sind. Diese Modelle (die von
  \Ref{ModelledObject} abstammen), enthalten u.A. die absoluten Koordinaten auf
  dem Spielfeld und die Geschwindigkeit.

  Initiative: Wer als naechster den Ball spielen darf. Die Mannschaft, die
  nicht die Initiative hat, muss warten, bis die Mannschaft mit Initiative den
  Ball beruehrt hat. Anfangs bekommt die auf der linken Seite spielende
  Mannschaft die Initiative.

  Die Klasse ist nicht zur Vererbung geeignet.
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/WorldModel.h} | [\URL[Implementierung]{src/WorldModel.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: WorldModel.h,v $
  *   $Revision: 1.63 $
  *   $Date: 1998/08/14 13:21:34 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class WorldModel
{
public:

  /** Interne Fehlerklasse. */
  class IncorrectOldSituationException: public Exception {
  public:
    IncorrectOldSituationException()
      : Exception() {}
  };
    
  /** @name M\"ogliche Spielzust\"ande
    Ein Wechsel des Spielzustands wird normalerweise vom Schiedsrichter
    bekanntgegeben. */
  //@{
    ///
    const int GAME_STATE_MIN = 0;
    ///
    const int BEFORE_KICK_OFF = 0;
    ///
    const int TIME_OVER  = 1;
    ///
    const int PLAY_ON = 2;
    ///
    const int KICK_OFF = 3;
    ///
    const int KICK_IN = 4;
    ///
    const int FREE_KICK = 5;
    ///
    const int CORNER_KICK = 6;
    ///
    const int GOAL_KICK = 7;
    ///
    const int GOAL = 8;
    ///
    const int FOUL = 9;
    ///
    const int HALF_TIME = 10;
    ///
    const int TIME_UP = 11;
    ///
    const int TIME_UP_WITHOUT_A_TEAM = 12;
    ///
    const int TIME_EXTENDED = 13;
    ///
    const int OFFSIDE = 14;
    ///
    const int DROP_BALL = 15;
    ///
    const int GAME_STATE_MAX = 15;
  //@}
  
  /** @name M\"ogliche Entscheidungsgr\"unde. */
  //@{
    ///
    const int DECISION_REASON_MIN = 0;
    ///
    const int DECISION_REASON_SEE_INFO = 0;
    ///
    const int DECISION_REASON_REFEREE_DECISION = 1;
    ///
    const int DECISION_REASON_PLAN_ENDED = 2;
    ///
    const int DECISION_REASON_MAX = 2;
  //@}

  /** @name WorldModel anlegen */
  //@{
  
    /** Erzeugt ein neues Weltmodell, das seine Informationen aus der
      Sensorik-Komponente Sensors bezieht. Das Weltmodell legt ein internes
      Feld (Field) an, welches die Positionen der moeglichen Flaggen 
      enthaehlt und eine Berechnung des eigenen Blickpunktes aus gesehenen
      Objekten erlaubt.
      @param Pointer auf das aktive Sensorik-Objekt */
    WorldModel( Sensors *sensors );

  //@}

  
  /** @name Anfragen zum Agenten */
  //@{

    /** Liefert die Spielernummer des Agenten. */
    int playerNumber() const;
    
    /** Liefert den Namen des Teams, zu dem der Agent gehoert. */
    const String& teamName() const;

    /** Liefert die Seite des Agenten. */
      int side() const;

    /** Liefert die Id des Agenten. */
    int id() const;
    
    /** Liefert die Basisposition des Agenten. 
      Die BasisPosition wird anfangs aus der Konfigurationsdatei eingelesen,
      kann sich aber waehrend des Spiels \"andern. Je nach Spielzustand und
      Initiative wird eine andere Basisposition geliefert. Werden daf\"ur keine
      Werte angegeben, werden die aktuellen Werte aus dem Weltmodell 
      verwendet.
      @param aGameState Einer der m\"oglichen Spielzust\"ande.
      @param anInitiative Einer der Werte von in \Ref{Teams}.
      @return absoluter Vektor zur Basisposition.*/
    Vector basePosition( int aGameState = UNDEF_INT, 
      int anInitiative = UNDEF_INT ) const;

    /** Liefert die Anfangsposition des Agenten. 
      Je nach Spielzustand und Initiative wird eine andere Anfangsposition
      geliefert. Werden daf\"ur keine Werte angegeben, werden die aktuellen
      Werte aus dem Weltmodell verwendet. Die Anfangsposition wird aus der
      Konfigurationsdatei ausgelesen und \"andert sich nur bei einem
      Rollentausch. Als Laufziel sollte immer ein Wert von \Ref{basePosition}()
      verwendet werden.
      @param aGameState Einer der m\"oglichen Spielzust\"ande.
      @param anInitiative Einer der Werte von in \Ref{Teams}.
      @return absoluter Vektor zur Anfangsposition.*/
    Vector initialPosition( int aGameState = UNDEF_INT, 
      int anInitiative = UNDEF_INT ) const;
  //@}
  
        
  /** @name Werte des Agenten setzen */
  //@{
  
    /** Setzt die Spielernummer des Agenten.
      Ein ModelledPlayer mit seiner Id wird in die aktuelle 
      Situation aufgenommen, falls er dort noch nicht vorhanden ist.
      Die Anfangs- und die Basispositionen  werden aus \Ref{globalConf} gelesen und 
      gesetzt. */
    void setPlayerNumber( int newPlayerNumber );

    /** Setzt den Namen des Teams, zu dem der Agent gehoert. */
    void setTeamName( const String& newTeamName );

    /** Setzt die Seite, auf der der Agent spielt. */
    void setSide( int newSide );

    /** Setzt die Basisposition des Agenten.
      Hier m\"ussen der Spielzustand und die Initiative angegeben werden. Gilt
      dieselbe Position f\"ur verschiedene Zust\"ande, so modifiziert das
      Setzen in einem dieser Zust\"ande die Position f\"ur alle diese
      Zust\"ande.
      @param aGameState Einer der m\"oglichen Spielzust\"ande.
      @param anInitiative Einer der Werte von in \Ref{Teams}. */
    void setBasePosition( int aGameState, int anInitiative, const Vector& pos );

    /** Setzt die Anfangsposition des Agenten.
      Hier m\"ussen der Spielzustand und die Initiative angegeben werden. Gilt
      dieselbe Position f\"ur verschiedene Zust\"ande, so modifiziert das
      Setzen in einem dieser Zust\"ande die Position f\"ur alle diese
      Zust\"ande.
      
      Nur die Initialisierungs- und der Rollentauschroutine setzen die
      Anfangspositionen um.
      @param aGameState Einer der m\"oglichen Spielzust\"ande.
      @param anInitiative Einer der Werte von in \Ref{Teams}. */
    void setInitialPosition( int aGameState, int anInitiative, const Vector& pos );
  
  //@}
    
  
  /**@name Anfragen zur aktuellen Situation */
  //@{
  
    /** Liefert aktuelle Daten ueber den gewuenschten Spieler. 
      Wird kein Spieler angegeben, werden die Daten ueber den Agenten selbst
      geliefert.

      @param playerId die Id des gewuenschten Spielers. (default ist id() )
      @exception Situation::NoPlayerWithThisIdException, wenn der angegebene
      Spieler in der aktuellen Situation nicht existiert.
      @exception InvalidIdException, wenn die playerId ungueltig ist.
      @return ModelledPlayer mit den gewuenschten Daten. */
    const ModelledPlayer& player( int playerId = SELF_ID ) const;

    /** Liefert aktuelle Daten ueber den Ball. 
      Wenn es keine Daten ueber den  Ball gibt, wird ein BallObjekt mit
      undefinierten Werten zurueckgegeben. */
    const ModelledBall& ball() const;

    /** Liefert den Simulationstakt der aktuellen Situation. 
      Die zu diesem Zeittakt gehoerende Situation wird von situation() 
      geliefert. */
    int simStep() const;  

    /** Liefert aktuelle, vergangene und zukuenftige Situationen.
      Es wird eine gewisse, vom historyCount-Parameter in agent.conf
      abhaengige Zahl von alten Situationen gespeichert. Aufgrund von
      Synchronisationsproblemen oder eines Spielzustands, in dem das Spiel
      nicht laeuft, lassen sich nicht immer alle Situationen bis zum Takt
      simStep-historyCount abfragen.

      Die haeufigste Verwendung ist sicher die Frage nach der aktuellen
      Situation. Der Parameter timeStepOffset ist per Default auf 0 gesetzt, so
      dass mit situation() auf die aktuelle Situation zugegriffen werden kann.

      @param timeStepOffset Mit timeStepOffset < 0 koennen vergangene
      (d.h. gespeicherte) Situationen abgerufen werden. Mit timeStepOffset >
      0 koennen zukuenftige Situationen vorausberechnet werden. Zulaessige
      Werte liegen zwischen -historyCount und +historyCount (wird in
      agent.conf gesetzt).
      @exception Array<Situation>::IndexOutOfBoundsException, wenn der
      timeStepOffset einen ungueltigen Wert hatte.
      @return Eine Referenz, die die angefragte Situation enth\"alt. Diese
      veraendert sich moeglicherweise mit Aenderung des simSteps(). */
    const Situation& situation( int timeStepOffset = 0 ) const;

    /** Liefert den Spielzustand zurueck. Dieser ist als int kodiert. */
    int gameState() const;

    /** Liefert den Entscheidungsgrund zur\"uck. Dieser kann die Werte
      DECISION_REASON_SEE_INFO, DECISION_REASON_REFEREE_DECISION oder
      DECISION_REASON_MIN_PLAN_ENDED annehmen. */
    int decisionReason() const;

    /** Liefert die Initiative.
      @return Einer der Werte OUR_TEAM_ID, THEIR_TEAM_ID, BOTH_TEAMS_ID,
        NO_TEAM_ID */
    int initiative() const;

    /** Gibt an, ob ein bestimmtes Team die Initiative hat. 
      Diese Funktion erleichtert das Umgehen mit der Initiative, da mehrfache
      Abfragen unn\"otig werden. Wir haben die Initiative, wenn, die Initiative
      auf OUR_TEAM_ID oder auf BOTH_TEAMS_ID steht. Entsprechendes gilt fuer
      THEIR_TEAM_ID.
      @param aTeamId Das Team, von erfragt werden soll, ob es die Initiative
        hat. M\"ogliche Werte sind OUR_TEAM_ID, THEIR_TEAM_ID.
      @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
        hat.
      @return YES oder NO. */
    bool hasInitiative( int aTeamId ) const;

    /** Sagt aus, ob der eigene Spieler einer derjenigen ist, der den Ball
      als schnellstes erreichen kann. YES wird zur\"uckgeben, wenn er der
      Spieler des eigenen Teams mit den wenigsten Schritten zum Ball ist
      oder er nur \Ref{ globalConf->interceptionTolerance } Schritte langsamer,
      als der schnellste Teamkamerad ist.
      @return YES oder NO. */
    bool amIOneOfTheNearestPlayerToBall() const;

    /** Liefert die Zahl der Tore, die ein Team bisher erzielt hat.
      @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
      @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
      hat.
      @return Die Zahl der Tore dieses Teams. */
    int score(int aTeamId) const;

    /** Liefert das Feld, auf dem der Agent spielt. */
    const Field& field() const;
    
   /** Ermoeglicht den Zugriff auf den RasteredPlayground, der aktuelle
      Werte ueber Aufenthaltshaeufigkeiten von Spielern und Ball enthaelt. */
   RasteredPlayground& rasteredPlayground();

    /** Der simStep, an dem das letzte Catch-Kommando an den Server
      geschickt wurde. */
    int simStepOfLastCatch() const;

    /** Der simStep, an dem das letzte Say-Kommando an den Server
      geschickt wurde. */
    int simStepOfLastSay() const;

    /** Der simStep, an dem die letzte Schiedsrichterentscheidung
      eingetroffen ist. */
    int simStepOfLastRefereeDecision() const;


    /** Liefert die Position des Mittelpunktes des gegenerischen Tores. */
    const Vector& enemyGoal() const;

    /** Liefert die Position des Mittelpunktes des eignen Tores. */
    const Vector& ourGoal() const;

    /** Liefert die Position des oberen Pfosten des eignen Tores. */
    const Vector& ourGoalTop() const;

    /** Liefert die Position des unteren Pfosten des eignen Tores. */
    const Vector& ourGoalBottom() const;
    
    /** Liefert die Rolle des Agenten.
      Die Rolle bestimmt das Verhalten des Spielers. Anfangs wird die Rolle aus
      der Konfigurationsdatei eingelesen. Steht dort ein ung\"ultiger Wert,
      wird auch von dieser Funktion ein ung\"ultiger Wert geliefert.
      @return Der Wert einer der Rollenkonstanten.
      @see \Ref{Rollen} */
    int role() const;
    
  //@}
  
  
  /** @name Werte der aktuellen Situation setzen */
  //@{
    
    /** Setzt den Simulationstakt der aktuellen Situation. 
      Diese Funktion sollte mit Vorsicht verwendet werden, damit nicht die
      interne Synchronisation des Weltmodells durcheinander geraet. */
    void setSimStep( int newValue );

    /** Setzt den Spielzustand. */
    void setGameState( int newGameState );

    /** Setzt den Entscheidungsgrund. Erlaubte Werte sind
      DECISION_REASON_SEE_INFO, DECISION_REASON_REFEREE_DECISION und
      DECISION_REASON_MIN_PLAN_ENDED. */
    void setDecisionReason( int newDecisionReason );

    /** Setzt die Initiative. 
      Dabei kann auch eine Seite angegeben werden.  Spielen wir gerade auf
      dieser Seite, wird die Initiative auf OUR_TEAM_ID gesetzt, andernfalls
      auf THEIR_TEAM_ID.

      @param newValue Entweder die Id eines Teams ( OUR_TEAM_ID, THEIR_TEAM_ID,
      BOTH_TEAMS_ID ) oder die Id einer Seite (LEFT_SIDE_ID, 
      RIGHT_SIDE_ID ) */
    void setInitiative(int newValue);

    /** Setzt die Zahl der Tore fuer ein Team.
      @param newScore Die neue Zahl der Tore
      @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
      @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
      hat */
    void setScore( int newScore, int aTeamId );

    /** Setzt die Rolle eines Spielers.
      Die Rolle bestimmt das Verhalten des Spielers. Anfangs wird die Rolle aus
      der Konfigurationsdatei eingelesen. Die G\"ultigkeit der Rolle wird nicht
      \"uberpr\"uft.
      @param newRole neue Rolle des Spielers.
      @see \Ref{Rollen} */
    void setRole( int newRole );
  //@}  
    
    
  /** @name Verbindung zu Sensors und Effektors */
  //@{
    /** Integriert die Sensorinformationen in das Weltmodell.
      Dabei wird eine neue Situation erzeugt und evtl. der Spielzustand
      geaendert. Sowohl Sicht- als auch Hoerinformationen werden ausgewertet.
      @see synchronizeSituations
      @see modelOfSelf
      @see addSeenObjects
      @see mergeOldSituationsIntoSituation
      @see processAudioInfo */
    void integrateSensorInfo();

    /** Integriert Schiedsrichtermeldungen in das Weltmodell.
      Dabei wird der Spielzustand der aktuellen Situation
      geaendert. Es werden nur Hoerinformationen ausgewertet.
      @see processAudioInfo */
    void integrateHearInfo();
    
    /** Registriert, da\3 eine Aktion abgesendet wurde.
      F\"uhrt gegebenenfalls ein updateWithAction durch.
      @param anAction Die abgeschickte Aktion */
    void registerExecution( const Action& anAction );
      

    /** Integriert die Folgen der Ausf\"uhrung einer Aktion.
      Ist die Aktion eine Hauptaktion, wird eine neue Situation unter der
      Annahme erzeugt, dass der Agent in der aktuellen Situation die angegebene
      Aktion  ausfuehrt. Die neue Situation wird zur aktuellen Situation. Ist
      die Aktion eine Nebenaktion, wird keine neue Situation angelegt! 
      @param Die Aktion, die integriert werden soll. */
    void updateWithAction( const Action& anAction );  

    /** F\"ugt Informationen aus der anderen Situation in die 
      vertrauensw\"urdigere Situation ein.
      Die \Ref{Reliablity} Objekte bilden die Grundlage f\"ur das Fusionieren
      der einzelnen Objekte. Verglichen werden nur Objekte die gleich sind.
      Es werden jeweils die aktuelleren Daten dieser Objekte \"ubernommen, bei
      Gleichheit siegt das Datum, da\3 der vertrauensw\"urdigeren Situation
      entspricht. Daten der anderen Situation, die keine Entsprechung in der
      vertauensw\"urdigen Situation haben werden nicht \"ubernommen.
      @param trustedSituation vertrauensw\"urdige Situation dessen Daten als
      Grundlage angesehen werden.
      @param anotherSituation andere Situation, dessen Daten einverleibt werden
      sollen. 
      @see mergeBalls
      */
    void mergeSituations( Situation& trustedSituation,
                          Situation& anotherSituation ) const;
  //@}
  
  /** @name Abfragen zur Vorzugsrichtung */
  //@{
    /** Liefert die bevorzugte Kickrichtung.
      Diese "Vorzugsrichtung" ist abh\"angig von der Position des Spielers.
      Durch sie werden wichtige Eigenschaften des Mannschaftsspiels umgesetzt,
      z.b. Flanken und Angriffe \"uber die Fl\"ugel.
        @param pos Position, f\"ur die die Vorzugsrichtung berechnet werden
          soll.
        @return Ein Richtungsvektor. Die Richtungskomponente ( \Ref{length()})
          direkt der kickBall-Skill \"ubergeben werden. Die Laenge gibt die
          optimale Ballgeschwindigkeit in m/simStep an. Die Division der Laenge
          durch globalConf->kickPowerRate ergibt eine Kraft, die kickBall
          \"ubergeben werden kann.*/
    Vector preferredKickDirection( const Vector& pos ) const;
    
    /** Liefert die Region, in der sich eine Position befindet.
      Ja nach Region kann sich ein Spieler anders verhalten. Auch die
      Definition der Vorzugsrichtungsfunktion basiert auf der Region, in der
      sich eine Position befindet. 
        @param pos Position, f\"ur die die Region ermittelt werden soll.
        @return Die Nummer der Region, eine Zahl zwischen 1 und 10. */
    int region( const Vector& pos ) const;
    
  //@}
      
  /** Verschl\"usselt die gegebene Zahl zu einem String
    @param number Die zu codierende Zahl. */
  String encode( int number );

 /** Entschl\"usselt den gegebenen String zu einer Zahl
   @param string Der zu decodierende String. */
  int decode( String string );
 

  /** Beschreibung als String fuer die Ausgabe. */
  String toString() const;
  
private:
  
  // Instanzvariablen
    
  /** Zeiger auf das Sensorik-Objekt */
  Sensors *theSensors;
  
  /** Das Feld, auf dem der Spieler spielt. */
  Field theField;

  /** Array, das vergangene (und zukuenftige) Situationen enthaelt. */
  mutable Array<Situation> theSituations;

  /** Index der aktuellen Situation im Array */
  int currentSituationIndex;

  /** Zahl der vergangenen Situtationen, die aufgehoben werden sollen und
    gleichzeitig auch Zahl der in die Zukunft simulierbaren Aktionen. */
  int historyCount;

  /** Interne Feldgroesse fuer das Array der Situationen. */
  int situationFieldSize;

  /** Der Teamname */
  String theTeamName;

  /** Die Spielernummer */
  int thePlayerNumber;

  /** Die Seite des Spielfelds, auf der das Tor des Teams des Agenten 
    steht. */
  int theSide;

  /** Die Zahl der bisher gefallenen Tore  */
  int theScore[2];

  /** Das Team, das die Initiative besitzt. */
  int theInitiative;

  /** Der aktuelle Spielzustand */
  int theGameState;

  /** Der aktuelle Entscheidungsgrunf. */
  int theDecisionReason;
  
  /** simStep des letzten catch-Kommandos */
  int theSimStepOfLastCatch;

  /** simStep des letzten say-Kommandos */
  int theSimStepOfLastSay;
  
  /** simStep der letzten Schiedsrichterentscheidung */
  int theSimStepOfLastRefereeDecision;

  
  /** Die Basispositionen des Agenten.
    Sie sind anfangs identisch mit den Anfangspositionen, k\"onnen sich aber
    w\"ahrend des Spiels durch Lernen ver\"andern. 
    Index 0: Position bei eigenem Ansto\3
    
    Index 1: Position bei Ansto\3 des Gegners
    
    Index 2: Position w\"ahrend des normalen Spiels. */
  Array<Vector> theBasePositions;

    /** Die Anfangspositionen des Agenten.
      Sie werden aus der Konfigurationsdatei ausgelesen.
      
    Index 0: Position bei eigenem Ansto\3
    
    Index 1: Position bei Ansto\3 des Gegners
    
    Index 2: Position w\"ahrend des normalen Spiels. */
  Array<Vector> theInitialPositions;
  
  /** Puffert die abgesendeten Aktionen. */
  Array<Action> theExecutedActions;
   
  /** \"Ausseres Limit. 
    y-Koordinate einer waagrechten Linie 6m vom Feldrand. */
  static const double OUTER_LIMIT;
  
  /** Inneres Limit. 
    y-Koordinate einer waagrechten Linie 14m vom Feldrand. */
  static const double INNER_LIMIT;
  
  /** Die Rolle des Agenten. */
  int theRole;
  
  /** RasteredPlayground verwaltet online erfasste Daten der gegnerischen
    Spielerpositionen und des Balles. 
    Es bietet die Mglichkeit eine passive Position zu ermitteln, die der Agent
    whrend des Spiels einnehmen kann.*/
  RasteredPlayground theRasteredPlayground;
  
   
  // Private Funktionen

  /** Setzt den simStep, an dem das letzte Catch-Kommando an den Server
      geschickt wurde. */
    void setSimStepOfLastCatch( int newSimStep );

  /** Setzt den simStep, an dem das letzte Say-Kommando an den Server
      geschickt wurde. */
    void setSimStepOfLastSay( int newSimStep );
  
  /** Setzt den simStep, an dem die letzte Schiedsrichterentscheidung
    eingetroffen ist. */
    void setSimStepOfLastRefereeDecision( int newSimStep );
  
  /** Speichert eine Situation als die (neue) gegenwaertige ab. 
    Die bisherige gegenwaertige Situation ist nach dem Aufruf mit 
    situation( -1 )  erreichbar. 
    @param aSituation Die zu speichernde Situation */  
  void addCurrentSituation( const Situation& aSituation );
  
  /** Entfernt die aktuelle Situation aus dem Weltmodell. 
    Dies kann sinnvoll sein, wenn man aus Versehen zu weit in die Zukunft
    simuliert hat und den aktuellen Stand wieder herstellen m\"ochte. */
  void removeCurrentSituation();

  /** Die alten Situationen werden f\"ur die Integration von neuen
    Sensorinformationen aufbereitet.
    Wenn die aktuelle Situation des Weltmodells
    zeitlich vor dem angegebenen Simulationsschritt liegt, dann werden
    Situationen bis zu dem Simulationsschritt dazugef\"ugt und simuliert.
    Wenn die aktuelle Situation
    dagegen einen h\"oheren Simulationsschritt als den angegebenen hat und
    die Zeit im SoccerServer laut Spielzustand weiter gelaufen ist,
    dann werden die Situationen des Weltmodells auf diesen Simulationsschritt
    zur\"uckgesetzt; wenn die Zeit im SoccerServer nicht weitergelaufen ist,
    dann werden die Zeitindizes der entsprechenden Situation zur\"uckgesetzt.
    @param simStep Simulationsschritt f\"ur die die Situationen aufbereitet
      werden sollen.
    @exception IncorrectOldSituationException wenn die Differenz des
      Zeitstempels der aktuellen Situation mit simStep vom Betrag her
      gr\"oser als \Ref{historyCount} ist, also keine entsprechenden alten
      Situationen vorliegen. */
  void synchronizeSituations( int simStep );
  
  /** Liefert einen \Ref{ModelledPlayer}, der die Daten
    \"uber den Spieler selbst enth\"alt.
    Diese Daten stammen entweder aus den gesehenen Informationen oder,
    wenn diese Informationen nicht ausreichen, aus den alten Situation.
    @return ModelledPlayer Spieler mit Daten \"uber den Spieler selbst */
  const ModelledPlayer modelOfSelf() const;
  
  /** F\"ugt alle gesehenen Objekte in die Situation ein.
    Diese Daten entstammen nur den gesehenen Informationen.
    @param newSituation Situation in die Daten eingef\"ugt werden sollen. */
  void addSeenObjects( Situation& newSituation );

  /** Sucht zum angegebenen Spieler einen korrespondierenden in der 
    angegebenen Situation. Bei der Suche nach dem korrespondierenden Spieler,
    werden nur Spieler mit bekannter Nummer betrachtet. Zur\"uckgegeben wird
    die Position an der sich der korrespondierenden Spieler in der Situation
    befindet. Gibt es keinen korrespondierenden Spieler wird eine
    NoCorrespondingPlayerException erzeugt.
    @param player Spieler zu dem ein korrespondierender erfragt wird
    @param situation Situation in der der korrespondierende Spieler gesucht
    wird
    @exception NoCorrespondingPlayerException falls kein korrespondierender
    Spieler in der Situation auffindbar ist
    @return index des korrespondierenden Spielers in der Situation situation
    */
  int indexOfCorrespondingPlayer( ModelledPlayer player, 
                                  Situation situation ) const;

  /** F\"ugt Informationen des anderen Balles in den vertrauensw\"urdigeren
    Ball ein.
    @param trustedBall vertrauensw\"urdiger Ball dessen Daten als
    Grundlage angesehen werden.
    @param anotherBall anderer Ball, dessen Daten einverleibt werden sollen.
    @param mySelf der Spieler, der den Ball sehen m\"u\3te, wenn 
    er vor ihm liegt. */
  ModelledBall mergeBalls( const ModelledBall& trustedBall, 
                           const ModelledBall& anotherBall, 
                           const ModelledPlayer& mySelf ) const;

  /** F\"ugt Informationen des anderen Spielers in den vertrauensw\"urdigeren
    Spieler ein.
    @param trustedPlayer vertrauensw\"urdiger Spieler dessen Daten als
    Grundlage angesehen werden.
    @param anotherPlayer anderer Spieler, dessen Daten einverleibt werden
    sollen.
    @param mySelf der Spieler, der den Spieler sehen m\"u\3te, wenn 
    er in seinem Blickfeld ist. */
  ModelledPlayer mergePlayers( const ModelledPlayer& trustedPlayer, 
                               const ModelledPlayer& anotherPlayer ) const;
  
  /** Berechnet die relative Position des Objekts. 
    Benoetigt werden die gesehene Position des Objektes und der eigene 
    Blickwinkel. 
    @param seenPos Vektor, der von der eigenen Position zum Objekt zeigt. Er
    ist also zu interpretieren in einem Koordinatensystem, dessen y-Achse
    in die Blickrichtung zeigt und in dessen Ursprung die eigene Position
    liegt.
    @param Der eigene Blickwinkel 
    @return die relative Position des Objektes */
  Vector relativePos( const Vector& seenPos, double faceDir );

  /** Berechnet die relative Geschwindigkeit eines Objekts.
    Benoetigt werden die relative Position des Objektes und desen
    gesehene Bewegungsaenderung.
    @param relativePos Vektor, der von der eigenen Position zum Objekt zeigt,
    aber im globalen Koordinatensystem ist. 
    @param seenChangeOfPos Vektor, der die relative Winkel- und 
    Entfernungsaenderung des Objekts enthaelt
    @return die relative Geschwindigkeit des Objektes im globalen
    Koordinatensystem. */
  Vector relativeSpeed( const Vector& relativePos,
			const Vector& seenChangeOfPos );

  /** Verarbeitet Zurufe. 
    Zurufe kommen entweder von einem selbst, von einem anderen Spieler des
    eigenen Teams, vom anderen Team oder vom Schiedsrichter.
    @param newSituation Fuehren Nachrichten von einem Teammitglied zu neuen
    Informationen ueber die aktuelle Situation, so werden sie in
    newSituation eingefuegt.
    @param Array mit den gehoerten Nachrichten */
  void processAudioInfo( Situation& newSituation, 
			 const Array<Message>& messages );

  /** Verarbeitet Schiedsrichterentscheidungen. 
    Dabei werden Spielzustand und Initiative neu gesetzt.
    @param message Die Nachricht des Schiedsrichters */  
  void processRefereeMessage( const Message& message );
  

  /** Verarbeitet Zurufe eines Mitspielers. 
    @param message Die Nachricht des Mitspielers */  
  void processOurTeamMessage( Situation& newSituation,
                              const Message& message );

  /** Liefert einen internen Index, an dem die Position 
    des Agenten fuer den angegebenen Spielzustand/Initiative steht. 
    Mit diesem Index kann auf Elemente von theBasePositions und
    theInitialPositions zgegriffen werden.
    @return Zahl zwischen 0 und 10
    */
  int positionIndexFromState( int aGameState, int anInitiative ) const;
  
  /** F\"uhrt Konsistenzchecks durch und pa\3t die aktuelle Situation und
    das restlichen Weltmodell an.
    Es werden Aktionen nachsimuliert, die schon abgesendet wurden, aber
    noch nicht im SoccerServer simuliert wurden.
    Die Stamina-Werte werden nochmal aus den BodyInfos ausgelesen, weil
    das Nachsimulieren diese fehlerhaft ver\"andert.
    Im  Tournamentmode wird der GameState ggf. auf PLAY_ON gestellt. */
  void improveConsistency();

  // Exceptions

  /** Interne Exception, die aufgerufen wird, wenn keine korrespondierenden
    Objekte gefunden werden k\"onnen */
  class NoCorrespondingPlayerException: public Exception {
  public:
    NoCorrespondingPlayerException()
      : Exception() {}
  };  
};


/** Global verfuegbares Weltmodell.  
  Enthaelt die Annahmen des Spielers ueber den realen Zustand  seiner Umwelt. */
extern WorldModel *globalWorldModel;

// inline Funktionsdefinitionen;

inline void WorldModel::setTeamName( const String& newTeamName )
{
  theTeamName = newTeamName;
}

inline const String& WorldModel::teamName() const
{
  return theTeamName;
}

inline int WorldModel::playerNumber() const 
{
  return thePlayerNumber;
}

inline void WorldModel::setSimStep( int newValue )
{
  theSituations[currentSituationIndex].setSimStep( newValue );
}

inline int WorldModel::simStep() const 
{
  return theSituations[currentSituationIndex].simStep();
}

inline void WorldModel::setSide( int newSide ) 
{
  theSide = newSide;
}

inline int WorldModel::side() const 
{
  return theSide;
}

inline int WorldModel::initiative() const
{
  return theInitiative;
}

inline int WorldModel::id() const
{
  return ( thePlayerNumber == UNDEF_INT) ? 
    UNDEF_PLAYER_ID : MIN_TEAMMATE_ID - 1 + thePlayerNumber;
}

inline void WorldModel::setGameState( int newGameState )
{
  theGameState = newGameState;
}

inline int WorldModel::gameState() const
{
  return theGameState;
}

inline void WorldModel::setDecisionReason( int newDecisionReason )
{
  theDecisionReason = newDecisionReason;
}

inline int WorldModel::decisionReason() const
{
  return theDecisionReason;
}

inline const Field& WorldModel::field() const
{
  return theField;
}

inline const Vector& WorldModel::enemyGoal() const
{
  return theField.posOfPoint( RIGHT_GOAL_CENTER_FLAG_ID );
}
inline const Vector& WorldModel::ourGoal() const
{
    return theField.posOfPoint( LEFT_GOAL_CENTER_FLAG_ID );
}

inline const Vector& WorldModel::ourGoalTop() const
{
    return theField.posOfPoint( LEFT_GOAL_TOP_FLAG_ID );
}

inline const Vector& WorldModel::ourGoalBottom() const
{
    return theField.posOfPoint( LEFT_GOAL_BOTTOM_FLAG_ID );
}

inline RasteredPlayground& WorldModel::rasteredPlayground()
{
  return theRasteredPlayground;
}

inline int WorldModel::simStepOfLastCatch() const
{
  return theSimStepOfLastCatch;
}

inline int WorldModel::simStepOfLastSay() const
{
  return theSimStepOfLastSay;
}

inline void WorldModel::setSimStepOfLastCatch( int newSimStep )
{
  theSimStepOfLastCatch = newSimStep;
}

inline void WorldModel::setSimStepOfLastSay( int newSimStep )
{
  theSimStepOfLastSay = newSimStep;
}


inline int WorldModel::simStepOfLastRefereeDecision() const
{
  return theSimStepOfLastRefereeDecision;
}

inline void WorldModel::setSimStepOfLastRefereeDecision( int newSimStep )
{
  theSimStepOfLastRefereeDecision = newSimStep;
}


#endif



